#!/bin/bash

# Paths are configured for the script to be run from the top 
# Ctalk build directory.

CONF_USER=`cat build/.confname`

CTALK=`which ctalk`

CTALKDIR=$HOME/.ctalk
CTALKCACHEDIR=$HOME/.ctalk/libcache
CLASSDIR=`cat ./build/.includedir`


# if [ ! -z $LOGNAME ]; then
#    ACTUAL_USER=$LOGNAME
#else if [ ! -z $EUID ]; then
#    ACTUAL_USER=$EUID
#fi
#fi

# 
# These are the classes that are used by preload.ca.  If rewriting
# preload.ca, then also check this list.  We want to replace only
# the base classes when installing.
#
PRELOADCLASSES="Application ArgumentList Array Boolean Character \
	Collection Event Exception FileStream Float Integer \
	Key List LongInteger Magnitude Method Object \
	ObjectInspector ReadFileStream SignalEvent SignalHandler \
    Stream String Symbol SystemErrnoException WriteFileStream"


if [ -f build/.nopreload ]; then
    echo
    echo Skipping base class preloading.
    echo
    exit 0
fi


echo 
echo Preloading base classes.
echo 

if [ -z $CONF_USER ]; then
    echo
    echo Warning: Cannot determine the actual user name. You
    echo can still use Ctalk without preloading during installation, 
    echo but it will run more slowly the first few times you compile 
    echo programs.
    echo
    exit 1
fi

if [ -d $CTALKCACHEDIR ]; then
    for i in $PRELOADCLASSES; do
	rm -f $CTALKCACHEDIR/$i*;
    done
fi

$CTALK --progress -I$CLASSDIR build/preload.ca -o build/preload.i

#
# Set some reasonable args at first.
#
echo -n -I$CLASSDIR >$CTALKDIR/args

chown -R $CONF_USER $CTALKDIR

echo
echo
