/* $Id: bitmap.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include <Xlib.h>
#include "x11defs.h"

extern Display *display;

#define RECTANGLE_GCV_MASK (GCFunction|GCFillStyle|\
			    GCForeground|GCBackground) 
int __xlib_clear_pixmap (Drawable d, unsigned int width, 
			 unsigned int height, GC gc){
  XGCValues rectangle_gcv, old_gcv;
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  rectangle_gcv.function = GXcopy;
  rectangle_gcv.fill_style = FillSolid;
  rectangle_gcv.background = old_gcv.background;
  rectangle_gcv.foreground = old_gcv.background;
  XChangeGC (display, gc, RECTANGLE_GCV_MASK, &rectangle_gcv);
  XFillRectangle (display, d, gc,
		  0, 0, width, height);
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

int __xlib_clear_rectangle_basic (Drawable d, int x, int y, 
				  unsigned int width, unsigned int height,
				  GC gc) {
  XGCValues rectangle_gcv, old_gcv;
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  rectangle_gcv.function = GXset;
  rectangle_gcv.fill_style = FillSolid;
  rectangle_gcv.background = old_gcv.background;
  rectangle_gcv.foreground = old_gcv.background;
  XChangeGC (display, gc, RECTANGLE_GCV_MASK, &rectangle_gcv);
  XFillRectangle (display, d, gc,
 		  x, y, width, height);
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

int __ctalkX11CreatePixmap (int parent,
			    int width,
			    int height, 
			    int depth) {
  Pixmap p;
  GC gc;
  XGCValues v;
  p = XCreatePixmap (display, 
		     (Drawable)parent, width, height, depth);
  v.function = GXcopy;
  v.fill_style = FillSolid;
  v.background = v.foreground = 
    BlackPixel (display, DefaultScreen (display));
  gc = XCreateGC (display, p, 
		  GCForeground|\
		  GCBackground|\
		  GCFunction|\
		  GCFillStyle, &v);
  XFillRectangle (display, p, gc,
		  0, 0, width, height);
  XFreeGC (display, gc);
  return (int)p;
}

void __ctalkX11DeletePixmap (int id) {
  XFreePixmap (display, (Drawable)id);
}

void __ctalkX11FreeGC (int gc_ptr) {
  XFreeGC (display, (GC)gc_ptr);
}

unsigned int __ctalkX11CreateGC (int drawable) {
  GC gc;
  XGCValues v;
  gc = XCreateGC (display, (Drawable)drawable, 0, &v);
  return (int)gc;
}

int __ctalkX11CreatePaneBuffer (OBJECT *pane,
				int width, int height, 
				int depth) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object,
    *xWindowID_object;
  OBJECT *xGC_object;
  GC gc;
  int win_id;
  char buf[MAXMSG];
  Pixmap p;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((xGC_object = 
	__ctalkGetInstanceVariable (pane_object, "xGC", TRUE))
       == NULL)
    return ERROR;
  sscanf (xGC_object->instancevars->__o_value, "0x%p", &gc);
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((xWindowID_object = 
       __ctalkGetInstanceVariable (pane_object, "xWindowID", TRUE))
      == NULL)
    return ERROR;
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  win_id = atoi (xWindowID_object->instancevars->__o_value);
  p = XCreatePixmap (display, win_id, width, height, depth);
  __xlib_clear_pixmap (p, width, height, gc);
  sprintf (buf, "%d", (int)p);
  __ctalkSetObjectValueVar (paneBuffer_object, buf);
  p = XCreatePixmap (display, win_id, width, height, depth);
  __xlib_clear_pixmap (p, width, height, gc);
  sprintf (buf, "%d", (int)p);
  __ctalkSetObjectValueVar (paneBackingStore_object, buf);
  return SUCCESS;
}

int __ctalkX11FreePaneBuffer (OBJECT *pane) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  XFreePixmap (display, 
	       atoi(paneBackingStore_object->instancevars->__o_value));
  XFreePixmap (display, 
	       atoi(paneBuffer_object->instancevars->__o_value));
  return SUCCESS;
}
int __ctalkX11ResizePaneBuffer (OBJECT *pane,
				int width, int height) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object,
    *xWindowID_object;
  OBJECT *xGC_object;
  int win_id;
  char buf[MAXMSG];
  unsigned int p_old;
  Pixmap p;
  GC gc;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((xWindowID_object = 
       __ctalkGetInstanceVariable (pane_object, "xWindowID", TRUE))
      == NULL)
    return ERROR;
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  if ((xGC_object = 
	__ctalkGetInstanceVariable (pane_object, "xGC", TRUE))
       == NULL)
    return ERROR;
  win_id = atoi (xWindowID_object->instancevars->__o_value);
  p = XCreatePixmap (display, win_id, width, height, CopyFromParent);
  sscanf (paneBuffer_object->instancevars->__o_value, "%u", &p_old);
  sscanf (xGC_object->instancevars->__o_value, "0x%p", &gc);
  XCopyArea (display, (Pixmap)p_old, p, gc, 
	     0, 0, 
	     width, height,
	     0, 0);
  sprintf (buf, "%u", (unsigned int)p);
  __ctalkSetObjectValueVar (paneBuffer_object, buf);
  XFreePixmap (display, (Pixmap)p_old);
  sscanf (paneBackingStore_object->instancevars->__o_value, "%u", &p_old);
  p = XCreatePixmap (display, win_id, width, height, CopyFromParent);
  XCopyArea (display, (Pixmap)p_old, p, gc, 
	     0, 0, 
	     width, height,
	     0, 0);
  sprintf (buf, "%u", (unsigned int)p);
  __ctalkSetObjectValueVar (paneBackingStore_object, buf);
  XFreePixmap (display, (Pixmap)p_old);
  return SUCCESS;
}

int __ctalkX11ClearBufferRectangle (OBJECT *pane,
				    int x, int y,
				    int width, int height) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object,
    *xWindowID_object, *xGC_object;
  GC gc;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((xGC_object = 
       __ctalkGetInstanceVariable (pane_object, "xGC", TRUE))
      == NULL)
    return ERROR;
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  sscanf (xGC_object->instancevars->__o_value, "0x%p", &gc);
  __xlib_clear_rectangle_basic 
    (atoi(paneBuffer_object->instancevars->__o_value), 
     x, y, width, height, gc);
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  __xlib_clear_rectangle_basic 
    (atoi(paneBackingStore_object->instancevars->__o_value), 
     x, y, width, height, gc);
  if ((xWindowID_object = 
       __ctalkGetInstanceVariable (pane_object, "xWindowID", TRUE))
      == NULL)
    return ERROR;
  __xlib_clear_rectangle_basic 
    (atoi(xWindowID_object->instancevars->__o_value), 
     x, y, width, height, gc);
  return SUCCESS;

}

int __get_pane_buffers (OBJECT *pane_object, int *panebuffer_xid,
			int *panebackingstore_xid) {
  OBJECT *paneBuffer_instance_var,
    *paneBackingStore_instance_var,
    *paneBuffer_xID_instance_var,
    *paneBackingStore_xID_instance_var;
  *panebuffer_xid = *panebackingstore_xid = 0;
  if ((paneBuffer_instance_var =
       __ctalkGetInstanceVariable (pane_object, 
				   PANEBUFFER_INSTANCE_VAR,
				   FALSE)) == NULL)
    return ERROR;
  if ((paneBackingStore_instance_var =
       __ctalkGetInstanceVariable (pane_object, 
				   PANEBACKINGSTORE_INSTANCE_VAR,
				   FALSE)) == NULL)
    return ERROR;
  if ((paneBuffer_xID_instance_var =
       __ctalkGetInstanceVariable (paneBuffer_instance_var, 
				   X11BITMAP_ID_INSTANCE_VAR,
				   FALSE)) == NULL) {
    *panebuffer_xid = 
      atoi(paneBuffer_instance_var -> instancevars -> __o_value);
  } else {
    *panebuffer_xid = 
      atoi(paneBuffer_xID_instance_var -> instancevars -> __o_value);
  }

  if ((paneBackingStore_xID_instance_var =
       __ctalkGetInstanceVariable (paneBackingStore_instance_var, 
				   X11BITMAP_ID_INSTANCE_VAR,
				   FALSE)) == NULL) {
    *panebackingstore_xid = 
      atoi(paneBackingStore_instance_var -> instancevars -> __o_value);
  } else {
    *panebackingstore_xid = 
      atoi(paneBackingStore_xID_instance_var -> 
	   instancevars -> __o_value);
  }
  return SUCCESS;
}

#else

int __ctalkX11CreatePaneBuffer (OBJECT *pane_object,
				  int width, int height, int depth) {
  return SUCCESS;
}

int __ctalkX11FreePaneBuffer (OBJECT *pane_object) {
  return SUCCESS;
}

int __ctalkX11ResizePaneBuffer (OBJECT *pane_object,
				int width, int height) {
  return SUCCESS;
}

int __ctalkX11ClearBufferRectangle (OBJECT *pane_object,
				    int x, int y,
				    int width, int height) {
  return SUCCESS;
}

#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
