/* $Id: guiclearrectangle.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  __ctalkX11 ... are here for backward compatibility but are
 *  deprecated.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include "x11defs.h"

extern Display *display;   /* Defined in x11lib.c. */
extern char *shm_mem;
extern int mem_id;

#if X11LIB_FRAME
int __ctalkX11PaneClearRectangle (OBJECT *self_object, int x, int y,
				   int width, int height) {
  return SUCCESS;
}
int __ctalkGUIPaneClearRectangle (OBJECT *self_object, int x, int y,
				   int width, int height) {
  return SUCCESS;
}
#else /* X11LIB_FRAME */

int __ctalkX11ClearRectangleBasic (int drawable_id,
				   int gc_ptr, 
				   int x, int y,
				   int width, int height) {
  char d_buf[MAXLABEL], gc_buf[MAXLABEL], 
    w_buf[MAXLABEL];
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif
  sprintf (gc_buf, "%#x", (unsigned int) gc_ptr);
  sprintf (w_buf, "%d", drawable_id);
  /* __xlib_clear_rectangle in x11lib.c
   * clears the rectangle in a pixmap if we send it
   * buffer XIDs.
   */
  sprintf (d_buf, "%d:%d:%d:%d:%d:%d:", x, y, 
	   width, height, drawable_id, drawable_id);
  make_req (shm_mem, PANE_CLEAR_RECTANGLE_REQUEST,
	    w_buf, gc_buf, d_buf);

#ifdef GRAPHICS_WRITE_SEND_EVENT
  send_event.xgraphicsexpose.type = GraphicsExpose;
  send_event.xgraphicsexpose.send_event = True;
  send_event.xgraphicsexpose.display = display;
  send_event.xgraphicsexpose.drawable = atoi(win_id_value->__o_value);
  XSendEvent (display, atoi(win_id_value->__o_value),
 	      False, 0L, &send_event);
#endif
  wait_req (shm_mem);
  return SUCCESS;
}

int __ctalkX11PaneClearRectangle (OBJECT *self, int x, int y,
				   int width, int height) {
  return __ctalkGUIPaneClearRectangle (self, x, y, width, height);
}
int __ctalkGUIPaneClearRectangle (OBJECT *self, int x, int y,
				   int width, int height) {
  OBJECT *self_object, *win_id_value, *gc_value;
  int panebuffer_xid, panebackingstore_xid;
  char d_buf[MAXLABEL], xid_buf[0xff];
  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  win_id_value = __x11_pane_win_id_value_object (self_object);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  __get_pane_buffers (self_object, &panebuffer_xid,
		      &panebackingstore_xid);
  if (panebuffer_xid && panebackingstore_xid) {
    sprintf (d_buf, "%d:%d:%d:%d:%d:%d:", 
	     x, y, width, height,
	     panebuffer_xid, panebackingstore_xid); 
    sprintf (xid_buf, "%d", panebuffer_xid);
    make_req (shm_mem, PANE_CLEAR_RECTANGLE_REQUEST, 
	      xid_buf,
	      gc_value->__o_value, d_buf);
  } else {
    sprintf (d_buf, "%d:%d:%d:%d:%d:%d:",
 	     x, y, width, height, 0, 0);
    make_req (shm_mem, PANE_CLEAR_RECTANGLE_REQUEST,
 	      win_id_value->__o_value,
 	      gc_value->__o_value, d_buf);
  }
  wait_req (shm_mem);
  return SUCCESS;
}

#endif /* X11LIB_FRAME */
#else /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
static void __gui_support_error (void) {
  _error ("__ctalkGUI ... () functions require Graphical User Interface support.\n");
int __ctalkX11PaneClearRectangle (OBJECT *self_object, int x, int y,
				  int width, int height) {__gui_support_error();return ERROR;}
int __ctalkGUIPaneClearRectangle (OBJECT *self_object, int x, int y,
				  int width, int height) {__gui_support_error();return ERROR;}
}
#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
