/* $Id: guidrawpoint.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  __ctalkX11PaneDrawPoint is here for backward compatibility but is
 *  deprecated.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include "x11defs.h"

extern Display *display;   /* Defined in x11lib.c. */
extern char *shm_mem;
extern int mem_id;

#if X11LIB_FRAME
int __ctalkX11PaneDrawPoint (OBJECT *self, OBJECT *pane, OBJECT *pen) {
  return __ctalkGUIPaneDrawPoint (self, pane, pen);
}
int __ctalkGUIPaneDrawPoint (OBJECT *self, OBJECT *pane, OBJECT *pen) {
  return SUCCESS;
}
#else /* X11LIB_FRAME */
int __ctalkX11PaneDrawPoint (OBJECT *self, OBJECT *pane, OBJECT *pen) {
  return __ctalkGUIPaneDrawPoint (self, pane, pen);
}
int __ctalkGUIPaneDrawPoint (OBJECT *self, OBJECT *point, OBJECT *pen) {
  OBJECT *self_object, *point_object, *pen_object, *pen_width_object,
    *point_x_var, *point_y_var, *pen_color_object;
  int panebuffer_xid, panebackingstore_xid;
  OBJECT *win_id_value, *gc_value;
  char colorname[MAXMSG];
  char d_buf[MAXLABEL];
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif

  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  win_id_value = __x11_pane_win_id_value_object (self_object);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  pen_object = (IS_VALUE_INSTANCE_VAR (pen) ? pen->__o_p_obj : pen);
  point_object = (IS_VALUE_INSTANCE_VAR (point) ? point->__o_p_obj : point);
  pen_width_object = 
    __ctalkGetInstanceVariable (pen_object, "width", TRUE);
  pen_color_object = 
    __ctalkGetInstanceVariable (pen_object, "colorName", TRUE);
  strcpy (colorname, pen_color_object->instancevars->__o_value);
  point_x_var = 
    __ctalkGetInstanceVariable (point_object, "x", TRUE);
  point_y_var = 
    __ctalkGetInstanceVariable (point_object, "y", TRUE);
  __get_pane_buffers (self_object, &panebuffer_xid,
		      &panebackingstore_xid);
  sprintf (d_buf, "%d:%d:%d:%d:%d:%s", 
	   atoi(point_x_var->instancevars->__o_value),
	   atoi(point_y_var->instancevars->__o_value),
	   atoi (pen_width_object->instancevars->__o_value),
	   panebuffer_xid, panebackingstore_xid,
	   colorname);
  make_req (shm_mem, PANE_DRAW_POINT_REQUEST,
	    win_id_value->__o_value,
	    gc_value->__o_value, d_buf);
#ifdef GRAPHICS_WRITE_SEND_EVENT
  send_event.xgraphicsexpose.type = GraphicsExpose;
  send_event.xgraphicsexpose.send_event = True;
  send_event.xgraphicsexpose.display = display;
  send_event.xgraphicsexpose.drawable = atoi(win_id_value->__o_value);
  XSendEvent (display, atoi(win_id_value->__o_value),
 	      False, 0L, &send_event);
#endif
  wait_req (shm_mem);
  return SUCCESS;
}
#endif /* X11LIB_FRAME */
#else /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
static void __gui_support_error (void) {
  _error ("__ctalkGUI ... () functions require Graphical User Interface support.\n");
}
int __ctalkX11PaneDrawPoint (OBJECT *self, OBJECT *pane, OBJECT *pen) {
  return __ctalkGUIPaneDrawPoint (self, pane, pen);
}
int __ctalkGUIPaneDrawPoint (OBJECT *self, OBJECT *pane, OBJECT *pen) {
  __gui_support_error();return ERROR;
}
#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
