/* $Id: guixpm.c,v 1.2 2014/05/09 16:24:13 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2014  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  __ctalkX11PanePutStr and __ctalkGUIPanePutStr are here for 
 *  compatibility, but they rely on hard-coded instance variable names.  
 *  Try to use __X11PanePutStrBasic in new code.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include "x11defs.h"

extern Display *display;   /* Defined in x11lib.c. */
/* extern char *shm_mem; */
/* extern int mem_id; */
/* extern char *fixed_font_name; */
/* extern Font fixed_font; */

#if X11LIB_FRAME

int __ctalkGUIXPMFromData (Drawable d, int x_org, int y_org, 
			   int width, int height, 
			   char **xpm_data) {
  return SUCCESS;
}

#else /* X11LIB_FRAME */

#define HEXASCIITOBINARY(d) (\
  (d >= '0' && d <= '9') ? d - 48 :				\
  ((d >= 'A' && d <= 'F') ? d - 55 :				\
   ((d >= 'a' && d <= 'f') ? d - 87 : 0)))

int __ctalkGUIXPMFromData (int drawable_id, int gc_ptr, 
			   int x_org, int y_org, 
			   int width, int height, 
			   char **data) {
  int x, y, x_out, y_out, nth_color,
    have_color;
  int n_colors, chars_per_pixel;
  int color_1_row, color_n_row;
  char old_color_pixels[3] = "";
  unsigned long int c;
  XColor xcolor;
  XGCValues gcv;

  sscanf (data[0], "%d %d %d %d",
   	  &width, &height, &n_colors, &chars_per_pixel);

  color_1_row = 1;
  color_n_row = n_colors;

  for (y_out = 0, y = color_n_row + 1; y <= (color_n_row + height);
       ++y_out, y++) {
    for (x = 0, x_out = 0; x_out < width; ++x_out, x += chars_per_pixel) {
      int rr = 0, gg = 0, bb = 0;
      if (!strncmp (&data[y][x], old_color_pixels, chars_per_pixel)) {
   	// Re-use a color if it's the same as the previous pixel.
	/* SetPixel (hdc, x_out, y_out, c); */
      } else {
   	for (nth_color = color_1_row, have_color = FALSE;
   	     (nth_color <= color_n_row) && !have_color; nth_color++) {
   	  if (!strncmp (&data[y][x], data[nth_color], chars_per_pixel)) {
   	    char *p, *q;
   	    if ((p = strchr ((char *)data[nth_color], 'c')) != NULL) {
   	      if ((q = strchr (p, '#')) != NULL) {
   		// A #rrggbb color spec
   		rr = (HEXASCIITOBINARY(*(q + 1)) << 4) + 
   		  HEXASCIITOBINARY(*(q + 2));
   		gg = (HEXASCIITOBINARY(*(q + 3)) << 4) + 
   		  HEXASCIITOBINARY(*(q + 4));
   		bb = (HEXASCIITOBINARY(*(q + 5)) << 4) + 
   		  HEXASCIITOBINARY(*(q + 6));
   		have_color = TRUE;

		xcolor.red = rr << 8;
		xcolor.green = gg << 8;
		xcolor.blue = bb << 8;
		XAllocColor (display, 
			     DefaultColormap (display, 
					      DefaultScreen (display)),
			     &xcolor);
		c = xcolor.pixel;
		gcv.foreground = gcv.background = c;
		XChangeGC (display, (GC)gc_ptr, 
			   GCForeground|GCBackground, &gcv);
		XDrawPoint (display, drawable_id, (GC)gc_ptr, x_out, y_out);
		switch (chars_per_pixel)
		  {
		  case 1:
		    old_color_pixels[0] = data[y][x];
		    old_color_pixels[1] = '\0';
		    break;
		  case 2:
		    old_color_pixels[0] = data[y][x];
		    old_color_pixels[1] = data[y][x+1];
		    old_color_pixels[2] = '\0';
		    break;
		  }
   	      } else {
   		// A named color.  TODO
   	      }
   	    }
   	  }
   	}
      }
    } // for (x = 0, x_out = 0; x_out < width; ++x_out, x += chars_per_pixel)
  } // for (y_out = 0, y = color_n_row + 1; y <= (color_n_row + height); ...

}

#endif /* X11LIB_FRAME */

#else /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
static void __gui_support_error (void) {
  _error ("__ctalkGUI ... () functions require Graphical User Interface support.\n");
int __ctalkGUIPanePutStr (OBJECT *self_object, int x, int y, char *s) {
  __gui_support_error(); return ERROR;
}
#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
