/* $Id: ismethd.c,v 1.2 2014/05/02 10:31:37 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(__sparc__) && defined(__svr4__)
#include <strings.h>       /* Prototype of bcmp. */
#endif
#include <ctype.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

static inline METHOD *ismethd_get_class_method_by_name (OBJECT *o, const char *name, int warn) {

  OBJECT *class, *value_var;
  METHOD *m;

  if (!IS_OBJECT(o))
    return NULL;

  if (str_eq (o -> __o_classname, (char *)"Expr")) {
    /* if ((value_var = __ctalkGetInstanceVariable (o, "value", FALSE))  */
    /* 	== NULL) { */
    /*   if (warn) { */
    /* 	_warning ("ismethd_get_class_method_by_name: %s: Undefined value.\n", */
    /* 		  o -> __o_name); */
    /*   } */
    /*   return NULL; */
    /* } */
    value_var = o -> instancevars;
    if (!value_var || !(value_var -> attrs & OBJECT_IS_VALUE_VAR)) {
      if (warn) {
     	_warning ("ismethd_get_class_method_by_name: %s: Undefined value.\n",
     		  o -> __o_name);
      }
      return NULL;
    }
    class = value_var -> __o_class;
   } else {
     class = o -> __o_class;
   }

   if (!class) {
     if (warn) {
       _warning ("ismethd_get_class_method_by_name: Undefined class.\n",
 		o -> __o_classname);
     }
     return NULL;
   }

   /* namelength = strlen (name); */
   for (m = class -> class_methods; m; m = m -> next) {
     if (str_eq (m -> name, (char *)name))
       return m;
     if (!m -> next)
       break;
   }
 
   if (class -> __o_superclass)
     return ismethd_get_class_method_by_name (class -> __o_superclass, name, warn);
   else
     return NULL;
}

static inline METHOD *ismethd_get_instance_method_by_name (OBJECT *o, 
					const char *name, int warn) {
  OBJECT *class, *value_var;
  register METHOD *m;

  if (!IS_OBJECT(o)) return NULL;

  /* 
   * Check for the name, "value," in case an object's
   * first instance var is different, as with some 
   * Collections, which have all Key objects as 
   * instancevars.
   */
  if (o -> instancevars) {
    if (str_eq (o -> instancevars -> __o_name, "value"))
      value_var = o -> instancevars;
    else
      value_var = o;
  } else {

    value_var = o;
  }

  if ((class=
       ((IS_CLASS_OBJECT(value_var)) ? value_var : value_var->__o_class))
      != NULL) {
    for (m = class -> instance_methods; m; m = m -> next) {
      if (str_eq (m -> name, (char *)name))
	return m;
    }
    for (class = class -> __o_superclass; 
	 class; class = class -> __o_superclass) {
      for (m = class -> instance_methods; m; m = m -> next) {
	if (str_eq (m -> name, (char *)name))
	  return m;
      }
    }
  }

  return NULL;

}

static inline METHOD *ismethd_find_class_method_by_name (OBJECT **rcvr, 
				      const char *name, int warn) {

  int i;
  OBJECT *tmp_rcvr;
  METHOD *m = NULL;

  if ((m = ismethd_get_class_method_by_name (*rcvr, name, warn)) == NULL) {
    for (i = __ctalkGetReceiverPtr () + 1; !m && (i <= MAXARGS); i++) {
      tmp_rcvr = __ctalk_receivers_at (i);
      if ((m = 
	   ismethd_get_class_method_by_name (tmp_rcvr, name, warn))
	  != NULL) {
	*rcvr = tmp_rcvr;
      }
    }
  }

  if (warn && !m)
    _warning
      ("ismethd_find_class_method_by_name: %s: %s (Class %s) does not understand %s.\n",
       __argvFileName (), (*rcvr) -> __o_name, (*rcvr) -> __o_classname, name);

  return m;
}

static inline METHOD *ismethd_find_instance_method_by_name (OBJECT **rcvr, 
					 const char *name, int warn) {

  int i;
  OBJECT *tmp_rcvr;
  METHOD *m = NULL;

  if ((m = ismethd_get_instance_method_by_name (*rcvr, name, warn)) == NULL) {
    for (i = __ctalkGetReceiverPtr () + 1; !m && (i <= MAXARGS); i++) {
      tmp_rcvr = __ctalk_receivers_at (i);
      if ((m = 
	   ismethd_get_instance_method_by_name(tmp_rcvr, name, warn))
	  != NULL) {
	*rcvr = tmp_rcvr;
      }
    }
  }

  if (IS_OBJECT(*rcvr)) {
    if (!m) {
      if ((!(*rcvr) -> instancevars) ||
	  ((*rcvr) -> instancevars && 
	   /***/
	   /* !str_eq ((*rcvr) -> instancevars -> __o_name, (char *)"value"))) { */
	   !((*rcvr) -> instancevars -> attrs & OBJECT_IS_VALUE_VAR))) {
	OBJECT *collection_value;
	collection_value = *rcvr;
	m = ismethd_get_instance_method_by_name (collection_value, name, warn);
      }
    }
  }

  if (warn && !m)
    _warning
      ("ismethd_find_instance_method_by_name: %s: %s (Class %s) does not understand %s.\n",
       __argvFileName (), (*rcvr) -> __o_name, (*rcvr) -> __o_classname, name);

  return m;
}

static inline METHOD *ismethd_find_method_by_name (OBJECT **rcvr_p, const char *name, int warn) {

  METHOD *m = NULL;

  if (((m = ismethd_find_instance_method_by_name (rcvr_p, name, warn)) == NULL) &&
      ((m = ismethd_find_class_method_by_name (rcvr_p, name, warn)) == NULL))
    if (warn)
_warning ("ismethd_find_method_by_name: Object %s (Class %s) does not understand method %s.\n",
  (*rcvr_p) -> __o_name, (*rcvr_p) -> __o_classname, name);

return m;
}

int __ctalk_isMethod (char *name, MESSAGE_STACK messages, int msg_ptr, 
	       int stack_top, int stack_end) {
  int prev_tok_ptr;
  MESSAGE *m_prev_tok;
  METHOD *m;
  OBJECT *m_prev_tok_obj;

  if (M_TOK(messages[msg_ptr]) == METHODMSGLABEL)
    return TRUE;

  if ((prev_tok_ptr = __ctalkPrevLangMsg (messages, msg_ptr, stack_top))
      != ERROR) {
    m_prev_tok = messages[prev_tok_ptr];
    if (M_OBJ_IS_VAR(m_prev_tok)) {

      if (m_prev_tok && m_prev_tok -> obj) {
 	if (IS_OBJECT (m_prev_tok -> obj -> instancevars) &&
	    str_eq (m_prev_tok->obj->instancevars->__o_name,
	     	    (char *)"value")) {
	  m_prev_tok_obj = m_prev_tok -> obj -> instancevars;
	  if ((m = 
	       ismethd_find_method_by_name (&m_prev_tok_obj, 
					name, FALSE)) != NULL)
	    return TRUE;
	} else {
	  if (IS_VALUE_INSTANCE_VAR(m_prev_tok->obj)) {
	    m_prev_tok_obj = m_prev_tok -> obj;
	    if ((m = 
		 ismethd_find_method_by_name (&m_prev_tok_obj, 
					  name, FALSE)) != NULL)
	      return TRUE;
	  } else {
	    /*
	     *  Should only be needed for Collection
	     *  subclasses that don't have value instance 
	     *  variables.
	     */
 	    if ((!m_prev_tok -> obj -> instancevars) ||
 		(m_prev_tok -> obj -> instancevars &&
 		 !str_eq (m_prev_tok -> obj -> instancevars -> __o_name,
			  (char *)"value"))) {
	      m_prev_tok_obj = m_prev_tok -> obj;
	      if ((m = 
		   ismethd_find_method_by_name (&m_prev_tok_obj, 
					    name, FALSE)) != NULL)
		return TRUE;
	    }
	  }
	}
      } else { /* if (m_prev_tok && m_prev_tok -> obj)  */
	if (m_prev_tok && m_prev_tok -> value_obj) {
 	  if (IS_OBJECT (m_prev_tok -> value_obj -> instancevars) &&
	      str_eq (m_prev_tok->value_obj->instancevars->__o_name, 
		      (char *)"value")) {
	    m_prev_tok_obj = m_prev_tok -> value_obj -> instancevars;
	    if ((m = 
		 ismethd_find_method_by_name (&m_prev_tok_obj, 
					  name, FALSE)) != NULL)
	      return TRUE;
	  } else {
	    if (IS_VALUE_INSTANCE_VAR(m_prev_tok->value_obj)) {
	      m_prev_tok_obj = m_prev_tok -> value_obj;
	      if ((m = 
		   ismethd_find_method_by_name (&m_prev_tok_obj, 
					    name, FALSE)) != NULL)
		return TRUE;
	    } else {
	      /*
	       *  Should only be needed for Collection
	       *  subclasses that don't have value instance 
	       *  variables.
	       */
 	      if ((!m_prev_tok -> value_obj -> instancevars) ||
 		  (m_prev_tok -> value_obj -> instancevars && 
 		   !str_eq (m_prev_tok -> value_obj -> instancevars -> __o_name,
			    (char *)"value"))) {
		m_prev_tok_obj = m_prev_tok -> value_obj;
		if ((m = 
		     ismethd_find_method_by_name (&m_prev_tok_obj, 
					      name, FALSE)) != NULL)
		  return TRUE;
	      }
	    }
	  }
	} else { /* if (m_prev_tok && m_prev_tok -> value_obj)  */
	  if ((M_TOK(m_prev_tok) == METHODMSGLABEL) &&
	      (m_prev_tok -> value_obj &&
	       (m_prev_tok -> value_obj == 
		messages[msg_ptr] -> receiver_obj))) {
	    if ((m_prev_tok_obj = messages[msg_ptr] -> receiver_obj) != NULL){
	      if ((m = 
		   ismethd_find_method_by_name (&m_prev_tok_obj, 
					    name, FALSE)) != NULL)
		return TRUE;
	    }
	  }
	}
      }

    } else {
      /*
       *  The previous label is a method - check for a value_obj
       *  result also.
       */
      if (M_TOK(m_prev_tok) == METHODMSGLABEL) {
	if (m_prev_tok && IS_OBJECT (M_VALUE_OBJ(m_prev_tok))) {
	  m_prev_tok_obj = M_VALUE_OBJ(m_prev_tok);
	  if ((m = ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
	      != NULL) {
	    return TRUE;
	  }  else {
	    /*
	     *  Check for a method from an instance variable value class.
	     */
	    if (m_prev_tok_obj && HAS_INSTANCEVAR_CLASS(m_prev_tok_obj)) {
	      if (IS_OBJECT(m_prev_tok_obj -> instancevars)) {
		if ((m = ismethd_find_method_by_name 
		     (&(m_prev_tok_obj->instancevars), name, FALSE))
		    != NULL)
		  return TRUE;
	      }
	    }
	  }
	} else {
	  /*
	   *  If the previous method doesn't have a result yet, look
	   *  up the method's return class.
	   */
	  if (IS_OBJECT(m_prev_tok->receiver_obj)){
	    METHOD *m_prev_tok_method;
	    OBJECT *m_prev_tok_rcvr_obj;
	    m_prev_tok_rcvr_obj = m_prev_tok->receiver_obj;
	    if ((m_prev_tok_method = 
		 ismethd_find_method_by_name (&m_prev_tok_rcvr_obj,
					  M_NAME(m_prev_tok), FALSE))
		!= NULL) {
	      OBJECT *m_prev_return_class;
	      if (str_eq (m_prev_tok_method -> returnclass, "Any")) {
		m_prev_return_class = 
		  ((m_prev_tok->receiver_obj) ?
		   (m_prev_tok->receiver_obj->__o_class) :
		   (NULL));
	      } else {
		if (m_prev_tok_method -> returnclass) {
		  m_prev_return_class = 
		    __ctalkGetClass(m_prev_tok_method->returnclass);
		} else {
		  m_prev_return_class = m_prev_tok_rcvr_obj -> __o_class;
		}
	      }
	      if (ismethd_find_method_by_name (&m_prev_return_class, 
					   name, FALSE)) {
		return TRUE;
	      }
	    }
	  }
	}
      } else {
	if (m_prev_tok -> attrs & TOK_IS_AGGREGATE_MEMBER_LABEL) {
	  if ((m_prev_tok_obj = m_prev_tok -> value_obj) != NULL) {
	    if ((m = ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
		!= NULL) {
	      return TRUE;
	    }
	  }
	} else {
	  m_prev_tok_obj = m_prev_tok -> obj;
	  if (m_prev_tok && IS_OBJECT (m_prev_tok -> obj)) {
	    if ((m = ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
		!= NULL) {
	      return TRUE;
	    }  else {
	      /*
	       *  Check for a method from an instance variable value class.
	       */
	      if (m_prev_tok_obj && HAS_INSTANCEVAR_CLASS(m_prev_tok_obj)) {
		if (IS_OBJECT(m_prev_tok_obj -> instancevars)) {
		  if ((m = ismethd_find_method_by_name 
		       (&(m_prev_tok_obj->instancevars), name, FALSE))
		      != NULL)
		    return TRUE;
		}
	      }
	    }
	  } else {
	    m_prev_tok_obj = m_prev_tok -> value_obj;
	    if (m_prev_tok && IS_OBJECT (m_prev_tok -> value_obj)) {
	      if ((m = 
		   ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
		  != NULL) {
		return TRUE;
	      }  else {
		/*
		 *  Check for a method from an instance variable value class.
		 */
		if (m_prev_tok_obj && HAS_INSTANCEVAR_CLASS(m_prev_tok_obj)) {
		  if (IS_OBJECT(m_prev_tok_obj -> instancevars)) {
		    if ((m = ismethd_find_method_by_name 
			 (&(m_prev_tok_obj->instancevars), name, FALSE))
			!= NULL)
		      return TRUE;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  return FALSE;
}

int __ctalk_op_isMethod (char *name, MESSAGE_STACK messages, int msg_ptr, 
	       int stack_top, int stack_end) {
  int prev_tok_ptr;
  MESSAGE *m_prev_tok;
  METHOD *m;
  OBJECT *m_prev_tok_obj;

  if ((prev_tok_ptr = __ctalkPrevLangMsg (messages, msg_ptr, stack_top))
      != ERROR) {
    m_prev_tok = messages[prev_tok_ptr];
    if (M_OBJ_IS_VAR(m_prev_tok)) {

      if (m_prev_tok && m_prev_tok -> obj) {
 	if (IS_OBJECT (m_prev_tok -> obj -> instancevars) &&
	    str_eq (m_prev_tok->obj->instancevars->__o_name, 
		    (char *)"value")) {
	  m_prev_tok_obj = m_prev_tok -> obj -> instancevars;
	  if ((m = 
	       ismethd_find_method_by_name (&m_prev_tok_obj, 
					name, FALSE)) != NULL)
	    return TRUE;
	} else {
	  if (IS_VALUE_INSTANCE_VAR(m_prev_tok->obj)) {
	    m_prev_tok_obj = m_prev_tok -> obj;
	    if ((m = 
		 ismethd_find_method_by_name (&m_prev_tok_obj, 
					  name, FALSE)) != NULL)
	      return TRUE;
	  } else {
	    /*
	     *  Should only be needed for Collection
	     *  subclasses that don't have value instance 
	     *  variables.
	     */
 	    if ((!m_prev_tok -> obj -> instancevars) ||
 		(m_prev_tok -> obj -> instancevars &&
 		 !str_eq (m_prev_tok -> obj -> instancevars -> __o_name,
			  (char *)"value"))) {
	      m_prev_tok_obj = m_prev_tok -> obj;
	      if ((m = 
		   ismethd_find_method_by_name (&m_prev_tok_obj, 
					    name, FALSE)) != NULL)
		return TRUE;
	    }
	  }
	}
      } else { /* if (m_prev_tok && m_prev_tok -> obj)  */
	if (m_prev_tok && m_prev_tok -> value_obj) {
 	  if (IS_OBJECT (m_prev_tok -> value_obj -> instancevars) &&
	      str_eq (m_prev_tok->value_obj->instancevars->__o_name, 
		      (char *)"value")) {
	    m_prev_tok_obj = m_prev_tok -> value_obj -> instancevars;
	    if ((m = 
		 ismethd_find_method_by_name (&m_prev_tok_obj, 
					  name, FALSE)) != NULL)
	      return TRUE;
	  } else {
	    if (IS_VALUE_INSTANCE_VAR(m_prev_tok->value_obj)) {
	      m_prev_tok_obj = m_prev_tok -> value_obj;
	      if ((m = 
		   ismethd_find_method_by_name (&m_prev_tok_obj, 
					    name, FALSE)) != NULL)
		return TRUE;
	    } else {
	      /*
	       *  Should only be needed for Collection
	       *  subclasses that don't have value instance 
	       *  variables.
	       */
 	      if ((!m_prev_tok -> value_obj -> instancevars) ||
 		  (m_prev_tok -> value_obj -> instancevars && 
 		   !str_eq (m_prev_tok -> value_obj -> instancevars -> __o_name,
			    (char *)"value"))) {
		m_prev_tok_obj = m_prev_tok -> value_obj;
		if ((m = 
		     ismethd_find_method_by_name (&m_prev_tok_obj, 
					      name, FALSE)) != NULL)
		  return TRUE;
	      }
	    }
	  }
	} else { /* if (m_prev_tok && m_prev_tok -> value_obj)  */
	  if ((M_TOK(m_prev_tok) == METHODMSGLABEL) &&
	      (m_prev_tok -> value_obj &&
	       (m_prev_tok -> value_obj == 
		messages[msg_ptr] -> receiver_obj))) {
	    if ((m_prev_tok_obj = messages[msg_ptr] -> receiver_obj) != NULL){
	      if ((m = 
		   ismethd_find_method_by_name (&m_prev_tok_obj, 
					    name, FALSE)) != NULL)
		return TRUE;
	    }
	  }
	}
      }

    } else {
      /*
       *  The previous label is a method - check for a value_obj
       *  result also.
       */
      if (M_TOK(m_prev_tok) == METHODMSGLABEL) {
	if (m_prev_tok && IS_OBJECT (M_VALUE_OBJ(m_prev_tok))) {
	  m_prev_tok_obj = M_VALUE_OBJ(m_prev_tok);
	  if ((m = ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
	      != NULL) {
	    return TRUE;
	  }  else {
	    /*
	     *  Check for a method from an instance variable value class.
	     */
	    if (m_prev_tok_obj && HAS_INSTANCEVAR_CLASS(m_prev_tok_obj)) {
	      if (IS_OBJECT(m_prev_tok_obj -> instancevars)) {
		if ((m = ismethd_find_method_by_name 
		     (&(m_prev_tok_obj->instancevars), name, FALSE))
		    != NULL)
		  return TRUE;
	      }
	    }
	  }
	} else {
	  /*
	   *  If the previous method doesn't have a result yet, look
	   *  up the method's return class.
	   */
	  if (IS_OBJECT(m_prev_tok->receiver_obj)){
	    METHOD *m_prev_tok_method;
	    OBJECT *m_prev_tok_rcvr_obj;
	    m_prev_tok_rcvr_obj = m_prev_tok->receiver_obj;
	    if ((m_prev_tok_method = 
		 ismethd_find_method_by_name (&m_prev_tok_rcvr_obj,
					  M_NAME(m_prev_tok), FALSE))
		!= NULL) {
	      OBJECT *m_prev_return_class;
	      if (str_eq (m_prev_tok_method -> returnclass, "Any")) {
		m_prev_return_class = 
		  ((m_prev_tok->receiver_obj) ?
		   (m_prev_tok->receiver_obj->__o_class) :
		   (NULL));
	      } else {
		if (m_prev_tok_method -> returnclass) {
		  m_prev_return_class = 
		    __ctalkGetClass(m_prev_tok_method->returnclass);
		} else {
		  m_prev_return_class = m_prev_tok_rcvr_obj -> __o_class;
		}
	      }
	      if (ismethd_find_method_by_name (&m_prev_return_class, 
					   name, FALSE)) {
		return TRUE;
	      }
	    }
	  }
	}
      } else {
	if (m_prev_tok -> attrs & TOK_IS_AGGREGATE_MEMBER_LABEL) {
	  if ((m_prev_tok_obj = m_prev_tok -> value_obj) != NULL) {
	    if ((m = ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
		!= NULL) {
	      return TRUE;
	    }
	  }
	} else {
	  m_prev_tok_obj = m_prev_tok -> obj;
	  if (m_prev_tok && IS_OBJECT (m_prev_tok -> obj)) {
	    if ((m = ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
		!= NULL) {
	      return TRUE;
	    }  else {
	      /*
	       *  Check for a method from an instance variable value class.
	       */
	      if (m_prev_tok_obj && HAS_INSTANCEVAR_CLASS(m_prev_tok_obj)) {
		if (IS_OBJECT(m_prev_tok_obj -> instancevars)) {
		  if ((m = ismethd_find_method_by_name 
		       (&(m_prev_tok_obj->instancevars), name, FALSE))
		      != NULL)
		    return TRUE;
		}
	      }
	    }
	  } else {
	    m_prev_tok_obj = m_prev_tok -> value_obj;
	    if (m_prev_tok && IS_OBJECT (m_prev_tok -> value_obj)) {
	      if ((m = 
		   ismethd_find_method_by_name (&(m_prev_tok_obj), name, FALSE))
		  != NULL) {
		return TRUE;
	      }  else {
		/*
		 *  Check for a method from an instance variable value class.
		 */
		if (m_prev_tok_obj && HAS_INSTANCEVAR_CLASS(m_prev_tok_obj)) {
		  if (IS_OBJECT(m_prev_tok_obj -> instancevars)) {
		    if ((m = ismethd_find_method_by_name 
			 (&(m_prev_tok_obj->instancevars), name, FALSE))
			!= NULL)
		      return TRUE;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  return FALSE;
}

