/* $Id: rt_return.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "rtinfo.h"
#include "parser.h"

OBJECT *__ctalkSaveOBJECTMemberResource (OBJECT *o) {
  OBJECT *__o_copy;
  if (o && IS_OBJECT (o)) {
    if (o -> nrefs == 0) {
      __ctalkCopyObject(OBJREF(o), OBJREF(__o_copy));
      __o_copy = __ctalkCreateObjectInit 
	(o -> __o_name, o -> __o_classname, 
	 o -> __o_superclassname, o -> scope, o -> __o_value);
      __ctalkRegisterUserObject(__o_copy);
      return __o_copy;
    } else {
      return o;
    }
  } else {
    return NULL;
  }
}

OBJECT *__ctalkRegisterIntReturn (int i) {
  OBJECT *__o; 
  char __b[MAXLABEL];
  __ctalkDecimalIntegerToASCII(i,__b);
  __o = __ctalkCreateObjectInit 
    ("result", "Integer", "Magnitude", LOCAL_VAR, __b);
  __ctalkRegisterUserObject(__o);
  return __o;
}

OBJECT *__ctalkRegisterLongLongIntReturn (long long int i) {
  OBJECT *__o; 
  char __b[MAXLABEL];
  __ctalkDecimalLongLongToASCII(i,__b);
  __o = __ctalkCreateObjectInit 
    ("result", "LongInteger", "Magnitude", LOCAL_VAR, __b);
  __ctalkRegisterUserObject(__o);
  return __o;
}

OBJECT *__ctalkRegisterBoolReturn (int i) {
  OBJECT *__o; 
  char __b[MAXLABEL];
  __ctalkDecimalLongLongToASCII(i,__b);
  __o = __ctalkCreateObjectInit 
    ("result", "Boolean", "Integer", LOCAL_VAR, __b);
  __ctalkRegisterUserObject(__o);
  return __o;
}

OBJECT *__ctalkRegisterCharPtrReturn (char *s) {
  OBJECT *__o; 
  __o = __ctalkCreateObjectInit 
    ("result", "String", "Character", LOCAL_VAR, s);
  __ctalkRegisterUserObject(__o);
  return __o;
}

OBJECT *__ctalkRegisterCharReturn (char c) {
  OBJECT *__o; 
  char __b[2];
  __b[0] = c;
  __b[1] = '\0';
  __o = __ctalkCreateObjectInit 
    ("result", "Character", "Magnitude", LOCAL_VAR, __b);
  __ctalkRegisterUserObject(__o);
  return __o;
}

OBJECT *__ctalkRegisterFloatReturn (double d) {
  OBJECT *__o; 
  char __b[MAXLABEL];
#if defined(__sparc__) && defined(__svr4__)
  sprintf (__b, "%f", d);
#else
  sprintf (__b, "%lf", (double)d);
#endif
  __o = __ctalkCreateObjectInit 
    ("result", "Float", "Magnitude", LOCAL_VAR, __b);
  __ctalkRegisterUserObject(__o);
  return __o;
}

static OBJECT *CVAR_receiver_return (char *name) {
  char buf[MAXLABEL];
  OBJECT *obj;
  CVAR *c;
  if ((c = get_method_arg_cvars (name)) != NULL) {
    if (!strcmp (c -> type, "OBJECT") && (c -> n_derefs == 1)) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
      sprintf (buf, "%#lx", (unsigned long int)c -> val.__value.__ptr);
#else
      sprintf (buf, "%#x", (unsigned int)c -> val.__value.__ptr);
#endif
      if ((obj = obj_ref_str (buf)) != NULL) {
	if (is_receiver (obj))
	  return obj;
	if (is_receiver (top_parent_object (obj)))
	  return obj;
	return NULL;
      }
    }
  }
  return NULL;
}

static OBJECT *CVAR_ref_return (char *name) {
  char buf[MAXLABEL];
  OBJECT *obj;
  CVAR *c;
  if ((c = get_method_arg_cvars (name)) != NULL) {
    if (!strcmp (c -> type, "OBJECT") && (c -> n_derefs == 1)) {
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
      sprintf (buf, "%#lx", (unsigned long int)c -> val.__value.__ptr);
#else
      sprintf (buf, "%#x", (unsigned int)c -> val.__value.__ptr);
#endif
      if ((obj = obj_ref_str (buf)) != NULL) {
	if (obj -> scope & VAR_REF_OBJECT) 
	  return obj;
      }
    }
  }
  return NULL;
}

/*
 * The fn can assume that there is a local_cvar list because
 * it's only called from __ctalkSaveCVARResource (), which
 * is only called after a call to __ctalk_register_c_method_arg ().
 */
static void __delete_last_method_return_resource_cvar (void) {
  METHOD *m;
  CVAR *c;
  if ((m = __ctalkRtGetMethod ()) != NULL) {
    for (c = m -> local_cvars; c && c -> next; c = c -> next)
      ;
    if (c == m -> local_cvars) {
      m -> local_cvars = NULL;
    } else {
      c -> prev -> next = NULL;
    }
    _delete_cvar (c);
  }
}

/*
 *  NOTE - This function can ONLY be called after a call 
 *  to __ctalk_register_c_method_arg (), because it deletes
 *  the last local CVAR registered by the method, when it 
 *  calls __delete_last_method_return_resource_cvar, above.
 */
OBJECT *__ctalkSaveCVARResource (char *name) {
  OBJECT *obj, *obj_copy;
  CVAR *c;

  if ((obj = CVAR_receiver_return (name)) != NULL) {
    __delete_last_method_return_resource_cvar ();
    return obj;
  }

  if ((obj = CVAR_ref_return (name)) != NULL) {
    __delete_last_method_return_resource_cvar ();
    return obj;
  }

  if ((c = get_method_arg_cvars (name)) != NULL) {
    if (!strcmp (c -> type, "OBJECT") && (c -> n_derefs == 1)) {
      obj = (OBJECT *)c -> val.__value.__ptr;
      __ctalkRegisterUserObject (obj);
      /*
       *  This call requires that the __ctalkSaveCVARResource
       *  get called ONLY after a call to 
       *  __ctalk_register_c_method_arg ().
       */
      __delete_last_method_return_resource_cvar ();
      return obj;
    }

    /*
     *  !!!!!
     *  The call to cvar_object should not be needed
     *  for an OBJECT *.
     */
    if ((obj = cvar_object (c)) != NULL) {
      /*
       *  The only type right now should be OBJECT *'s.
       */
      if (obj -> nrefs) {
	if (!strcmp (c -> type, "OBJECT")) {
	  __ctalkCopyObject (OBJREF(obj), OBJREF(obj_copy));
	  __ctalkRegisterUserObject (obj_copy);
	  return obj_copy;
	} else {
	  _warning ("__ctalkSaveCVARResource: Unsupported type %s.\n",
		    c -> type);
	  return NULL;
	}
      } else { /* if (obj -> nrefs) */
	if (!strcmp (c -> type, "OBJECT")) {
	  __ctalkRegisterUserObject (obj);
	  __delete_last_method_return_resource_cvar ();
	  return obj;
	} else {
	  if ((!strcmp (c -> type, "int") && (c -> n_derefs == 0)) ||
	      (!strcmp (c -> type, "char") && (c -> n_derefs == 1)) ||
	      (!strcmp (c -> type, "char") && (c -> n_derefs == 0)) ||
	      (!strcmp (c -> type, "int") && !strcmp(c -> qualifier, "long") &&
	       !strcmp (c -> qualifier2, "long") && (c -> n_derefs == 0)) ||
	      (!strcmp (c -> type, "double") && (c -> n_derefs == 0))) {
	    __ctalkRegisterUserObject (obj);
	    __delete_last_method_return_resource_cvar ();
	    return obj;
	  } else {
	    _warning ("__ctalkSaveCVARResource: Unsupported type %s.\n",
		      c -> type);
	    __delete_last_method_return_resource_cvar ();
	    return NULL;
	  }
	}
      }
    }
  }
  __delete_last_method_return_resource_cvar ();
  return NULL;
}

/*
 * See the comment above for __ctalkSaveCVARResource.
 */
OBJECT *__ctalkSaveCVARArrayResource_internal (char *name, int initializer_size,
				      void *var) {
  OBJECT *array_obj, *mbr_obj, *t, *member_class, *return_class,
    *elem_obj;
  CVAR *c;
  int i;
  int *i_array = NULL;
  long long int *ll_array = NULL;
  char *c_array = NULL;
  double *d_array = NULL;
  METHOD *m;
  char mbr_name[MAXLABEL], valbuf[MAXMSG];;

  array_obj = NULL;
  if ((c = get_method_arg_cvars (name)) != NULL) {
    if (!strcmp (c -> type, "int")) {
      if (*(c -> qualifier) && *(c -> qualifier2)) {
	if (!strcmp (c -> qualifier, "long") && 
	    !strcmp (c -> qualifier2, "long")) {
	  ll_array = var;
	  member_class = __ctalkGetClass (LONGINTEGER_CLASSNAME);
	} else {
	  i_array = var;
	  member_class = __ctalkGetClass (INTEGER_CLASSNAME);
	}
      } else {
	i_array = var;
	member_class = __ctalkGetClass (INTEGER_CLASSNAME);
      }
    } else {
      if (!strcmp (c -> type, "char")) {
	c_array = var;
	m = __ctalkRtGetMethod ();
	if (*(m -> returnclass)) {
	  if (!strcmp (m -> returnclass, ARRAY_CLASSNAME)) {
	    member_class = __ctalkGetClass (CHARACTER_CLASSNAME);
	  } else {
	    member_class = __ctalkGetClass (STRING_CLASSNAME);
	    array_obj = __ctalkCreateObjectInit
	      (name, STRING_CLASSNAME,
	       STRING_SUPERCLASSNAME,
	       LOCAL_VAR, c_array);
	    __ctalkRegisterUserObject (array_obj);
	    return array_obj;
	  }
	} else {
	  return_class = __ctalk_rtReceiverClassObject ();
	  if (!strcmp (return_class -> __o_name, ARRAY_CLASSNAME)) {
	    member_class = __ctalkGetClass (CHARACTER_CLASSNAME);
	  } else {
	    member_class = __ctalkGetClass (STRING_CLASSNAME);
	  }
	}
      } else {
	if (!strcmp (c -> type, "float") || !strcmp (c -> type, "double")) {
	  d_array = var;
	  member_class = __ctalkGetClass (FLOAT_CLASSNAME);
	} else {
	  _warning ("__ctalkSaveCVARResource: Array, \"%s,\" type, \"%s,\" not supported directly by class library.\n", name, c -> type);
	  return NULL;
	}
      }
    } /* if (!strcmp (c -> type, "int")) */
    array_obj = __ctalkCreateObjectInit
      (name, ARRAY_CLASSNAME,
       ARRAY_SUPERCLASSNAME,
       LOCAL_VAR, NULLSTR);
    for (i = 0, t = array_obj -> instancevars; i < initializer_size; i++) {
      sprintf (mbr_name, "%d", i);
      if (!strcmp (member_class -> __o_name, INTEGER_CLASSNAME)) {
	if (i_array != NULL)  {
	  sprintf (valbuf, "%d", i_array[i]);
	}
      } else {
	if (!strcmp (member_class -> __o_name, LONGINTEGER_CLASSNAME)) {
	  if (ll_array != NULL) {
	    sprintf (valbuf, "%lld", ll_array[i]);
	  }
	} else {
	  if (!strcmp (member_class -> __o_name, CHARACTER_CLASSNAME)) {
	    if (c_array != NULL) {
	      sprintf (valbuf, "\'%c\'", c_array[i]);
	    }
	  } else {
	    if (!strcmp (member_class -> __o_name, STRING_CLASSNAME)) {
	      if (c_array != NULL) {
		strcpy (valbuf, c_array);
	      }
	    } else {
	      if (!strcmp (member_class -> __o_name, FLOAT_CLASSNAME)) {
		if (d_array != NULL) {
#if defined(__sparc__) && defined(__svr4__)
		sprintf (valbuf, "%f", d_array[i]);
#else
		sprintf (valbuf, "%lf", d_array[i]);
#endif
		}
	      }
	    }
	  }
	}
      }
      elem_obj = __ctalkCreateObjectInit (mbr_name,
					 member_class -> __o_name,
					 member_class -> __o_superclassname,
					 LOCAL_VAR,
					 valbuf);
#if defined(__GNUC__) && defined(__sparc__) && defined(__svr4__)
      sprintf (valbuf, "0x%p", elem_obj);
#else
# if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
      sprintf (valbuf, "%#lx", (unsigned long int)elem_obj);
# else
      sprintf (valbuf, "%#x", (unsigned int)elem_obj);
# endif
#endif
      mbr_obj = __ctalkCreateObjectInit (mbr_name,
					 "Key", "Symbol",
					 LOCAL_VAR,
					 valbuf);
      t -> next = mbr_obj;
      mbr_obj -> prev = t;
      t = mbr_obj;
    }
    return array_obj;
  }
  return NULL;
}

OBJECT *__ctalkSaveCVARArrayResource (char *name, int initializer_size,
				      void *var) {
  OBJECT *array_obj;
  if ((array_obj = __ctalkSaveCVARArrayResource_internal (name, initializer_size,
							  var)) != NULL) {
    __ctalkRegisterUserObject (array_obj);
  }
  return array_obj;
}
