/* $Id: rtinfo.c,v 1.2 2014/05/04 02:20:18 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "rtinfo.h"
#include "parser.h"

#ifdef CLASSLIBDIR 
char *classlibdir = CLASSLIBDIR;
#else
char *classlibdir = NULL;
#endif
#ifdef PKGNAME
char *pkgname = PKGNAME;
#else
char *pkgname = NULL;
#endif

char *library_include_paths[MAXUSERDIRS] = { NULL, };

RT_INFO rtinfo;

RT_INFO *__call_stack[MAXARGS+1];
int __call_stack_ptr = MAXARGS;

I_PASS interpreter_pass;

RT_CLEANUP_MODE global_cleanup_mode = rt_cleanup_null;

static char argv_name[FILENAME_MAX];

static char classlib_name[FILENAME_MAX];

PARSER *parsers[MAXARGS+1];
int current_parser_ptr;

void __argvName (char *s) {
  strcpy (argv_name, s);
}

char *__argvFileName (void) {
  return argv_name;
}

void __classLibName (char *s) {
  strcpy (classlib_name, s);
}

char *__classLibFileName (void) {
  return classlib_name;
}

void __setClassLibRead (Boolean b) {
  rtinfo.classlib_read = b;
}

Boolean __getClassLibRead (void) {
  return rtinfo.classlib_read;
}

int __ctalkRtSaveSourceFileName (char *s) {
  strcpy (rtinfo.source_file, s);
  return 0;
}

char *__ctalkClassSearchPath (void) {
  return __ctalkClassLibraryPath ();
}

char *__ctalkClassLibraryPath (void) {
  static char classlibpath[MAXMSG];
  char **dir, direntry[FILENAME_MAX];
  if (*library_include_paths == NULL) return NULL;
  dir = library_include_paths;
  strcpy (classlibpath, *dir);
  ++dir;
  while (*dir) {
    sprintf (direntry, ":%s", *dir);
    if ((strlen (classlibpath) + strlen (direntry)) > MAXMSG)
      break;
    strcat (classlibpath, direntry);
    ++dir;
  }
  return classlibpath;
}

char *__source_filename (void) {
  return rtinfo.source_file;
}

void __ctalkRtReceiver (OBJECT *o) {
  rtinfo.rcvr_obj = o;
}

OBJECT *__ctalkRtReceiverObject (void) {
  return rtinfo.rcvr_obj;
}

void __ctalk_rtReceiverClass (OBJECT *o) {
  rtinfo.rcvr_class_obj = o;
}

OBJECT *__ctalk_rtReceiverClassObject (void) {
  return rtinfo.rcvr_class_obj;
}

void __ctalkRtMethodClass (OBJECT *o) {
  rtinfo.method_class_obj = o;
}

OBJECT *__ctalkRtMethodClassObject (void) {
  return rtinfo.method_class_obj;
}

void __ctalk_rtMethodFn (OBJECT *(*fn)()) {
  rtinfo.method_fn = fn;
}

void *__ctalkRtGetMethodFn (void) {
  return rtinfo.method_fn;
}

/*
 *  Only saved by __save_rt_info (). 
 */
METHOD *__ctalkRtGetMethod (void) {
  if (__call_stack_ptr < MAXARGS) {
    return (IS_METHOD(__call_stack[__call_stack_ptr+1]->method) ?
	    __call_stack[__call_stack_ptr+1] -> method : 
	    NULL);
  } else {
    return NULL;
  }
}

RT_FN *get_fn (void) { 
  if (__call_stack_ptr >= MAXARGS) return NULL;
  return (__call_stack[__call_stack_ptr+1]->_rt_fn ? 
	  __call_stack[__call_stack_ptr+1]->_rt_fn :
	  NULL); 
}

RT_FN *get_calling_fn (void) { 
  if ((__call_stack_ptr + 1) >= MAXARGS) return NULL;
  return (__call_stack[__call_stack_ptr+2]->_rt_fn ? 
	  __call_stack[__call_stack_ptr+2]->_rt_fn :
	  NULL); 
}

METHOD *get_calling_method (void) { 
  if ((__call_stack_ptr + 1) >= MAXARGS) return NULL;
  return (__call_stack[__call_stack_ptr+2]->method ? 
	  __call_stack[__call_stack_ptr+2]->method :
	  NULL); 
}

int __ctalk_initFn (char *name) {

  RT_FN *r;
  RT_INFO *r_info;

  r_info = _new_rtinfo ();

  r = new_rt_fn ();

  strcpy (r -> name, name);

  r_info -> _rt_fn = r;
  r_info -> _arg_frame_top = __ctalkGetArgPtr () + 1;
  strcpy (r_info -> source_file, rtinfo.source_file);
  __call_stack[__call_stack_ptr--] = r_info;

  return 0;
}

static void __delete_fn_local_objs (RT_FN *r) {
  VARENTRY *v_ptr, *v_ptr_prev;

  for (v_ptr = r -> local_objects.vars; v_ptr && v_ptr -> next;
       v_ptr = v_ptr -> next)
    ;
  if (v_ptr == r -> local_objects.vars) {
    if (v_ptr && IS_OBJECT(v_ptr -> var_object)) {
      __objRefCntZero (OBJREF(v_ptr -> var_object));
      __ctalkDeleteObject (v_ptr -> var_object);
      delete_varentry (v_ptr);
    }
    r -> local_objects.vars = NULL;
  } else {
    while (v_ptr != r -> local_objects.vars) {
      v_ptr_prev = v_ptr -> prev;
      if (v_ptr && IS_OBJECT(v_ptr -> var_object)) {
	__objRefCntZero (OBJREF(v_ptr -> var_object));
	__ctalkDeleteObject (v_ptr -> var_object);
      }
      delete_varentry (v_ptr);
      v_ptr = v_ptr_prev;
    }
    if (v_ptr && IS_OBJECT(v_ptr -> var_object)) {
      __objRefCntZero (OBJREF(v_ptr -> var_object));
      __ctalkDeleteObject (v_ptr -> var_object);
    }
    delete_varentry (v_ptr);
    r -> local_objects.vars = NULL;
  }
}

extern int __cleanup_deletion;

int __ctalk_exitFn (int app_exit) {

  RT_FN *r;
  RT_INFO *r_info;
  int cleanup_deletion_save;

  if (__call_stack_ptr < MAXARGS) {
    r_info = __call_stack[++__call_stack_ptr];
    __call_stack[__call_stack_ptr] = NULL;

    if (!r_info) return SUCCESS;
    r = r_info -> _rt_fn;
    free (r_info);

    if (r) {
      if (r -> local_objects.vars != NULL) {
	cleanup_deletion_save = __cleanup_deletion;
	__cleanup_deletion = app_exit;
	__delete_fn_local_objs (r);
	__cleanup_deletion = cleanup_deletion_save;
      }
      free (r);
    }
  }

  if (app_exit) {
    __ctalkClassDictionaryCleanup ();
    delete_return_valptrs ();
  }

  return 0;
}

int __ctalkErrorExit (void) {

  RT_FN *r;
  RT_INFO *r_info;

  for ( ++__call_stack_ptr; 
	__call_stack_ptr <= MAXARGS; ++__call_stack_ptr) {
    if ((r_info = __call_stack[++__call_stack_ptr]) == NULL)
      continue;
    __call_stack[__call_stack_ptr] = NULL;
    r = r_info -> _rt_fn;
    free (r_info);
    if (r) {
      if (r -> local_objects.vars != NULL) {
	__delete_fn_local_objs (r);
      }
      free (r);
    }
  }
  __ctalkClassDictionaryCleanup ();
  delete_return_valptrs ();
  return 1;
}

RT_FN *new_rt_fn (void) {

  RT_FN *r;

  if ((r = (RT_FN *)__xalloc (sizeof (RT_FN))) == NULL)
    _error ("new_rt_fn: allocation error.\n");

  return r;
}

void __init_data (void) {
  /*
   *  Not used with DJGPP for the moment because I haven't 
   *  been able to test it.
   *
   *  Sparcs need more work on the swap space allocation
   *  because this affects fork () calls adversely.
   */
#if !defined(DJGPP) && !defined(__sparc__)
  struct rlimit rlim;
  unsigned long long int sbrk_arg = 0x40000000;
  void *b;
  rlim.rlim_cur = RLIM_INFINITY;
  rlim.rlim_max = RLIM_INFINITY;
  setrlimit (RLIMIT_DATA, &rlim);
  while ((b = sbrk (sbrk_arg)) == (void *)-1) {
#ifdef SHOW_SBRK_INFO
    fprintf (stderr, "__init_data: sbrk_arg = %lld.\n",
	     (unsigned long long int)sbrk_arg);
#endif
     sbrk_arg /= 2;
  }
#endif
}

/*
 *  This doesn't seem to be very useful - delete? 
 */
void __init_time (void) {
  time_t t;
  struct tm *tm_ptr;
  t = time (NULL);
  tm_ptr = localtime (&t);
}

int is_recursive_method_call (METHOD *);

int __save_rt_info (OBJECT *__rcvr_obj, OBJECT *__rcvr_class_obj, 
		    OBJECT *__method_class_obj, METHOD *method,
		    OBJECT *(*fn) (), Boolean inline_call) {
  RT_INFO *r;
  r = _new_rtinfo ();

  r -> inline_call = inline_call;
  r -> rcvr_obj = rtinfo.rcvr_obj = __rcvr_obj;
  r -> rcvr_class_obj = rtinfo.rcvr_class_obj = __rcvr_class_obj;
  r -> method_class_obj = rtinfo.method_class_obj = __method_class_obj;
  r -> method = method;
  r -> method_fn = rtinfo.method_fn = fn;
  r -> _arg_frame_top = __ctalkGetArgPtr () + 1;
  strcpy (r -> source_file, rtinfo.source_file);
  if (is_recursive_method_call (r -> method)) {
    r -> local_object_cache[r -> local_obj_cache_ptr++] = 
      M_LOCAL_VAR_LIST(r -> method);
    M_LOCAL_VAR_LIST(r -> method) = NULL;
  }
  __call_stack[__call_stack_ptr--] = r;

  return SUCCESS;
}

int __restore_rt_info (void) {

  RT_INFO *r;
  VARENTRY *__v, *__v_prev;
  METHOD *__m;

  r = __call_stack[++__call_stack_ptr];
  __call_stack[__call_stack_ptr] = NULL;

  if (__call_stack_ptr >= MAXARGS) {
    rtinfo.rcvr_obj = rtinfo.rcvr_class_obj = rtinfo.method_class_obj = NULL;
    rtinfo.method_fn = NULL;
  } else {
    rtinfo.rcvr_obj = __call_stack[__call_stack_ptr+1] -> rcvr_obj;
    rtinfo.rcvr_class_obj = __call_stack[__call_stack_ptr+1] -> rcvr_class_obj;
    rtinfo.method_class_obj = 
      __call_stack[__call_stack_ptr+1] -> method_class_obj;
    rtinfo.method_fn = __call_stack[__call_stack_ptr+1] -> method_fn;
    rtinfo.method = __call_stack[__call_stack_ptr+1] -> method;
  }
  if (is_recursive_method_call (r -> method)) {
    if (r -> method && M_LOCAL_VAR_LIST(r -> method)) {
      __m = r -> method;
      for (__v = M_LOCAL_VAR_LIST(__m);
	   __v && IS_OBJECT(__v -> var_object) && 
	     __v -> next; 
	   __v = __v -> next)
	; 
      if (__v == M_LOCAL_VAR_LIST(__m)) {
	if (M_LOCAL_VAR_LIST(__m) && M_LOCAL_VAR_LIST(__m) -> var_object) {
	  __objRefCntZero (OBJREF(M_LOCAL_VAR_LIST(__m)->var_object));
	  __ctalkDeleteObject (M_LOCAL_VAR_LIST(__m) -> var_object);
	}
	delete_varentry (M_LOCAL_VAR_LIST(__m));
	M_LOCAL_VAR_LIST(__m) = NULL;
      } else {
	while (__v != M_LOCAL_VAR_LIST(__m)) {
	  __v_prev = __v -> prev;
	  if (__v && __v -> var_object) {
	    __objRefCntZero (OBJREF (__v -> var_object));
	    __ctalkDeleteObject ( __v -> var_object);
	  }
	  delete_varentry (__v);
	  __v = __v_prev;
	  if (!__v) break;
	  __v -> next = NULL;
	}
	if (M_LOCAL_VAR_LIST(__m) && M_LOCAL_VAR_LIST(__m) -> var_object) { 
	  __objRefCntZero (OBJREF (M_LOCAL_VAR_LIST(__m)->var_object));
	  __ctalkDeleteObject (M_LOCAL_VAR_LIST(__m) -> var_object); 
	  delete_varentry (M_LOCAL_VAR_LIST(__m));
	  M_LOCAL_VAR_LIST(__m) = NULL;
	}
      }
    }
    M_LOCAL_VAR_LIST(r -> method) =
      r -> local_object_cache[--r -> local_obj_cache_ptr];
  }
  free (r);

  return SUCCESS;
}

int __get_call_stack_ptr (void) {
  return __call_stack_ptr;
}

int __current_call_stack_idx (void) {
  return __call_stack_ptr + 1;
}

Boolean __ctalkIsInlineCall (void) {
  return rtinfo.inline_call;
}

Boolean __ctalkCallerIsInlineCall (void) {
  return __call_stack[__call_stack_ptr + 1]->inline_call;
}

Boolean __ctalkIsArgBlkScope (void) {
  return __call_stack[__call_stack_ptr + 1]->block_scope;
}

RT_INFO *_new_rtinfo (void) {
  static RT_INFO *r;
  r = (RT_INFO *)__xalloc (sizeof (struct _rtinfo));
  return r;
}

void _delete_rtinfo (RT_INFO *r) {
  __xfree (r);
}

/*
 *  Returns the call stack index of the previous method
 *  call, or zero.
 */
int is_recursive_method_call (METHOD *m) {
  int i;
  for (i = __call_stack_ptr + 1; i <= MAXARGS; i++) {
    if (__call_stack[i] && (m == __call_stack[i]->method))
      return i;
  }
  return FALSE;
}

/*
 *  The *successive_call* functions are called only from
 *  __ctalkInitLocalObjects at the moment.  Local objects are restored
 *  in __restore_rt_info (), above.
 *
 *  See the notes for last_eval_result () in rt_expr.c. 
 */
int is_successive_method_call (METHOD *m) {
  VARENTRY *v;

  for (v = M_LOCAL_VAR_LIST(m); v; v = v -> next) {
    if (v -> var_object && 
	(is_receiver (v -> var_object) || is_arg (v -> var_object)))
      return TRUE;

    if (v -> var_object == last_eval_result ())
      return TRUE;

  }
  return FALSE;
}

void register_successive_method_call (void) {
  __call_stack[__call_stack_ptr + 1] -> _successive_call = TRUE;
  __call_stack[__call_stack_ptr + 1] -> method -> local_object_ptr += 1;
}

void __rt_init_library_paths (void) {
  char libdir[FILENAME_MAX],
    libenvdir[FILENAME_MAX],
    libenvdir2[FILENAME_MAX],
    libenvdir3[FILENAME_MAX],
    *libenv, *p, *q;
  int n_paths = 0;

  if ((libenv = getenv ("CLASSLIBDIRS")) != NULL) {
    p = libenv;
    do {
      q = strchr (p, ':');
      if (q) {
	substrcpy (libenvdir, p, 0, q - p);
	p = q + 1;
      } else {
	strcpy (libenvdir, p);
      }
      strcpy (libenvdir2, expand_path(libenvdir));
      library_include_paths[n_paths++] = strdup (libenvdir2);
      sprintf (libenvdir3, "%s/%s", libenvdir2, pkgname);
      if (is_dir (libenvdir3)) {
	library_include_paths[n_paths++] = strdup (libenvdir3);
      }
    } while (q != NULL);
  }

  library_include_paths[n_paths++] = strdup (classlibdir);

  sprintf (libdir, "%s/%s", classlibdir, pkgname);
  if (is_dir (libdir)) {
    library_include_paths[n_paths++] = strdup (libdir);
  }

}

/*
 *  Should only be called by error ().  
 */
void cleanup_rt_fns (void) {
  int i;
  RT_FN *r;
  for (i = __call_stack_ptr + 1; i <= MAXARGS; i++) {
    if ((r = __call_stack[i] -> _rt_fn) != NULL) {
      if (r -> local_objects.vars) {
	DELETE_VAR_LIST (r -> local_objects.vars);
      }
    }
  }
}
