/* $Id: rtobjref.c,v 1.3 2014/05/09 03:28:42 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  Functions to handle object reference counts.  The arguments to 
 *  the function calls should use the OBJREF() macro, which is defined
 *  in object.h.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

/*
 *  OBJREF_T is defined in object.h.
 */

void __refObj (OBJREF_T ref1, OBJREF_T ref2) {
  OBJECT *var;
  if (IS_OBJECT (*ref1)) {
    --(*ref1) -> nrefs;
    for (var = (*ref1) -> instancevars; var; var = var -> next)
      __objRefCntDec (OBJREF (var));
  }
  *ref1 = *ref2;
  ++(*ref2) -> nrefs;
  for (var = (*ref2) -> instancevars; var; var = var -> next)
    __objRefCntInc (OBJREF (var));
}

/***/
static void __obj_ref_cnt_store (OBJREF_T obj_p, int nrefs) {
  OBJECT *vars;

  if (IS_OBJECT (*obj_p)) {
    (*obj_p) -> nrefs = nrefs;
    for (vars = (*obj_p) -> instancevars; vars; vars = vars -> next)
      __obj_ref_cnt_store (OBJREF (vars), nrefs);
  } else {
#ifdef DEBUG_INVALID_OBJ_REFS
    _warning ("__objRefCntSet: Not an object.\n");
#endif
  }
}

void __objRefCntInc (OBJREF_T obj_p) {

  __obj_ref_cnt_store (obj_p, (*obj_p) -> nrefs + 1);
/*   OBJECT *vars; */
/* #ifdef DEBUG_INVALID_OBJ_REFS */
/*   if (!IS_OBJECT (*obj_p)) { */
/*     _warning ("__objRefCntInc: Not an object.\n"); */
/*   } else { */
/* #endif */
/*     /\* This helps insure that a function call here doesn't */
/*        get optimized out. *\/ */
/*     asm (""); */
/*     ((*obj_p) -> nrefs)++; */
/*     for (vars = (*obj_p) -> instancevars;  */
/* 	 IS_OBJECT(vars);  */
/* 	 vars = vars -> next) { */
/*       __objRefCntInc (OBJREF (vars)); */
/*     } */
/* #ifdef DEBUG_INVALID_OBJ_REFS */
/*   } */
/* #endif */
}

void __objRefCntInc_old (OBJREF_T obj_p) {
  OBJECT *vars;
#ifdef DEBUG_INVALID_OBJ_REFS
  if (!IS_OBJECT (*obj_p)) {
    _warning ("__objRefCntInc: Not an object.\n");
  } else {
#endif
    /* This helps insure that a function call here doesn't
       get optimized out. */
    asm ("");
    ((*obj_p) -> nrefs)++;
    for (vars = (*obj_p) -> instancevars; 
	 IS_OBJECT(vars); 
	 vars = vars -> next) {
      __objRefCntInc (OBJREF (vars));
    }
#ifdef DEBUG_INVALID_OBJ_REFS
  }
#endif
}

void __objRefCntDec (OBJREF_T obj_p) {

  if ((*obj_p) -> nrefs == 0)
    return;

  __obj_ref_cnt_store (obj_p, (*obj_p) -> nrefs - 1);

/*   OBJECT *vars; */
/* #ifdef DEBUG_INVALID_OBJ_REFS */
/*   if (!IS_OBJECT (*obj_p)) { */
/*     _warning ("__objRefCntDec: Not an object.\n"); */
/*   } else { */
/* #endif */
/*     /\* This helps insure that a function call here doesn't */
/*        get optimized out. *\/ */
/*      asm (""); */
/*     ((*obj_p) -> nrefs)--; */
/*     for (vars = (*obj_p) -> instancevars; */
/*      	 IS_OBJECT(vars); */
/*      	 vars = vars -> next) { */
/*        __objRefCntDec (OBJREF (vars)); */
/*     } */
/* #ifdef DEBUG_INVALID_OBJ_REFS */
/*   } */
/* #endif */
}


void __objRefCntDec_old (OBJREF_T obj_p) {
  OBJECT *vars;
#ifdef DEBUG_INVALID_OBJ_REFS
  if (!IS_OBJECT (*obj_p)) {
    _warning ("__objRefCntDec: Not an object.\n");
  } else {
#endif
    /* This helps insure that a function call here doesn't
       get optimized out. */
     asm ("");
    ((*obj_p) -> nrefs)--;
    for (vars = (*obj_p) -> instancevars;
     	 IS_OBJECT(vars);
     	 vars = vars -> next) {
       __objRefCntDec (OBJREF (vars));
    }
#ifdef DEBUG_INVALID_OBJ_REFS
  }
#endif
}


void __objRefCntSet (OBJREF_T obj_p, int nrefs) {
  OBJECT *vars;

  /* 
     This line helps prevent any call from zapping an object.  
     If we want an object to be expressly deleted, call 
     __objRefCntZero () before __ctalkDeleteObject* () - which 
     is what the Object::delete method does.
  */
  if (nrefs == 0) return;

  if (!IS_OBJECT (*obj_p)) {
#ifdef DEBUG_INVALID_OBJ_REFS
    _warning ("__objRefCntSet: Not an object.\n");
#endif
  } else {
    (*obj_p) -> nrefs = nrefs;
    for (vars = (*obj_p) -> instancevars; vars; vars = vars -> next)
      __objRefCntSet (OBJREF (vars), nrefs);
  }
}

void __objRefCntZero (OBJREF_T obj_p) {
  OBJECT *vars;

  if (!IS_OBJECT (*obj_p)) {
#ifdef DEBUG_INVALID_OBJ_REFS
    _warning ("__objRefCntSet: Not an object.\n");
#endif
  } else {
    (*obj_p) -> nrefs = 0;
    for (vars = (*obj_p) -> instancevars; vars; vars = vars -> next)
      if (IS_OBJECT(vars))
	__objRefCntZero (OBJREF (vars));
  }
}

#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
#define OBJREF_IS_OBJECT(x) (((long int)(x) > 0) && IS_OBJECT(x))
#else
#define OBJREF_IS_OBJECT(x) (((int)(x) > 0) && IS_OBJECT(x))
#endif

OBJECT *obj_ref_str (char *__s) {
  OBJECT *__r;
  if (__s == NULL)
    return NULL;
#if ((defined (__sparc__) && defined (__GNUC__)) || \
     (defined (__APPLE__) && defined (__GNUC__)))
  if ((__s[0] && (__s[0] == '0')) && 
      (__s[1] && (__s[1] == 'x')) && 
      (__s[2] && isxdigit((int)__s[2])) && 
      (__s[3] && isxdigit((int)__s[3])) && 
      (__s[4] && isxdigit((int)__s[4])) && 
      (__s[5] && isxdigit((int)__s[5])) && 
      (__s[6] && isxdigit((int)__s[6])) && 
      (__s[7] && isxdigit((int)__s[7]))) {
    if ((__r = (OBJECT *)__ctalkStrToPtr (__s)) != NULL) {
      if (OBJREF_IS_OBJECT(__r)) {
	return __r;
      }
    }
  }
  /*
   *  In case Solaris or OS X uses different addressing.
   */
  if ((__s[0] && (__s[0] == '0')) && 
      (__s[1] && (__s[1] == 'x')) && 
      (__s[2] && isxdigit((int)__s[2])) && 
      (__s[3] && isxdigit((int)__s[3])) && 
      (__s[4] && isxdigit((int)__s[4])) && 
      (__s[5] && isxdigit((int)__s[5])) && 
      (__s[6] && isxdigit((int)__s[6]))) {
    if ((__r = (OBJECT *)__ctalkStrToPtr (__s)) != NULL) {
      /***/
/*       if (((int)__r > 0) && IS_OBJECT(__r)) { */
      if (OBJREF_IS_OBJECT(__r)) {
	return __r;
      }
    }
  }
#else
  if ((__s[0] && (__s[0] == '0')) && 
      (__s[1] && (__s[1] == 'x')) && 
      (__s[2] && isxdigit((int)__s[2])) && 
      (__s[3] && isxdigit((int)__s[3])) && 
      (__s[4] && isxdigit((int)__s[4])) && 
      (__s[5] && isxdigit((int)__s[5])) && 
      (__s[6] && isxdigit((int)__s[6])) && 
      (__s[7] && isxdigit((int)__s[7])) && 
      (__s[8] && isxdigit((int)__s[8]))) {
    if ((__r = (OBJECT *)__ctalkStrToPtr (__s)) != NULL) {
/*       if (IS_OBJECT(__r)) { */
      if (OBJREF_IS_OBJECT(__r)) {
	return __r;
      }
    }
  }
#endif
  return NULL;
}

void *generic_ptr_str (char *__s) {
  void *__r;
  int i;
  if ((__s[0] && (__s[0] == '0')) && 
      (__s[1] && (__s[1] == 'x'))) {
    for (i = 2; __s[i]; ++i) {
      if (!isxdigit ((int)__s[i])) {
	return NULL;
      }
    }
  } else {
    return NULL;
  }
  if ((__r = __ctalkStrToPtr (__s)) != NULL) {
    return __r;
  }
  return NULL;
}

void *__ctalkGenericPtrFromStr (char *s) {
  return generic_ptr_str (s);
}
