/* $Id: rttmpobj.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

int delete_all_objs_expr_frame (MESSAGE_STACK messages,
				int idx,
				int start_idx,
				int end_idx,
				int *all_objects_deleted) {
  OBJECT *o;
  int i;
  int have_created_param;

  o = messages[idx] -> obj;

  have_created_param = 0;

  if (!IS_OBJECT(o))
    return SUCCESS;

  if (o -> nrefs <= 0) {
    for (i = start_idx; i >= end_idx; i--) {

      if (!messages[i] || !IS_MESSAGE(messages[i]))
	break;

      if ((messages[i] -> obj == o) ||
	  is_p_obj (o, messages[i] -> obj)) {
	if ((messages[i] -> obj -> nrefs == 0) || 
	    (messages[i] -> attrs & TOK_OBJ_IS_CREATED_PARAM)) {
	  have_created_param = TRUE;
	  messages[i] -> attrs &= ~TOK_OBJ_IS_CREATED_PARAM;
	}
	messages[i] -> obj = NULL;
	if (messages[i] -> attrs & OBJ_IS_INSTANCE_VAR)
	  messages[i] -> attrs &= ~OBJ_IS_INSTANCE_VAR;
	if (messages[i] -> attrs & OBJ_IS_CLASS_VAR)
	  messages[i] -> attrs &= ~OBJ_IS_CLASS_VAR;
      }

      if ((messages[i] -> value_obj == o) ||
	  is_p_obj (o, messages[i] -> value_obj)) {
	messages[i] -> value_obj = NULL;
	if (messages[i] -> attrs & VALUE_OBJ_IS_INSTANCE_VAR)
	  messages[i] -> attrs &= ~VALUE_OBJ_IS_INSTANCE_VAR;
	if (messages[i] -> attrs & VALUE_OBJ_IS_CLASS_VAR)
	  messages[i] -> attrs &= ~VALUE_OBJ_IS_CLASS_VAR;
      }

      if ((messages[i] -> receiver_obj == o) ||
	  is_p_obj (o, messages[i] -> receiver_obj)) {
	messages[i] -> receiver_obj = NULL;
      }

    }

    if (!IS_VALUE_INSTANCE_VAR (o) || (have_created_param > 0)) { 
      __objRefCntZero (OBJREF(o));
      __ctalkDeleteObjectInternal (o);
      *all_objects_deleted = TRUE;
    }

    return -1;

  } else {
    if (have_created_param) {
      __objRefCntZero (OBJREF(o));
      __ctalkDeleteObject (o);
      *all_objects_deleted = TRUE;
    } else {
      __ctalkDeleteObjectInternal (o);
      *all_objects_deleted = FALSE;
    }
    return o -> nrefs;
  }
}

int delete_all_value_objs_expr_frame (MESSAGE_STACK messages,
				int idx,
				int start_idx,
				int end_idx,
				int *all_objects_deleted) {
  OBJECT *o;
  int i;
  o = messages[idx] -> value_obj;
  if (!IS_OBJECT(o))
    return SUCCESS;
  if (o -> nrefs <= 0) {
    for (i = start_idx; i >= end_idx; i--) {
      if (!messages[i] || !IS_MESSAGE(messages[i]))
	break;
      if ((messages[i] -> value_obj == o) ||
	  is_p_obj (o, messages[i] -> value_obj)) {
	messages[i] -> value_obj = NULL;
	if (messages[i] -> attrs & VALUE_OBJ_IS_INSTANCE_VAR)
	  messages[i] -> attrs &= ~VALUE_OBJ_IS_INSTANCE_VAR;
	if (messages[i] -> attrs & VALUE_OBJ_IS_CLASS_VAR)
	  messages[i] -> attrs &= ~VALUE_OBJ_IS_CLASS_VAR;
      }
    }
    __ctalkDeleteObjectInternal (o);
    *all_objects_deleted = TRUE;
    return -1;
  } else {
    __ctalkDeleteObjectInternal (o);
    *all_objects_deleted = FALSE;
    return o -> nrefs;
  }
}

int _cleanup_temporary_objects (OBJECT *obj, OBJECT *result_obj, OBJECT *subexpr_obj,
				RT_CLEANUP_MODE cleanup_mode) {

  int nrefs;
  if (!IS_OBJECT(obj)) {
    return ERROR;
  }
  if (object_is_deletable(obj, result_obj, subexpr_obj)) {
    if (obj -> scope & METHOD_USER_OBJECT) {
      if (obj -> nrefs > 0)
	__ctalkDeleteObjectInternal (obj);
      return obj -> nrefs;
    } else {
      __ctalkDeleteObject (obj);
      nrefs = 0;
      return nrefs;
    }
  }
  if (IS_OBJECT(obj) && 
      obj->__o_p_obj && is_receiver(obj->__o_p_obj))
    return ERROR;

  if (IS_OBJECT(obj) &&
      object_is_decrementable(obj, result_obj, subexpr_obj)) {
    nrefs = obj -> nrefs;
    --nrefs;
    if (obj -> nrefs < 0)
      __objRefCntSet (OBJREF(obj), 0);
    __ctalkDeleteObjectInternal (obj);
    if (nrefs < 0)
      return 0;
    if ((obj -> scope & CREATED_PARAM) &&
	(obj -> nrefs == 0)) {
      __ctalkDeleteObjectInternal (obj);
      return 0;
    }
    if ((obj -> scope & SUBEXPR_CREATED_RESULT) &&
	(obj -> nrefs == 0)) {
      __ctalkDeleteObjectInternal (obj);
      return 0;
    }
    if (cleanup_mode == rt_cleanup_obj_ref) {
      if (obj && *(char *)obj == 'O' && 
	  IS_OBJECT(obj) && (obj -> nrefs == 0)) 
	__ctalkDeleteObjectInternal (obj);
    }
    return ((nrefs > 0) ? nrefs : 0);
  }
  return ERROR;
}

int _cleanup_temporary_objects_all_instances (OBJECT *obj, 
					      OBJECT *result_obj, 
					      OBJECT *subexpr_obj,
					      RT_CLEANUP_MODE cleanup_mode,
					      MESSAGE_STACK messages,
					      int idx, 
					      int start_idx,
					      int end_idx,
					      int *all_objects_deleted) {

  int nrefs;
  if (!IS_OBJECT(obj)) {
    return ERROR;
  }
  if (object_is_deletable(obj, result_obj, subexpr_obj)) {
    if (obj -> scope & METHOD_USER_OBJECT) {
      if (obj -> nrefs > 0)
	delete_all_objs_expr_frame (messages, idx, start_idx, end_idx,
				    all_objects_deleted);
      return obj -> nrefs;
    } else {
      delete_all_objs_expr_frame (messages, idx, start_idx, end_idx,
				  all_objects_deleted);
      nrefs = 0;
      return nrefs;
    }
  }
  if (obj->__o_p_obj && is_receiver(obj->__o_p_obj))
    return ERROR;

  if (object_is_decrementable(obj, result_obj, subexpr_obj)) {
    nrefs = obj -> nrefs;
    --nrefs;
    if (obj -> nrefs < 0)
      __objRefCntSet (OBJREF(obj), 0);
    delete_all_objs_expr_frame (messages, idx, start_idx, end_idx,
				all_objects_deleted);
    if (nrefs < 0)
      return 0;
    if ((obj -> scope & CREATED_PARAM) &&
	(obj -> nrefs == 0))
      delete_all_objs_expr_frame (messages, idx, start_idx, end_idx,
				  all_objects_deleted);
    if (cleanup_mode == rt_cleanup_obj_ref) {
      if (obj && *(char *)obj == 'O' && 
	  IS_OBJECT(obj) && (obj -> nrefs == 0)) 
	delete_all_objs_expr_frame (messages, idx, start_idx, end_idx,
				    all_objects_deleted);
    }
    return ((nrefs > 0) ? nrefs : 0);
  }
  return ERROR;
}

