/* $Id: ssearch.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

/* 
 *   __ctalkSearchBuffer () uses the Boyer-Moore algorithm's technique 
 *   of searching from the rightmost character of the pattern 
 *   string.  It does not maintain a skip table due to the 
 *   overhead of memory allocation.
 *
 *   Returns a -1 terminated array of match positions in the offets
 *   argument.
 *
 *   Adapted from the Perl StringSearch.pm module, available at
 *   http://www.cpan.org/.
 */
int __ctalkSearchBuffer (char *pattern, char *buffer, long long *offsets) {

  int s, skip;
  long long n;
  int n_offsets;
  int buffer_length, searchable_length;
  int pattern_length;

  buffer_length = strlen (buffer) - 1;
  pattern_length = strlen (pattern) - 1;
  if (buffer_length < pattern_length) {
    offsets[0] = -1L;
    return 0;
  }
  searchable_length = buffer_length - pattern_length;

  n_offsets = 0;
  n = 0L;
  while (n <= searchable_length) {
    skip = 0;
    for (s = pattern_length; s >= 0; s--) {
      if (pattern[s] == buffer[n + s]) {
	++skip;
      } else {
	++n;
	break;
      }
    }
    /***/
    /*
     *  Older compilers can decrement s after finishing the
     *  loop.  Newer compilers seem not to do this.  Still
     *  checking.
     */
   /* if (s < 0) { */
    if (s <= 0) {
      offsets[n_offsets++] = n;
      offsets[n_offsets] = -1;
    }
    n += skip;
  }
  return n_offsets;
}

