/* $Id: val.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2012  Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/* Functions to check and set VAL's. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lex.h"
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "typeof.h"

extern I_PASS interpreter_pass;   /* Declared in rtinfo.c. */

int m_print_val (MESSAGE *m, VAL *val) {
  switch (val -> __type)
      {
      case INTEGER_T:
	sprintf (m -> value, "%d", val -> __value.__i);
	break;
      case LONG_T:
	sprintf (m -> value, "%ld", val -> __value.__l);
	break;
      case LONGLONG_T:
	sprintf (m -> value, "%lld", val -> __value.__ll);
	break;
      case DOUBLE_T:
      case FLOAT_T:
	sprintf (m -> value, "%f", val -> __value.__d);
	break;
#ifndef __APPLE__
      case LONGDOUBLE_T:
	sprintf (m -> value, "%Lf", val -> __value.__ld);
#endif
	break;
      case PTR_T:
	if (IS_OBJECT ((OBJECT *)val -> __value.__ptr)) {
	  OBJECT *o, *o_value;
	  o = (OBJECT *)val -> __value.__ptr;
	  if ((o_value = __ctalkGetInstanceVariable (o, "value", FALSE))
	      == NULL)
	    o_value = o;

	  if (strlen (o_value -> __o_value) >= MAXLABEL) {
	    if ((m -> value = realloc (m -> value, 
				       strlen ((char *)m -> value) + 1)) 
		== NULL) 
	      _error ("m_print_val: allocation error.\n");
	  }
	  sprintf (m -> value, "%s", o_value -> __o_value);
	} else {
	  if (strlen ((char *)val -> __value.__ptr) >= MAXLABEL) {
	    if ((m -> value = realloc (m -> value, 
				       strlen ((char *)m -> value) + 1)) 
		== NULL) 
	      _error ("m_print_val: allocation error.\n");
	  }
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
	  sprintf (m -> value, "%#lx", (unsigned long int)val -> __value.__ptr);
#else
	  sprintf (m -> value, "%#x", (unsigned int)val -> __value.__ptr);
#endif
	}
	break;
      default:
	_warning ("Unknown type %d in m_print_val.\n", val -> __type);
	break;
      }
  return val -> __type;
}
    
int d_print_val (DEFINITION *d, VAL *val) {
  switch (val -> __type)
      {
      case INTEGER_T:
	sprintf (d -> value, "%d", val -> __value.__i);
	break;
      case LONG_T:
	sprintf (d -> value, "%ld", val -> __value.__l);
	break;
      case LONGLONG_T:
	sprintf (d -> value, "%lld", val -> __value.__ll);
	break;
      case DOUBLE_T:
	sprintf (d -> value, "%f", val -> __value.__d);
	break;
#ifndef __APPLE__
      case LONGDOUBLE_T:
	sprintf (d -> value, "%Lf", val -> __value.__ld);
#endif
	break;
      case PTR_T:
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
 	sprintf (d -> value, "%#lx", (unsigned long int)val -> __value.__ptr);
#else
 	sprintf (d -> value, "%#x", (unsigned int)val -> __value.__ptr);
#endif
	break;
      default:
	_warning ("Unknown type %d in d_print_val.\n", val -> __type);
	break;
      }
  return val -> __type;
}
    
int copy_val (VAL *s, VAL *d) {
  d -> __type = s -> __type;
  switch (s -> __type)
    {
    case INTEGER_T:
      d -> __value.__i = s -> __value.__i;
      break;
    case LONG_T:
      d -> __value.__l = s -> __value.__l;
      break;
    case LONGLONG_T:
      d -> __value.__ll = s -> __value.__ll;
      break;
    case DOUBLE_T:
      d -> __value.__d = s -> __value.__d;
      break;
#ifndef __APPLE__
    case LONGDOUBLE_T:
      d -> __value.__ld = s -> __value.__ld;
#endif
      break;
    case PTR_T:
      d -> __value.__ptr = s -> __value.__ptr;
      break;
    default:
      _warning ("Unknown type %d in copy_val.\n", s -> __type);
      break;
    }
  return SUCCESS;
}

int is_val_true (VAL *val) {
  switch (val -> __type)
    {
    case PTR_T:
      return (val -> __value.__ptr != NULL) ? TRUE : FALSE;
      break;
    case INTEGER_T:
      return (val -> __value.__i != 0) ? TRUE : FALSE;
      break;
    case BOOLEAN_T:
      return (val -> __value.__b != 0) ? TRUE : FALSE;
      break;
    case DOUBLE_T:
    case FLOAT_T:
      return (val -> __value.__d != 0.0) ? TRUE : FALSE;
      break;
    case LONG_T:
      return (val -> __value.__l != 0l) ? TRUE : FALSE;
      break;
    case LONGLONG_T:
      return (val -> __value.__ll != 0ll) ? TRUE : FALSE;
      break;
#ifndef __APPLE__
    case LONGDOUBLE_T:
      return (val -> __value.__ld != 0.0l) ? TRUE : FALSE;
#endif
      break;
    case 0:           /* Unitialized __value. */
      return FALSE;
      break;
    }
  _warning ("Unknown type %d in is_val_true.\n", val -> __type);
  return FALSE;
}

/*
 *  NOTE - Numeric_value is used only during expression parsing
 *  by the preprocessor, at this time, so an empty valbuf indicates 
 *  an undefined symbol, which we can evaluate silently to 
 *  FALSE unless we specify otherwise.
 */
int numeric_value (char *valbuf, VAL *val) {

  int r;

  if (interpreter_pass == preprocessing_pass) {
    if (!strlen (valbuf)) {
      val -> __type = INTEGER_T;
      val -> __value.__i = 0;
#ifdef DEBUG_UNDEFINED_PREPROCESSOR_SYMBOLS
      debug ("Undefined preprocessor symbol, \"%s.\"", m -> name);
#endif
      return val -> __type;
    }
  }

  switch (_lextype (valbuf))
  {
  case INTEGER_T:
    val -> __type = INTEGER_T;
    switch (radix_of (valbuf))
      {
      case decimal:
	r = sscanf (valbuf, "%d", &(val-> __value.__i));
	break;
      case octal:
	r = sscanf (valbuf, "%o", &val-> __value.__i);
	break;
      case hexadecimal:
	r = sscanf (valbuf, "%x", &val-> __value.__i);
#if defined (__GNUC__) && defined (__x86_64) && defined (__amd64__)
#else
	if ((long int *)val -> __value.__i == (long int *)val -> __value.__ptr)
#endif
	  val -> __type = PTR_T;
	break;
      case binary:
	break;
      }
    break;
  case LONG_T:
    val -> __type = LONG_T;
    switch (radix_of (valbuf))
      {
      case decimal:
	r = sscanf (valbuf, "%ld", &val-> __value.__l);
	break;
      case octal:
	r = sscanf (valbuf, "%lo", &val-> __value.__l);
	break;
      case hexadecimal:
	r = sscanf (valbuf, "%lx", &val-> __value.__l);
	if ((long int *)val -> __value.__l == (long int *)val -> __value.__ptr)
	  val -> __type = PTR_T;
	break;
      case binary:
	break;
      }
    break;
  case LONGLONG_T:
    val -> __type = LONGLONG_T;
    switch (radix_of (valbuf))
      {
      case decimal:
	r = sscanf (valbuf, "%lld", &val-> __value.__ll);
	break;
      case octal:
	r = sscanf (valbuf, "%llo", &val-> __value.__ll);
	break;
      case hexadecimal:
	r = sscanf (valbuf, "%llx", &val-> __value.__ll);
	if ((long long int *)val -> __value.__l == (long long int *)val -> __value.__ptr)
	  val -> __type = PTR_T;
	val -> __type = PTR_T;
	break;
      case binary:
	break;
      }
    break;
  case FLOAT_T:
  case DOUBLE_T:
    sscanf (valbuf, "%lf", &val-> __value.__d);
    val -> __type = DOUBLE_T;
    break;
#ifndef __APPLE__
  case LONGDOUBLE_T:
    sscanf (valbuf, "%Lf", &val-> __value.__ld);
    val -> __type = LONGDOUBLE_T;
#endif
    break;
  case BOOLEAN_T:
    sscanf (valbuf, "%d", (int *) &val-> __value.__b);
    val -> __type = BOOLEAN_T;
    break;
  case PTR_T:
    val -> __value.__ptr = valbuf;
    val -> __type = PTR_T;
    break;
  case LITERAL_T:
    val -> __value.__ptr = valbuf;
    val -> __type = LITERAL_T;
    break;
  default:
    val -> __type = INTEGER_T;
    val -> __value.__b = FALSE;
#ifdef DEBUG_CODE
    _warning ("Undefined type of %s in numeric___value.\n", valbuf);
#endif
    break;
  }

  return val -> __type;
}

int val_eq (VAL *val1, VAL *val2) {

  int bool_result = FALSE;

  if (val1 -> __type == val2 -> __type) {
    switch (val1 -> __type) 
      {
      case INTEGER_T:
	if (val1 -> __value.__i == val2 -> __value.__i)
	  bool_result = TRUE;
	break;
      case LONG_T:
	if (val1 -> __value.__l == val2 -> __value.__l)
	  bool_result = TRUE;
	break;
      case LONGLONG_T:
	if (val1 -> __value.__ll == val2 -> __value.__ll)
	  bool_result = TRUE;
	break;
      case DOUBLE_T:
	if (val1 -> __value.__d == val2 -> __value.__d)
	  bool_result = TRUE;
	break;
#ifndef __APPLE__
      case LONGDOUBLE_T:
	if (val1 -> __value.__ld == val2 -> __value.__ld)
	  bool_result = TRUE;
#endif
	break;
      case PTR_T:
	if (val1 -> __value.__ptr == val2 -> __value.__ptr)
	  bool_result = TRUE;
	break;
    }
  }

  return bool_result;
}

char *hex_from_numeric_val (char *s) {

  static char buf[MAXMSG];
  char tmpbuf[MAXLABEL];

  switch (_lextype (s))
    {
    case INTEGER_T:
      sprintf (buf, "0x%x", atoi(s));
      break;
    case LONG_T:
      sprintf (buf, "0x%lx", atol(s));
      break;
    case LONGLONG_T:
# if defined(__GNUC__) && defined(__sparc__) && defined(__svr4__)
      /*
       *  SPARC uses atol right now.
       */
      sprintf (buf, "0x%lx", atol(s));
# else
      sprintf (buf, "0x%llx", atoll(s));
# endif
      break;
    case LITERAL_CHAR_T:
      strcpy (tmpbuf, s);
      TRIM_CHAR (tmpbuf);
      sprintf (buf, "0x%x", (int) *tmpbuf);
      break;
    case DOUBLE_T:
    default:
      __ctalkExceptionInternal (NULL, ptr_conversion_x, "",0);
      break;
    }
  return buf;
}
