/* $Id: ltime.c,v 1.1.1.1 2014/05/01 19:46:37 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *   ltime.c - Print the 24-hour time in large block characters.
 */

/* #include <time.h> */

/*
 *  Terminal sequences for XTERM/ANSI/VT100 terminals.
 *  Change these for other terminal types.
 */
#define VT_REV    "\033[7m"     /* Reverse video.                    */
#define VT_NORM   "\033[0m"     /* Normal video.                     */
#define VT_CPOS   "\033[%d;%df" /* Set cursor position: row, column. */
#define VT_CLS    "\033[2J"     /* Clear the screen.                 */
#define VT_RESET  "\033[!p"     /* Soft reset.                       */

#define DIGIT_WIDTH  6
#define DIGIT_HEIGHT 7

static int row, column, digitN;

Array class ClockDigit;         /* The clock digit subclass and      */
                                /* clock digits, which are initial-  */
ClockDigit new digit0;          /* ized below.                       */
ClockDigit new digit1;
ClockDigit new digit2;
ClockDigit new digit3;
ClockDigit new digit4;
ClockDigit new digit5;
ClockDigit new digit6;
ClockDigit new digit7;
ClockDigit new digit8;
ClockDigit new digit9;
ClockDigit new colon;

Array instanceMethod getTime (void) {

  CTime new timeNow;
  Array new currentLocalTime;

  timeNow utcTime;

  currentLocalTime = timeNow localTime;

  self atPut 0, (currentLocalTime at 2);
  self atPut 1, (currentLocalTime at 1);
  self atPut 2, (currentLocalTime at 0);

  return NULL;
}

ClockDigit classMethod initDigits (void) {

  digit0 atPut 0, "001110";
  digit0 atPut 1, "010001";
  digit0 atPut 2, "010001";
  digit0 atPut 3, "010001";
  digit0 atPut 4, "010001";
  digit0 atPut 5, "010001";
  digit0 atPut 6, "001110";

  digit1 atPut 0, "000100";
  digit1 atPut 1, "001100";
  digit1 atPut 2, "010100";
  digit1 atPut 3, "000100";
  digit1 atPut 4, "000100";
  digit1 atPut 5, "000100";
  digit1 atPut 6, "011111";

  digit2 atPut 0, "001110";
  digit2 atPut 1, "010001";
  digit2 atPut 2, "000001";
  digit2 atPut 3, "000110";
  digit2 atPut 4, "001000";
  digit2 atPut 5, "010000";
  digit2 atPut 6, "011111";

  digit3 atPut 0, "011111";
  digit3 atPut 1, "000001";
  digit3 atPut 2, "000010";
  digit3 atPut 3, "001111";
  digit3 atPut 4, "000001";
  digit3 atPut 5, "010001";
  digit3 atPut 6, "001110";

  digit4 atPut 0, "000001";
  digit4 atPut 1, "000011";
  digit4 atPut 2, "000101";
  digit4 atPut 3, "001001";
  digit4 atPut 4, "011111";
  digit4 atPut 5, "000001";
  digit4 atPut 6, "000001";

  digit5 atPut 0, "011111";
  digit5 atPut 1, "010000";
  digit5 atPut 2, "010110";
  digit5 atPut 3, "011001";
  digit5 atPut 4, "000001";
  digit5 atPut 5, "010001";
  digit5 atPut 6, "001110";

  digit6 atPut 0, "001110";
  digit6 atPut 1, "010001";
  digit6 atPut 2, "010000";
  digit6 atPut 3, "011110";
  digit6 atPut 4, "010001";
  digit6 atPut 5, "010001";
  digit6 atPut 6, "001110";

  digit7 atPut 0, "011111";
  digit7 atPut 1, "000001";
  digit7 atPut 2, "000010";
  digit7 atPut 3, "000100";
  digit7 atPut 4, "000100";
  digit7 atPut 5, "001000";
  digit7 atPut 6, "001000";

  digit8 atPut 0, "001110";
  digit8 atPut 1, "010001";
  digit8 atPut 2, "010001";
  digit8 atPut 3, "001110";
  digit8 atPut 4, "010001";
  digit8 atPut 5, "010001";
  digit8 atPut 6, "001110";

  digit9 atPut 0, "001110";
  digit9 atPut 1, "010001";
  digit9 atPut 2, "010001";
  digit9 atPut 3, "001111";
  digit9 atPut 4, "000001";
  digit9 atPut 5, "010001";
  digit9 atPut 6, "001110";

  colon atPut 0, "000000";
  colon atPut 1, "001100";
  colon atPut 2, "001100";
  colon atPut 3, "000000";
  colon atPut 4, "001100";
  colon atPut 5, "001100";
  colon atPut 6, "000000";

  return NULL;
}

ClockDigit instanceMethod printDigitLine (void) {

  int i;
  OBJECT *self_var;

  self_var = __ctalk_self_internal ();

  printf (VT_CPOS, row, column);
  for (i = 0; i < DIGIT_WIDTH; i++) {
    if (self_var -> __o_value[i] == '1') {
      printf ("%s %s", VT_REV, VT_NORM);
    } else {
      printf (" ");
    }
  }
  ++row;
  return NULL;
}

ClockDigit instanceMethod printDigit (void) {

  row = 1; 
  column = digitN * (DIGIT_WIDTH + 1);
  self map printDigitLine;

  return NULL;
}

Integer instanceMethod selectDigit (void) {

  ClockDigit new digit;

  returnObjectClass ClockDigit;

  if (self == 0) {digit = digit0;}
  if (self == 1) {digit = digit1;}
  if (self == 2) {digit = digit2;}
  if (self == 3) {digit = digit3;}
  if (self == 4) {digit = digit4;}
  if (self == 5) {digit = digit5;}
  if (self == 6) {digit = digit6;}
  if (self == 7) {digit = digit7;}
  if (self == 8) {digit = digit8;}
  if (self == 9) {digit = digit9;}

  /* Return the complete object, not "digit value." */
  return digit;

}

int main (int argc, char **argv) {

  Array new clockTime;
  Integer new hourDigit1;
  Integer new hourDigit2;
  Integer new minuteDigit1;
  Integer new minuteDigit2;
  ClockDigit new digit;

  ClockDigit initDigits;

  /* Clear the screen and place the cursor at the upper 
   * left-hand corner.
   */
  printf (VT_CLS);
  printf (VT_CPOS, 1, 1);

  clockTime getTime;

  hourDigit1 = (clockTime at 0) / 10;
  hourDigit2 = (clockTime at 0) % 10;
  minuteDigit1 = (clockTime at 1) / 10;
  minuteDigit2 = (clockTime at 1) % 10;

  digit = hourDigit1 selectDigit;
  digitN = 0;
  digit printDigit;

  digit = hourDigit2 selectDigit;
  digitN = 1;
  digit printDigit;

  digitN = 2;
  colon printDigit;

  digit = minuteDigit1 selectDigit;
  digitN = 3;
  digit printDigit;

  digit = minuteDigit2 selectDigit;
  digitN = 4;
  digit printDigit;

  printf ("\n");

  exit (0);
}

