/* $Id: arg.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"
#include "typeof.h"

extern I_PASS interpreter_pass;    /* Declared in main.c.               */

extern int arg_has_leading_unary;  /* Declared in method.c.             */

extern int ctrlblk_pred;           /* Declared in control.c.            */

extern int argblk;                 /* Declared in argblk.c.             */

/* extern METHOD *new_methods[MAXARGS+1];*/ /* Declared in lib/rtnwmthd.c.   */
extern NEWMETHOD *new_methods[MAXARGS+1]; /* Declared in lib/rtnwmthd.c.   */
extern int new_method_ptr;

extern HASHTAB declared_global_variables;
extern HASHTAB declared_typedefs;
extern HASHTAB declared_functions;

extern OBJECT *rcvr_class_obj;     /* Declared in primitives.c.        */

/*
 *  Find the end of an argument list, recursing if necessary
 *  for functions and methods as arguments.  
 *
 *  If the argument list begins with a parenthesis, simply
 *  find the matching closing parenthesis.
 *
 *  The parameter, n_args, should come from the method, or -1 
 *  if the number of args is unknown.  
 * 
 *  Empty argument lists get handled immediately.  They generate
 *  warning messages in method_args () if method -> n_args > 0.
 */

int paren_expr (MESSAGE_STACK messages, int open_paren_idx) {

  int _m_paren;

  if ((_m_paren = match_paren (messages, open_paren_idx, 
			       get_stack_top(messages)))
      == ERROR) {
    __ctalkExceptionInternal (messages[open_paren_idx], 
			      mismatched_paren_x, NULL,0);
    return ERROR;
  }

  return _m_paren;
}

int method_arg_limit (MESSAGE_STACK messages, int arg_start) {

  int n_parens, n_subscripts, i, stack_end_idx, lookahead;
  int arg_arglist_limit;

  n_parens = n_subscripts = i = stack_end_idx = 0;

  switch (M_TOK(messages[arg_start]))
    {
    case OPENPAREN:
      if ((i = paren_expr (messages, arg_start)) == ERROR)
	return ERROR;
      if (interpreter_pass == expr_check) {
	int __stack_top;
	__stack_top = get_stack_top (messages);
	if (i == (__stack_top + 1))
	  return i;
      }
      if ((lookahead = nextlangmsg (messages, i)) == ERROR)
 	return ERROR;
      /*
       *  If we don't find an argument terminator immediately
       *  after the expression's closing parenthesis, scan
       *  forward until we do.
       */
      switch (M_TOK(messages[lookahead]))
	{
	case CLOSEPAREN:
	case SEMICOLON:
	case ARGSEPARATOR:
	  return i;
	  /* break; */   /* Not reached. */
	default:
	  /*
	   *  The function also needs to handle cases like 
	   *    a = (1 + 1) + (2 + 2); 
	   *    a = (1 + 1) + (2 + 2) + (3 + 3);
	   *  etc., here.
	   */
	  if (IS_C_BINARY_MATH_OP(M_TOK(messages[lookahead]))) {
	    if (!n_parens) {
 	      if ((lookahead = nextlangmsg (messages, lookahead)) == ERROR)
 		return ERROR;
	      return method_arg_limit (messages, lookahead);
	    }
	  } else {
	    for (; i > stack_end_idx; i--) {
 	      if ((lookahead = nextlangmsg (messages, i)) == ERROR)
 		return ERROR;
	      switch (M_TOK(messages[lookahead]))
		{
		case CLOSEPAREN:
		case SEMICOLON:
		case ARGSEPARATOR:
		  return i;
		  /* break; */  /* Not reached. */
		}
	    }
	  }
	}
      break;
    default:
      /*
       *  Arguments that don't begin with a parenthesis.
       */
      stack_end_idx = get_stack_top (messages);
      for (i = arg_start; i > stack_end_idx; i--) {
	if (M_ISSPACE(messages[i]))
	  continue;
	if ((lookahead = nextlangmsg (messages, i)) == ERROR) {
	  if (EXPR_CHECK) {
	    return i;
	  } else {
	    return ERROR;
	  }
	}
	switch (M_TOK(messages[i]))
	  {
	  case LABEL:
	    if ((arg_arglist_limit = resolve_arg2 (messages, i)) != 0) {
	      /*
	       *  Reset i and lookahead to end of an internal method
	       *  arglist.
	       */
	      i = arg_arglist_limit;
 	      if ((lookahead = nextlangmsg (messages, i)) == ERROR) {
		if (EXPR_CHECK) {
		  return i;
		} else {
		  return ERROR;
		}
	      }
	    }
	    break;
	  case OPENPAREN:
	    ++n_parens;
	    break;
	  case ARRAYOPEN:
	    ++n_subscripts;
	    break;
	  case ARRAYCLOSE:
	    if (--n_subscripts < 0) return i;
	    break;
	  case CLOSEPAREN:
	    if (--n_parens < 0) return i;
	    break;
	  }
	switch (M_TOK(messages[lookahead])) 
	  {
	  case ARGSEPARATOR:
	    if (!n_parens) return i;
	    break;
	  case CLOSEPAREN:
	    if (!n_parens) return i;
	    break;
	  case ARRAYCLOSE:
	    if (!n_subscripts) return i;
	    break;
	  case SEMICOLON:
	    return i;
	    /* 	    break; */  /* Not reached. */
	  }
      }
      break;
    }
  return ERROR;
}

/*
 * Like above, but it doesn't do any object checking.  Only for 
 * method_arglist_n_args (), below.
 */
static int method_arg_limit_2 (MESSAGE_STACK messages, int arg_start) {

  int n_parens, n_subscripts, i, stack_end_idx, lookahead;
  int arg_arglist_limit;

  n_parens = n_subscripts = i = stack_end_idx = 0;

  switch (M_TOK(messages[arg_start]))
    {
    case OPENPAREN:
      if ((i = paren_expr (messages, arg_start)) == ERROR)
	return ERROR;
      if (interpreter_pass == expr_check) {
	int __stack_top;
	__stack_top = get_stack_top (messages);
	if (i == (__stack_top + 1))
	  return i;
      }
      if ((lookahead = nextlangmsg (messages, i)) == ERROR)
 	return ERROR;
      /*
       *  If we don't find an argument terminator immediately
       *  after the expression's closing parenthesis, scan
       *  forward until we do.
       */
      switch (M_TOK(messages[lookahead]))
	{
	case CLOSEPAREN:
	case SEMICOLON:
	case ARGSEPARATOR:
	  return i;
	  /* break; */   /* Not reached. */
	default:
	  /*
	   *  The function also needs to handle cases like 
	   *    a = (1 + 1) + (2 + 2); 
	   *    a = (1 + 1) + (2 + 2) + (3 + 3);
	   *  etc., here.
	   */
	  if (IS_C_BINARY_MATH_OP(M_TOK(messages[lookahead]))) {
	    if (!n_parens) {
 	      if ((lookahead = nextlangmsg (messages, lookahead)) == ERROR)
 		return ERROR;
	      return method_arg_limit (messages, lookahead);
	    }
	  } else {
	    for (; i > stack_end_idx; i--) {
 	      if ((lookahead = nextlangmsg (messages, i)) == ERROR)
 		return ERROR;
	      switch (M_TOK(messages[lookahead]))
		{
		case CLOSEPAREN:
		case SEMICOLON:
		case ARGSEPARATOR:
		  return i;
		  /* break; */  /* Not reached. */
		}
	    }
	  }
	}
      break;
    default:
      /*
       *  Arguments that don't begin with a parenthesis.
       */
      stack_end_idx = get_stack_top (messages);
      for (i = arg_start; i > stack_end_idx; i--) {
	if (M_ISSPACE(messages[i]))
	  continue;
	if ((lookahead = nextlangmsg (messages, i)) == ERROR) {
	  if (EXPR_CHECK) {
	    return i;
	  } else {
	    return ERROR;
	  }
	}
	switch (M_TOK(messages[i]))
	  {
	  case LABEL:
	  case METHODMSGLABEL:
	    if ((arg_arglist_limit = resolve_arg2 (messages, i)) != 0) {
	      /*
	       *  Reset i and lookahead to end of an internal method
	       *  arglist.
	       */
	      i = arg_arglist_limit;
	      if ((lookahead = nextlangmsg (messages, i)) == ERROR) {
	     	if (EXPR_CHECK) {
	     	  return i;
	     	} else {
	     	  return ERROR;
	     	}
	      }
	    }
	    break;
	  case OPENPAREN:
	    ++n_parens;
	    break;
	  case ARRAYOPEN:
	    ++n_subscripts;
	    break;
	  case ARRAYCLOSE:
	    if (--n_subscripts < 0) return i;
	    break;
	  case CLOSEPAREN:
	    if (--n_parens < 0) return i;
	    break;
	  }
	switch (M_TOK(messages[lookahead])) 
	  {
	  case ARGSEPARATOR:
	    if (!n_parens) return i;
	    break;
	  case CLOSEPAREN:
	    if (!n_parens) return i;
	    break;
	  case ARRAYCLOSE:
	    if (!n_subscripts) return i;
	    break;
	  case SEMICOLON:
	    return i;
	    /* 	    break; */  /* Not reached. */
	  }
      }
      break;
    }
  return ERROR;
}

int method_arglist_limit (MESSAGE_STACK messages, int arglist_start,
			  int n_params, int varargs) {

  int arg_start, arg_end, lookahead, n_args;

  arg_start = arglist_start;
  n_args = 0;

  if (METHOD_ARG_TERM_MSG_TYPE(messages[arglist_start]))
    return arglist_start;

  /*
   *  Instance and class variables.
   */
  if ((arg_end = complex_arglist_limit (messages, arglist_start)) != ERROR) {
    return arg_end;
  }

  while (TRUE) {
    if ((arg_end = method_arg_limit (messages, arg_start)) == ERROR)
      return ERROR;
    if ((lookahead = nextlangmsg (messages, arg_end)) == ERROR) {
      if (EXPR_CHECK) {
	return arg_end;
      } else {
	__ctalkExceptionInternal (messages[arglist_start], 
				  parse_error_x, NULL,0);
	return ERROR;
      }
    }

    switch (M_TOK(messages[lookahead]))
      {
      case CLOSEPAREN:
      case SEMICOLON:
      case ARRAYCLOSE:
	return arg_end;
	/* 	break; */  /* Not reached.*/
      case ARGSEPARATOR:
	if ((n_args < (n_params - 1)) || varargs) {
	  ++n_args;
 	  if ((arg_start = nextlangmsg (messages, lookahead)) == ERROR) {
	    __ctalkExceptionInternal (messages[arglist_start], 
				      parse_error_x, NULL,0);
	    return ERROR;
	  }
	} else {
	  return arg_end;
	}
	break;
      }
  }

  return ERROR;
}

/*
 *  Same as above, but also calls compound_method_limit, for now
 *  called only by method_args ().
 */
int method_arglist_limit_2 (MESSAGE_STACK messages, 
			    int method_msg_ptr, int arglist_start,
			    int n_params, int varargs) {

  int arg_start, arg_end, lookahead, n_args;

  arg_start = arglist_start;
  n_args = 0;

  if (METHOD_ARG_TERM_MSG_TYPE(messages[arglist_start]))
    return arglist_start;

  /*
   *  Instance and class variables.
   */
  if ((arg_end = complex_arglist_limit (messages, arglist_start)) != ERROR) {
    return arg_end;
  }

  if ((arg_end = compound_method_limit (messages, method_msg_ptr,
					arglist_start,
					n_params, varargs)) != ERROR) {
    return arg_end;
  }

  while (TRUE) {
    if ((arg_end = method_arg_limit (messages, arg_start)) == ERROR)
      return ERROR;
    if ((lookahead = nextlangmsg (messages, arg_end)) == ERROR) {
      if (EXPR_CHECK) {
	return arg_end;
      } else {
	__ctalkExceptionInternal (messages[arglist_start], 
				  parse_error_x, NULL,0);
	return ERROR;
      }
    }

    switch (M_TOK(messages[lookahead]))
      {
      case CLOSEPAREN:
      case SEMICOLON:
      case ARRAYCLOSE:
	return arg_end;
	/* 	break; */  /* Not reached.*/
      case ARGSEPARATOR:
	if ((n_args < (n_params - 1)) || varargs) {
	  ++n_args;
 	  if ((arg_start = nextlangmsg (messages, lookahead)) == ERROR) {
	    __ctalkExceptionInternal (messages[arglist_start], 
				      parse_error_x, NULL,0);
	    return ERROR;
	  }
	} else {
	  return arg_end;
	}
	break;
      }
  }

  return ERROR;
}

/*
 * Check for a comma preceding the receiver.
 * Doesn't look count parens within the arglist.
 */
static int is_comma_before_receiver (MESSAGE_STACK messages, 
				   int method_msg_idx) {
  int stack_start_idx, i, i_prev;
  stack_start_idx = stack_start (messages);

  if (messages[method_msg_idx] -> receiver_msg) {
    for (i = method_msg_idx; i <= stack_start_idx; i++) 
      if (messages[i] == messages[method_msg_idx] -> receiver_msg)
	break;
    if ((i_prev = prevlangmsg (messages, i)) != ERROR) {
      if (M_TOK(messages[i_prev]) == ARGSEPARATOR) {
	return TRUE;
      } else {
	return FALSE;
      }
    } else {
      return FALSE;
    }
  } else {
    return FALSE;
  }
}

/*
 * Returns 
 *   n The number of arguments in the argument list, or:
 *
 *  -1 on parser error (like a stack error).
 *  -2 if the arglist is for defineInstanceMethod () and defineClassMethod (),
 *    which have the form of a function declaration.
 *  -3 if the arglist is an instance or class variable declaration.
 *  -4 if the arglist is in a control statement predicate.
 *
 */
int method_arglist_n_args (MESSAGE_STACK messages, int method_msg_idx) {
			  
  int arglist_start, arg_start, arg_end, lookahead, n_args,
    close_paren_idx; 
  int comma_delimits_arglist;
  METHOD *method = NULL;

  if (ctrlblk_pred)
    return -4;

  /* defineInstanceMethod () and defineClassMethod () */
  if (is_ctalk_keyword (M_NAME(messages[method_msg_idx])))
    return -2;

  if (str_eq (M_NAME(messages[method_msg_idx]), 
	      "instanceVariable") ||
      str_eq (M_NAME(messages[method_msg_idx]),
	      "classVariable"))
    return -3;

  /*
   * Do a pre-check here if we've already found a method.
   */
  if (M_TOK(messages[method_msg_idx]) == METHODMSGLABEL) {
    if (IS_OBJECT(messages[method_msg_idx]->receiver_obj)) {
      if (((method = get_instance_method 
	   (messages[method_msg_idx], 
	    messages[method_msg_idx] -> receiver_obj,
	    M_NAME(messages[method_msg_idx]), 
	    ERROR, FALSE)) != NULL) ||
	  ((method = get_class_method 
	    (messages[method_msg_idx], 
	     messages[method_msg_idx] -> receiver_obj,
	     M_NAME(messages[method_msg_idx]),
	     ERROR, FALSE)) != NULL)) {

	/* 
	 * TODO - Check for *any* method with this class
	 * and name and see if any of them can fit with
	 * the arglist.
	 */
	if (method -> n_params == 0)
	  return 0;

      }
    }
  }

  /*
   *  If there's a comma before the receiver, then 
   *  the next comma (at this level of parentheses)
   *  is end of the arglist.
   */
  comma_delimits_arglist = is_comma_before_receiver
    (messages, method_msg_idx);

  if ((arglist_start = nextlangmsg (messages, method_msg_idx)) == ERROR)
    return ERROR;

  /* Adjust for parentheses. */
  if (M_TOK(messages[arglist_start]) == OPENPAREN) {
    if ((close_paren_idx = match_paren (messages, arglist_start,
					get_stack_top (messages)))
	== ERROR)
      return ERROR;

    if ((lookahead = nextlangmsg (messages, close_paren_idx)) == ERROR)
      return ERROR;

    /* Make sure that the close paren is the end of the complete argument
       list before adjusting. */
    if ((M_TOK(messages[lookahead]) == CLOSEPAREN) ||
	(M_TOK(messages[lookahead]) == SEMICOLON) ||
	(M_TOK(messages[lookahead]) == ARRAYCLOSE)) {

      if ((arglist_start = nextlangmsg (messages, arglist_start)) ==
	  ERROR)
	return ERROR;
    }

  }

  arg_start = arglist_start;

  if (METHOD_ARG_TERM_MSG_TYPE(messages[arglist_start]))
    return 0;

  n_args = 1;

  while (TRUE) {
    if ((arg_end = method_arg_limit_2 (messages, arg_start)) == ERROR)
      return ERROR;
    if ((lookahead = nextlangmsg (messages, arg_end)) == ERROR) {
      if (EXPR_CHECK) {
	return ERROR;
      } else {
	__ctalkExceptionInternal (messages[arg_start], 
				  parse_error_x, NULL,0);
	return ERROR;
      }
    }

    switch (M_TOK(messages[lookahead]))
      {
      case CLOSEPAREN:
	/* We rely on method_arg_limit_2 () to match the parens
	   that belong to the last argument, so this paren 
	   is the one that encloses the entire arglist. */
      case SEMICOLON:
      case ARRAYCLOSE:
	return n_args;
	/* 	break; */  /* Not reached.*/
      case ARGSEPARATOR:
	/* Using a greedy match if there are no parens. */
	if (comma_delimits_arglist) {
	  if (method) {
	    if (n_args == method -> n_params) {
	      return n_args;
	    }
	  }
	}
	++n_args;
	if ((arg_start = nextlangmsg (messages, lookahead)) == ERROR)
	  return ERROR;
	break;
      }
  }

  return ERROR;
}

int global_constructor_arg (METHOD *method, int method_msg_ptr) {

  MESSAGE *m_method,
    *m_arg;
  METHOD *m_actual;
  OBJECT *rcvr;
  int scope,
    arg_ptr;

  if (strcmp (method -> name, "new") ||
      ((scope = parser_frame_scope ()) != GLOBAL_VAR))
    return ERROR;

  m_method = message_stack_at (method_msg_ptr);

  rcvr = m_method -> receiver_obj;

  m_actual = method_replaces_primitive (method, rcvr);

  if ((arg_ptr = nextlangmsg (message_stack (), method_msg_ptr))
      == ERROR) 
   _error ("global_constructor_arg (): Argument for %s not found.",
	    m_method -> name);

  m_arg = message_stack_at (arg_ptr);

  m_actual -> args[0] = 
    create_arg_init (create_object (rcvr -> __o_name, m_arg -> name));
  strcpy (m_actual -> args[0] -> obj -> __o_superclassname, 
	  rcvr -> __o_superclassname);

  ++m_actual -> n_args;

  if (m_arg -> obj) {_warning ("Object needs to be released in global_constructor_arg.\n"); }
  m_arg -> obj = ARG_OBJECT(m_actual -> args[0]);

  return SUCCESS;
}

/*
 *  Return the index of a function argument label matching arg_object
 *  if it is within a parenthesized argument list after start_idx, 
 *  FALSE otherwise.
 */
int fn_is_arg (OBJECT *arg_object, MESSAGE_STACK messages, 
	       int arglist_start_idx) {

  int i;
  int arglist_end_idx;
  MESSAGE *m_tok;

  arglist_end_idx = method_arglist_limit (messages, arglist_start_idx, 0, 
					  FALSE);

  for (i = arglist_start_idx; i > arglist_end_idx; i--) {
    m_tok = messages[i];
    if (M_ISSPACE(m_tok)) continue;
    if (M_TOK(m_tok) == LABEL) {
      if (!strstr (M_NAME(m_tok), arg_object -> __o_name))
	return i;
    }
  }

  return FALSE;
}

int fn_arg_expression_call = FALSE;

OBJECT *fn_arg_expression (OBJECT *rcvr_class, METHOD *method, 
			 MESSAGE_STACK messages, int fn_idx) {
  int n_args, arg_start, arg_end, n_th_arg;
  int stack_start_idx;
  char *args[MAXARGS];
  char expr_buf[MAXMSG];
  char namebuf[MAXLABEL];
  OBJECT *arg_object;
  CVAR *arg_cvar = NULL;
  CFUNC *fn_cfunc = NULL;

  stack_start_idx = get_stack_top (messages);

  if ((arg_start = scanforward (messages, fn_idx, 
				stack_start_idx,
				OPENPAREN)) == ERROR) {
    warning (messages[fn_idx], "Function syntax error.");
    return NULL;
  }

  if ((arg_end = match_paren (messages, arg_start, stack_start_idx))
      == ERROR) {
    warning (messages[fn_idx], "Function syntax error.");
    return NULL;
  }

  split_args (messages, arg_start, arg_end, args, &n_args);

  /*
   *  For now, transfer a function CVAR to a CFUNC, which is 
   *  quicker to find.
   *  Superceded ... 
   */
  *expr_buf = 0;
  for (n_th_arg = 0; n_th_arg < n_args; n_th_arg++) {

    if (!fn_cfunc) {

      if (strchr (args[n_th_arg], '[')) {
	strcpy (namebuf, subscript_var_basename(args[n_th_arg]));
      } else {
	if (strchr (args[n_th_arg], '.')) {
	  if (_lextype(args[n_th_arg]) == DOUBLE_T) {
	    strcpy (namebuf, args[n_th_arg]);
	  } else {
	    strcpy (namebuf, deref_var_basename(args[n_th_arg]));
	  }
	} else {
	  if (strstr (args[n_th_arg], "->")) {
	    strcpy (namebuf, deref_var_basename(args[n_th_arg]));
	  } else {
	    strcpy (namebuf, args[n_th_arg]);
	  }
	}
      }

      if (((arg_cvar = get_local_var (namebuf)) != NULL) ||
	  ((arg_cvar = get_global_var (namebuf)) != NULL)) {
	strcat (expr_buf, args[n_th_arg]);
	if (n_th_arg < (n_args - 1))
	  strcat (expr_buf, ", ");
	continue;
      }

      /* Look for method parameters. */
      if (interpreter_pass == method_pass) {
	METHOD *__n_method;
	int __n_th_param;
	char __argbuf[MAXMSG];
	PARAM *p;
	__n_method = new_methods[new_method_ptr+1] -> method;
	for (__n_th_param = 0; __n_th_param < __n_method -> n_params; 
	     __n_th_param++) {
	  if (!strcmp (__n_method -> params[__n_th_param] -> name, 
		       args[n_th_arg])) {
	    p = __n_method -> params[__n_method->n_params-__n_th_param-1];
 	    SNPRINTF (__argbuf, MAXMSG, "%s (%d)", METHOD_ARG_ACCESSOR_FN,
		      __n_method->n_params-__n_th_param-1);
	    strcat (expr_buf, fmt_rt_return (__argbuf, p -> class, TRUE));
	  }
	}
	continue;
      }
    }

    arg_object = eval_arg (method, rcvr_class, args[n_th_arg], fn_idx);

    if (arg_object) {
      OBJECT *__tmp_arg_object;
      if (!strcmp (arg_object -> __o_classname, "Expr")) {
	/*
	 *  Within a function, an arg can be used unchanged, 
	 *  unless it uses, "self" as the receiver, in which
	 *  case the arg needs a __ctalkEvalExpr call.
	 * 
	 *  TO DO - Signal to the eval_arg call above that it 
	 *  is not necessary to register the variable as a 
	 *  method arg.
	 */
	if (strstr (arg_object -> __o_name, "self")) {
	  strcat (expr_buf, fmt_eval_expr (arg_object -> __o_name));
	  delete_object (arg_object);
	} else {
	  strcat (expr_buf, args[n_th_arg]);
	  delete_object (arg_object);
	}
      } else {
	/* 
	 *  A CFunction object. 
	 */
	if (!strcmp (arg_object -> __o_classname, "CFunction")) {
	  strcat (expr_buf, arg_object -> __o_name);
	} else {
	  /*
	   *  A genuine, declared object.
	   */
	  if (((__tmp_arg_object = get_object (arg_object -> __o_name,
					       arg_object -> __o_classname))
	       != NULL) && 
	      (__tmp_arg_object == arg_object)) {
	    strcat (expr_buf, 
		    obj_arg_to_c_expr 
		    (messages, fn_idx, arg_object, n_args));
	  } else {
	    /*
	     *  Handle "self" as a receiver in an argument.
	     */
	    if (strstr (arg_object -> __o_name, "self")) {
	      strcat (expr_buf, 
		      obj_arg_to_c_expr (messages, fn_idx, arg_object, 
					 n_args));
	      delete_object (arg_object);
	    } else {
	      if (get_local_var (arg_object -> __o_name) || 
		  global_var_is_declared (arg_object -> __o_name)) {
		/*
		 *  Eval_arg gives C variables an "Expr" class, so
		 *  they are handled above.
		 */
		delete_object (arg_object);
	      }  else {
		/*
		 *  Anything else - make sure that the argument is
		 *  added to expr_buf.
		 */
		strcat (expr_buf, arg_object -> __o_name);
		delete_object (arg_object);
		arg_object = NULL;
	      }
	    }
	  }
	}
      }
    }
    if (n_th_arg < (n_args - 1))
      strcat (expr_buf, ", ");
  }

  for (n_th_arg = 0; n_th_arg < n_args; n_th_arg++)
    free (args[n_th_arg]);

  if (*expr_buf == '\0') {
    /* 
     *  Look for a constant token.
     */
    int __arg_tok_start, __arg_tok_end;
    if ((((__arg_tok_start = nextlangmsg (messages, arg_start)) != ERROR) &&
	 ((__arg_tok_end = prevlangmsg (messages, arg_end)) != ERROR)) &&
	(__arg_tok_start == __arg_tok_end)) {
      if (_lextype (namebuf) != PTR_T) {
	strcpy (expr_buf, namebuf);
      }
    }
  }

  arg_object = create_object (CFUNCTION_CLASSNAME, expr_buf);
  return arg_object;
}

int prev_msg_arg (MESSAGE **messages, int this_msg, int stack_start_idx) {

  int i;

  int top_ptr, top_ptr_frame;

  /* This is from prevlangmsg (), might not be needed here. */
#if 0
  if (interpreter_pass == var_pass) {
    top_ptr = stack_start (messages);
  } else {
    if (messages == message_stack ()) {
      top_ptr_frame = message_frame_top_n (parser_frame_ptr ());
      /* top_ptr_stack = P_MESSAGES; */
      /* top_ptr = (top_ptr_frame < top_ptr_stack) ? top_ptr_frame : */
      /*  	top_ptr_stack; */
      top_ptr = (top_ptr_frame < stack_start_idx) ? top_ptr_frame :
   	stack_start_idx;
    } else {
      top_ptr = stack_start (messages);
      
    }
  }
#endif

  top_ptr_frame = message_frame_top_n (parser_frame_ptr ());

  top_ptr = (top_ptr_frame < stack_start_idx) ? top_ptr_frame :
    stack_start_idx;

  for (i = this_msg + 1; i <= top_ptr; i++) {
    if ((messages[i] -> tokentype != NEWLINE) &&
	(messages[i] -> tokentype != WHITESPACE))
      return i;
  }
  return ERROR;
}

static inline int next_msg_arg (MESSAGE **messages, int this_msg, 
				int stack_end_idx) {

  int i;
  int last_ptr, last_ptr_frame;

  switch (interpreter_pass)
    {
    case library_pass:
    case parsing_pass:
    case method_pass:
      if (messages == message_stack ()) {
	last_ptr_frame = message_frame_top_n (parser_frame_ptr () - 1);
	last_ptr = (last_ptr_frame > stack_end_idx) ? last_ptr_frame :
	  stack_end_idx;
      } else {
	last_ptr = stack_end_idx;
      }
    break;
    default:
      last_ptr = stack_end_idx;
      break;
    }

  for (i = this_msg - 1; i > last_ptr; i--) {
    if (!messages[i] || !IS_MESSAGE (messages[i]))
      return ERROR;
      if ((M_TOK(messages[i]) != NEWLINE) && 
	  (M_TOK(messages[i]) != WHITESPACE))
	return i;
  }
  return ERROR;
}

int split_args (MESSAGE_STACK messages, int start_paren, int end_paren,
		char **args, int *n_args) {

  int i, arg_start_idx, arg_end_idx, n_parens;
  int stack_start_idx, stack_ptr_idx;
  char *t_buf;

  stack_ptr_idx = get_stack_top (messages);

  stack_start_idx = stack_start (messages);

  for (i = start_paren, *n_args = 0, n_parens = 0, arg_start_idx = -1; 
       i >= end_paren; i--) {
    
    if (M_ISSPACE (messages[i])) continue;

    switch (M_TOK(messages[i]))
      {
      case OPENPAREN:
	++n_parens;
	if (n_parens == 1) {
 	  if ((arg_start_idx = next_msg_arg (messages, i, stack_ptr_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    i = arg_start_idx + 1;
	  }
	}
	break;
      case CLOSEPAREN:
	--n_parens;
	if (n_parens == 0) {
	  if ((arg_end_idx = prev_msg_arg (messages, i, stack_start_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    t_buf = collect_tokens (messages, arg_start_idx, arg_end_idx);
	    args[(*n_args)++] = strdup (t_buf);
	    free (t_buf);
	  }
	}
	break;
      case ARGSEPARATOR:
	if (n_parens == 1) {
	  if ((arg_end_idx = prev_msg_arg (messages, i, stack_start_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    t_buf = collect_tokens (messages, arg_start_idx, arg_end_idx);
	    args[(*n_args)++] = strdup (t_buf);
	    free (t_buf);
	  }
 	  if ((arg_start_idx = next_msg_arg (messages, i, stack_ptr_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    i = arg_start_idx + 1;
	  }
	}
	break;
      }
  }
  return SUCCESS;
}

/*
 *  Returns array with arg_start_index, arg_end_index,... 
 *
 */
int split_args_idx (MESSAGE_STACK messages, int start_paren, int end_paren,
		    int *arg_indexes, int *n_arg_indexes) {

  int i, arg_start_idx, arg_end_idx, n_parens;
  int stack_ptr_idx, stack_start_idx;

  stack_ptr_idx = get_stack_top (messages);
  stack_start_idx = stack_start (messages);

  for (i = start_paren, *n_arg_indexes = 0, n_parens = 0, arg_start_idx = -1; 
       i >= end_paren; i--) {
    
    if (M_ISSPACE (messages[i])) continue;

    switch (M_TOK(messages[i]))
      {
      case OPENPAREN:
	++n_parens;
	if (n_parens == 1) {

 	  if ((arg_start_idx = next_msg_arg (messages, i, stack_ptr_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    arg_indexes[(*n_arg_indexes)++] = arg_start_idx;
 	    i = arg_start_idx + 1;
	  }
	}
	break;
      case CLOSEPAREN:
	--n_parens;
	if (n_parens == 0) {
	  if ((arg_end_idx = prev_msg_arg (messages, i, stack_start_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    arg_indexes[(*n_arg_indexes)++] = arg_end_idx;
	  }
	}
	break;
      case ARGSEPARATOR:
	if (n_parens == 1) {
	  if ((arg_end_idx = prev_msg_arg (messages, i, stack_start_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    arg_indexes[(*n_arg_indexes)++] = arg_end_idx;
	  }
 	  if ((arg_start_idx = next_msg_arg (messages, i, stack_ptr_idx)) 
	      == ERROR) {
	    _warning ("Parser error.\n");
	    return ERROR;
	  } else {
	    arg_indexes[(*n_arg_indexes)++] = arg_start_idx;
 	    i = arg_start_idx + 1;
	  }
	}
	break;
      }
  }
  return SUCCESS;
}

char *obj_arg_to_c_expr (MESSAGE_STACK messages, int fn_idx, 
			 OBJECT *arg_object, int n_args) {
  CFUNC *fn;
  CVAR *c, *param_c = NULL;  /* Avoid a warning. */
  int n_th_param;
  char t_buf[MAXMSG];
  char cls_buf[MAXLABEL];
  static char t_buf2[MAXMSG];

  SNPRINTF (t_buf, MAXMSG, "%s(\"%s\")", 
	    EVAL_EXPR_FN, arg_object -> __o_name);
  fn = get_function (M_NAME(messages[fn_idx]));
  c = get_global_var (M_NAME(messages[fn_idx]));
  if (fn) {
    for (n_th_param = 0, param_c = fn -> params; 
	 (n_th_param < (n_args - 1)); n_th_param++) {
      param_c = param_c -> next;
    }
  } else {
    if (c) {
      for (n_th_param = 0, param_c = c -> params; 
	   (n_th_param < (n_args - 1)) && param_c; n_th_param++) {
	param_c = param_c -> next;
      }
    } else {
      warning (messages[fn_idx], "Could not find function definition %s.",
		     M_NAME(messages[fn_idx]));
    }
  }

  if (param_c) {
    strcpy (cls_buf, basic_class_from_cvar (messages[fn_idx], param_c, 0));
    strcpy (t_buf2, fmt_rt_return (t_buf, cls_buf, TRUE));
  } else {
    if (is_translatable_basic_class (arg_object -> __o_classname)) {
      /*
       *  TO DO - Provide an optional warning if we can't find the
       *  type of the argument.
       */
      strcpy (t_buf2, 
	      fmt_rt_return 
	      (t_buf, arg_object->__o_classname,TRUE));
    } else {
      /*
       *  TO DO! - If evaluating a control block predicate, see if we can
       *  avoid creating a function expression altogether, by avoiding
       *  the call to fn_arg_obj () from eval_arg () in method.c.  It should
       *  be okay to avoid evaluating the args completely if we are simply
       *  going to write a template expression.
       */
      if (!ctrlblk_pred && !EXPR_CHECK) {
	warning (messages[fn_idx], 
		 "Could not find %s parameter definition. Type defaulting to int.", 
		 arg_object -> __o_name);
      }
      strcpy (t_buf2, fmt_rt_return (t_buf, INTEGER_CLASSNAME, TRUE));
    }
  }

  return t_buf2;
}

int arg_is_constant (MESSAGE_STACK messages, int start_idx, int end_idx) {

  int i;
  MESSAGE *m_tok;

  for (i = start_idx; i > end_idx; i--) {

    m_tok = messages[i];

    if (M_ISSPACE(m_tok)) continue;

    switch (M_TOK(m_tok))
      {
      case LABEL:
	if (!is_c_keyword(M_NAME(m_tok)))
	  return FALSE;
	break;
      default:
	break;
      }
  }

  return TRUE;
}

char *fmt_c_fn_obj_args_expr (MESSAGE_STACK messages, int fn_idx) {
  static char expr_buf[MAXMSG];
  char arg_buf[MAXMSG];
  char arg_class[MAXLABEL];
  int arglist_start_ptr,
    arglist_end_ptr;
  char *args[MAXARGS];
  int n_args, n_th_arg;
  CVAR *fn_param;
  CFUNC *fn_cfunc;
  MESSAGE *m_tok;
  int arg_tokentype;
  long long int lex_idx;

  if ((arglist_start_ptr = nextlangmsg (messages, fn_idx)) == ERROR)
    error (messages[fn_idx], "fmt_c_fn_obj_args_expr: Parser error.");
  if ((arglist_end_ptr = match_paren (messages, arglist_start_ptr, 
				      get_stack_top (messages))) == ERROR)
    error (messages[fn_idx], "fmt_c_fn_obj_args_expr: Parser error.");
  
  split_args (messages, arglist_start_ptr, arglist_end_ptr, args, &n_args);

  if ((fn_cfunc = get_function (M_NAME(messages[fn_idx]))) == NULL) {
    warning (messages[fn_idx], "Undefined function %s.\n", 
	       M_NAME(messages[fn_idx]));
  }
  SNPRINTF (expr_buf, MAXMSG, "%s (", M_NAME(messages[fn_idx]));
  for (n_th_arg = 0, fn_param = fn_cfunc -> params; 
       (n_th_arg < n_args);
       n_th_arg++, fn_param = fn_param -> next) {
    if (!fn_param) {
      _warning ("fmt_c_fn_obj_args_expr: wrong number of arguments.\n");
      break;
    }
    if (strstr (args[n_th_arg], "self")) {
      strcpy (arg_buf, fmt_eval_expr(args[n_th_arg]));
      strcat (expr_buf, 
	      fmt_rt_return 
	      (arg_buf, 
	       basic_class_from_cvar (messages[fn_idx], fn_param, 0), TRUE));
      if (n_th_arg < (n_args - 1))
	strcat (expr_buf, ",");
      continue;
    }

    /*
     *  TO DO - Getting the argument's token type here is easier 
     *  than having split_args record constant tokens, which 
     *  should not be lvalues in expressions, so this should work
     *  okay, although the function should not need to re-lex
     *  the arguments.
     */
    m_tok = new_message ();
    lex_idx = 0L;
    arg_tokentype = lexical (args[n_th_arg], &lex_idx, m_tok);
    delete_message (m_tok);

    strcpy (arg_class, 
	    basic_class_from_cvar (messages[fn_idx], fn_param, 0));
    strcpy (arg_buf, 
	    fmt_rt_return (fmt_eval_expr (args[n_th_arg]), 
			   arg_class, 
			   (((arg_tokentype == INTEGER) || 
			     (arg_tokentype == DOUBLE) ||
			     (arg_tokentype == LONG) ||
			     (arg_tokentype == LONGLONG) ||
			     (arg_tokentype == LITERAL) ||
			     (arg_tokentype == LITERAL_CHAR)) ? 
			    FALSE : TRUE)));
    strcat (expr_buf, arg_buf);
    if (n_th_arg < (n_args - 1))
      strcat (expr_buf, ",");
    /*
     *  basic_class_from_cvar should print a warning if the
     *  parameter CVAR is is invalid.
     */
    if (!fn_param || !IS_CVAR(fn_param))
      break;
  }
  strcat (expr_buf, ")");

  for (n_th_arg = 0; n_th_arg < n_args; n_th_arg++)
    free (args[n_th_arg]);
  return expr_buf;
}

MESSAGE *var_messages[N_VAR_MESSAGES + 1];
extern int var_messageptr;

int resolve_arg2 (MESSAGE_STACK messages, int idx) {

  OBJECT *result_object;
  OBJECT *shadow_instance_var_object;
  int prev_msg_idx;
  int m_next_idx;
  int arg_method_arglist_limit;
  METHOD *arg_method;

  /*
   *  Look for a receiver.
   */
  if ((result_object = class_object_search (M_NAME(messages[idx]), 
					    FALSE)) != NULL) {
    if (!M_OBJ(messages[idx])) {
      messages[idx] -> obj = result_object;
    } else {
      messages[idx] -> value_obj = result_object;
    }
    return SUCCESS;
  }
  
  if (!strcmp (M_NAME(messages[idx]), "self")) {
    if ((interpreter_pass != method_pass) && !argblk) {
      __ctalkExceptionInternal (messages[idx], self_without_receiver_x, NULL,0);
      return ERROR;
    }
    if (!M_OBJ(messages[idx])) {
      messages[idx] -> obj = instantiate_self_object ();
    } else {
      messages[idx] -> value_obj = instantiate_self_object ();
    }
    return SUCCESS;
  }

  if ((result_object = get_local_object (M_NAME(messages[idx]), NULL)) 
      != NULL) {
    if (!M_OBJ(messages[idx])) {
      messages[idx] -> obj = result_object;
    } else {
      messages[idx] -> value_obj = result_object;
    }
    return SUCCESS;
  }

  /*
   *  Expression check pass.
   */
  if ((prev_msg_idx = prevlangmsg (messages, idx)) == ERROR)
    return SUCCESS;

  /*
   *  TO DO - This does not check for constant receivers.
   */
  if ((M_TOK(messages[prev_msg_idx]) != LABEL) &&
      (M_TOK(messages[prev_msg_idx]) != METHODMSGLABEL))
    return SUCCESS;

  if (M_VALUE_OBJ(messages[prev_msg_idx])) {
    OBJECT *_rcvr_obj, *_rcvr_value_obj;

    _rcvr_obj = M_VALUE_OBJ(messages[prev_msg_idx]);
    _rcvr_value_obj = (IS_OBJECT (_rcvr_obj -> instancevars) ?
		       _rcvr_obj -> instancevars : 
		       _rcvr_obj);
    if ((arg_method = get_instance_method (messages[idx], 
					   /* M_VALUE_OBJ(messages[prev_msg_idx]), */
					   _rcvr_value_obj,
					   M_NAME(messages[idx]),
					   ERROR,
					   FALSE)) != NULL) {
      if (strcmp (arg_method -> name, "value") && 
	  (shadow_instance_var_object = 
	   get_instance_variable 
	   (M_NAME(messages[idx]),
	    (IS_OBJECT(M_VALUE_OBJ(messages[prev_msg_idx])) ? 
	     M_VALUE_OBJ(messages[prev_msg_idx])->__o_class->__o_name :
	     NULL),
	    FALSE)) != NULL) {
	method_shadows_instance_variable_warning_1
	  (messages[idx], 
	   ((interpreter_pass == method_pass) ? 
	    new_methods[new_method_ptr+1] -> method -> name : NULL),
	   rcvr_class_obj -> __o_name,
	   arg_method -> name, 
	   M_VALUE_OBJ(messages[prev_msg_idx])->__o_classname);
	return ERROR;
      } else {
	m_next_idx = nextlangmsg (messages, idx);
	if (METHOD_ARG_TERM_MSG_TYPE(messages[m_next_idx]))
	  return SUCCESS;
	arg_method_arglist_limit = 
	  method_arglist_limit (messages, m_next_idx, 
				arg_method -> n_params, 
				arg_method -> varargs);
	return arg_method_arglist_limit;
      }
    } else {
      if (M_VALUE_OBJ(messages[prev_msg_idx])) {
	OBJECT *__rcvr_obj = M_VALUE_OBJ(messages[prev_msg_idx]);
	char __p[MAXMSG];
	int start_idx, end_idx, __param_start_idx,
	  __param_end_idx, __idx;
	if (is_method_proto ((IS_CLASS_OBJECT (__rcvr_obj) ?
			      __rcvr_obj -> __o_name : 
			      __rcvr_obj -> __o_classname),
			     M_NAME(messages[idx]))) {
	  /*
	   *  Outputting a method in the middle of another
	   *  method's argument can screw up the output 
	   *  queue, so here we'll just get the number of 
	   *  arguments that the arg_method expects.  
	   *  fn_params needs the opening and closing parens
	   *  of the parameter list.
	   *
	   *  The var_messages stack should always be safe to use
	   *  here.
	   */
	  strncpy (__p, 
		  method_proto 
		  ((IS_CLASS_OBJECT (__rcvr_obj) ?
		    __rcvr_obj -> __o_name : 
		    __rcvr_obj -> __o_classname),
		   M_NAME(messages[idx])), MAXMSG);
	  start_idx = stack_start (var_message_stack ());
	  end_idx = tokenize (var_message_push, __p);
	  __param_end_idx = __param_start_idx = -1;
	  for (__idx = start_idx; 
	       (__idx >= end_idx) & (__param_end_idx == -1); 
	       __idx--) {
	    if (M_TOK(var_messages[__idx]) == OPENPAREN) {
	      __param_start_idx = __idx;
	      __param_end_idx = match_paren (var_messages, 
					     __param_start_idx,
					     end_idx);
	    }
	  }
	  fn_params (var_message_stack (), __param_start_idx, 
		     __param_end_idx);
	  DELETE_MESSAGES(var_messages, var_messageptr, start_idx)

          m_next_idx = nextlangmsg (messages, idx);
          arg_method_arglist_limit = 
	    method_arglist_limit (messages, m_next_idx, 
				  get_param_ptr () + 1, 
				  get_vararg_status ());
	  return arg_method_arglist_limit;
	} else { /* if (is_method_proto ((IS_CLASS_OBJECT (__rcvr_obj) ? */
	  OBJECT *var_object;
	  
	  if (M_VALUE_OBJ(messages[prev_msg_idx])) {
	    if (((var_object = get_instance_variable 
		 (M_NAME(messages[idx]),
		  M_VALUE_OBJ(messages[prev_msg_idx]) -> __o_classname, 
		  FALSE)) != NULL) ||
		((var_object = get_class_variable 
		 (M_NAME(messages[idx]),
		  M_VALUE_OBJ(messages[prev_msg_idx]) -> __o_classname, 
		  FALSE)) != NULL)) {
	      
	      if (!M_OBJ(messages[idx])) {
		messages[idx] -> obj = var_object;
	      } else {
		messages[idx] -> value_obj = var_object;
	      }
	      messages[idx] -> receiver_msg = messages[prev_msg_idx];

	    }

	    return SUCCESS;
	  }

#if 0
	  /***/
	  /***/
	  /* For a later clause, probably. */
	  int stack_end_ptr;
	  stack_end_ptr = get_stack_top (messages);
	  /***/
	  /* start_idx = stack_start (var_message_stack ()); */
	  /* end_idx = tokenize (var_message_push, __p); */
	  __param_end_idx = __param_start_idx = -1;
	  start_idx = idx;
	  if (((end_idx = 
		scanforward (messages, idx, stack_end_ptr, 
			     SEMICOLON)) != -1) ||
	      ((end_idx = 
		scanforward (messages, idx, stack_end_ptr, 
			     CLOSEBLOCK)) != -1) ||
	      ((end_idx = 
		scanforward (messages, idx, stack_end_ptr, 
			     ARRAYCLOSE)) != -1)) {
	      
	    for (__idx = start_idx; 
		 (__idx >= end_idx) & (__param_end_idx == -1); 
	       __idx--) {
	      if (M_TOK(messages[__idx]) == OPENPAREN) {
		__param_start_idx = __idx;
		__param_end_idx = match_paren (messages, 
					       __param_start_idx,
					       end_idx);
	      }
	    }
	    fn_params (messages, __param_start_idx, 
		       __param_end_idx);
	  /* DELETE_MESSAGES(var_messages, var_messageptr, start_idx) */

	    m_next_idx = nextlangmsg (messages, idx);
	    arg_method_arglist_limit = 
	      method_arglist_limit (messages, m_next_idx, 
				    get_param_ptr () + 1, 
				    get_vararg_status ());
	    return arg_method_arglist_limit;
	  }
#endif
	} /* if (is_method_proto ((IS_CLASS_OBJECT (__rcvr_obj) ? ... */
      }
    }
  }

  return 0;
}

/*
 *  Param evaluation needs to be lazy in the front end, in case we
 *  don't know what its class is.  If we evaluate the expression as a
 *  printf argument or a normal function argument, translate it to the
 *  type of the function parameter's prototype.  If the expression isn't a 
 *  function argument, translate into a basic C type.  
 *  TODO: Simple expression evaluation only in this version.
 */
int c_param_expr_arg (MESSAGE_STACK messages, int param_idx) {
  int n, n_th_arg, n_th_param, fn_idx, arg_limit, struct_expr_limit;
  CFUNC *fn;
  CVAR *c, *param_cvar;
  char *buf, buf_esc[MAXMSG], expr_buf[MAXMSG], c_expr_buf[MAXMSG];
  char expr_class[MAXLABEL];
  METHOD *m;

  arg_limit = method_arg_limit (messages, param_idx);
  buf = collect_tokens (messages, param_idx, arg_limit);
  strcpy (buf_esc, escape_str_quotes (buf));
  free (buf);
  sprintf (expr_buf, "%s(\"%s\")", EVAL_EXPR_FN, buf_esc);

  for (n = param_idx; n >= arg_limit; n--) {
    if (((c = get_local_var (M_NAME(messages[n]))) != NULL) ||
	((c = get_global_var (M_NAME(messages[n]))) != NULL)) {
      if (is_struct_or_union_expr (messages, n, 
				   stack_start(messages),
				   get_stack_top (messages))) {
	generate_register_c_method_arg_call 
	  (c, 
	   struct_or_union_expr (messages, n, 
				 get_stack_top(messages), 
				 &struct_expr_limit),
	   LOCAL_VAR,
	   param_idx);
	n = struct_expr_limit;
      } else {
	/*
	 *  Adds the --warnduplicatenames option here.
	 */
	if (global_var_is_declared_not_duplicated 
	    (messages, param_idx, M_NAME(messages[param_idx]))) {
	  generate_register_c_method_arg_call (c, M_NAME(messages[n]),
					       LOCAL_VAR,
					       param_idx);
	}
      }
    }
  }

  /*
   *  This can be much faster.
   */
  if (is_fmt_arg (messages, param_idx, 
		  stack_start (messages),
		  get_stack_top (messages))) {
    strcpy 
      (c_expr_buf, fmt_printf_fmt_arg
       (messages, param_idx,
	stack_start (messages), get_stack_top (messages), expr_buf));
  } else {
    if ((n_th_arg = obj_expr_is_arg (messages, param_idx,
				     stack_start (messages),
				     get_stack_top (messages),
				     &fn_idx)) != ERROR) {
      if ((fn = get_function (M_NAME(messages[fn_idx]))) == NULL) {
	warning (messages[param_idx],
		 "Prototype of C function, \"%s,\" not found.",
		 M_NAME(messages[fn_idx]));
	warning (messages[param_idx],
		 "Argument type of, \"%s,\" defaulting to, \"int.\"",
		 M_NAME(messages[param_idx]));
	strcpy (c_expr_buf, fmt_rt_return (expr_buf, INTEGER_CLASSNAME, TRUE));
	fileout (c_expr_buf, 0, param_idx);
	for (n = param_idx; n >= arg_limit; n--) {
	  ++messages[n] -> evaled;
	  ++messages[n] -> output;
	}
	return ERROR;
      }
      for (n = 0, param_cvar = fn -> params; 
	   (n < n_th_arg) && 
	     (param_cvar && IS_CVAR(param_cvar)); ++n)
	param_cvar = param_cvar -> next;
      if (param_cvar) {
	strcpy (expr_class, basic_class_from_cvar 
		(messages[fn_idx], param_cvar, 0));
	strcpy (c_expr_buf, fmt_rt_return (expr_buf, expr_class, TRUE));
      } else {
	/*
	 *  Use the class of the parameter definition.
	 */
	m = new_methods[new_method_ptr + 1] -> method;
	for (n_th_param = 0; n_th_param < m -> n_params; n_th_param++) {
	  if (!strcmp (m->params[n_th_param]->name, M_NAME(messages[param_idx])))
	    break;
	}
	strcpy 
	  (c_expr_buf, fmt_rt_return 
	   (expr_buf, m->params[n_th_param]->class, TRUE));
      }
    } else { /* ... obj_expr_is_arg ... */
      /*
       *  Simple expression only for now.
       */
      m = new_methods[new_method_ptr+1] -> method;
      for (n_th_param = 0; n_th_param < m -> n_params; n_th_param++) {
	if (!strcmp (m->params[n_th_param]->name, M_NAME(messages[param_idx])))
	  break;
      }
      strcpy 
	(c_expr_buf, fmt_rt_return 
	 (expr_buf, m->params[n_th_param]->class, TRUE));
    }
  } /* is_fmt_arg ... */

  fileout (c_expr_buf, 0, param_idx);
  for (n = param_idx; n >= arg_limit; n--) {
    ++messages[n] -> evaled;
    ++messages[n] -> output;
  }
  return SUCCESS;
}

/*
 *  obj_expr_is_arg can sometimes return an arg for
 *  an "if" statement, so check here.
 */
int obj_expr_is_fn_arg (MESSAGE_STACK messages, int expr_start_idx,
			int stack_start_idx, int stack_end_idx,
			int *__fn_idx) {
  CFUNC *fn;
  int n_th_arg;
  if ((n_th_arg = obj_expr_is_arg (messages, expr_start_idx,
				   stack_start_idx, stack_end_idx,
				   __fn_idx)) != ERROR) {
    if ((fn = get_function (M_NAME(messages[*__fn_idx]))) != NULL) {
      return n_th_arg;
    }
  }
  return ERROR;
}

/*
 *  Only class or instance variable expressions so far - doesn't handle
 *  "self," either.
 */
int complex_arglist_limit (MESSAGE_STACK messages, int arglist_start) {
  OBJECT *rcvr_obj;
  int next_tok_idx, expr_end_idx;
  if ((rcvr_obj = get_object (M_NAME(messages[arglist_start]), NULL)) != NULL) {
    if (!rcvr_obj -> __o_class) {
      if ((rcvr_obj -> __o_class = 
	   get_class_object (rcvr_obj -> __o_classname)) == NULL) {
	warning (messages[arglist_start], 
		 "complex_arglist_limit: could not find class, \"%s.\"",
		 rcvr_obj -> __o_classname);
      }
    }
    messages[arglist_start]->obj = rcvr_obj;
    if ((next_tok_idx = nextlangmsg (messages, arglist_start)) != ERROR) {
      if (get_class_variable (M_NAME(messages[next_tok_idx]),
			      rcvr_obj -> __o_class -> __o_name,
			      FALSE) ||
	  get_instance_variable (M_NAME(messages[next_tok_idx]),
				 rcvr_obj -> __o_class -> __o_name,
				 FALSE)) {
	messages[next_tok_idx]->receiver_obj = rcvr_obj;
	if (complex_var_or_method_message (rcvr_obj, messages, next_tok_idx)) {
	  fmt_rt_expr (messages, arglist_start, &expr_end_idx);
	  return expr_end_idx;
	}
      }
    }
  }
  return ERROR;
}

char *writable_arg_rt_arg_expr (MESSAGE_STACK messages, int start_idx, 
			  int end_idx) {

  int i, j, lookahead;
  int stack_start_idx;
  char expr_buf[MAXMSG];
  static char eval_expr_buf[MAXMSG];
  MESSAGE *m = NULL;   /* Avoid a warning. */
  RT_EXPR_CLASS rt_expr_class = rt_expr_null;

  *expr_buf = 0;

  stack_start_idx = get_stack_top (messages);

  for (i = start_idx; i >= end_idx; i--) {
    m = message_stack_at (i);

    if (libc_fn_needs_writable_args (M_NAME(m))) {
      if ((lookahead = nextlangmsg (messages, i)) != ERROR) {
	if ((messages[i] -> tokentype == LABEL) && 
	    (messages[lookahead] -> tokentype == OPENPAREN)) {
	  char *e;
	  if ((e = clib_fn_expr (messages, i, i, m -> obj)) != NULL) {
	    strcat (expr_buf, e);
	    rt_expr_class = rt_expr_fn;
	    i =  match_paren (messages, lookahead, stack_start_idx);
	    for (j = lookahead; j >= i; j--) {
	      ++messages[j] -> evaled;
	      ++messages[j] -> output;
	    }
	  }
	}
      }
    } else {
      if (IS_C_OP(M_TOK(m))) {
	operand_type_check (message_stack (), i);
	strcat (expr_buf, m -> name);
      } else {
	if (get_function (M_NAME(m))) {
	  if (rt_expr_class == rt_expr_null)
	    rt_expr_class = rt_c_fn;
	  strcat (expr_buf, m -> name);
	}  else {
	  strcat (expr_buf, m -> name);
	}
      }
    }

    /*
     *  If the expression also contains C variables or functions,
     *  they get registered now in register_c_var () called by
     *  method_args (), etc., so check here if the token has
     *  been evaluated and register the C variable here if 
     *  necessary.
     */
    if ((m -> tokentype == LABEL) && !m -> evaled) {
      CVAR *c;
      if (((c = get_local_var (m -> name)) != NULL) ||
	  ((c = get_global_var (m -> name)) != NULL)) {
	register_c_var (message_stack (), i);
      } else {
	if (!strcmp (M_NAME(m), "self")) {
	  if (self_class_or_instance_variable_lookahead 
	      (messages, i)) {
	    rt_expr_class = rt_self_instvar_simple_expr;
	  }
	}
      }
    }
    ++m -> evaled;
    ++m -> output;
  }
      
  sprintf (eval_expr_buf, "%s(%s (\"%s\"), %d)", INT_TRANS_FN, 
	   EVAL_EXPR_FN, escape_str_quotes(expr_buf), FALSE);
  return eval_expr_buf;
}

ARG *create_arg (void) {
  ARG *a;
  a = __xalloc (sizeof (struct _arg));
  a -> sig = ARG_SIG;
  return a;
}

ARG *create_arg_init (OBJECT *o) {
  ARG *a;
  a = create_arg ();
  a -> obj = o;
  return a;
}

void delete_arg (ARG *a) {
  __xfree (a);
}

char *stdarg_fmt_arg_expr (MESSAGE_STACK messages, int method_idx, 
			   METHOD *method) {
  int arglist_start, arglist_end, rcvr_end, i;
  char *tok_buf;
  static char result_buf[MAXMSG];
  if ((rcvr_end = prevlangmsg (messages, method_idx)) == ERROR)
    error (messages[method_idx], "Parser error.");
  if (method -> n_params) {
    if ((arglist_start = nextlangmsg (messages, method_idx)) == ERROR)
      error (messages[method_idx], "Parser error.");
    if ((arglist_end = method_arglist_limit (messages, arglist_start,
					     method -> n_params,
					     method -> varargs)) == ERROR)
      error (messages[method_idx], "Parser error.");
    
    tok_buf = collect_tokens (messages, rcvr_end, arglist_end);
    de_newline_buf (tok_buf);
    sprintf (result_buf, "%s (\"%s\")", EVAL_EXPR_FN, 
	     escape_str_quotes (tok_buf));
    free (tok_buf);
    for (i = rcvr_end; i >= arglist_end; i--) {
      ++messages[i] -> evaled;
      ++messages[i] -> output;
    }
  } else {
    strcpy (result_buf, rt_library_method_call 
	    (M_VALUE_OBJ(messages[rcvr_end]),
	     method, messages, method_idx));
  }
  return result_buf;
}

int method_expr_is_c_fmt_arg (MESSAGE_STACK messages, int method_msg_ptr,
			      int stack_start, int stack_top) {
  int rcvr_tok_idx, r;
  int array_open_idx;
  OBJECT_CONTEXT context;
  if ((rcvr_tok_idx = prevlangmsg (messages, method_msg_ptr)) != ERROR) {
    if (M_TOK(messages[rcvr_tok_idx]) == CLOSEPAREN) {
      if ((rcvr_tok_idx = 
	   match_paren_rev (messages, rcvr_tok_idx, stack_start)) == ERROR) {
	warning (messages[method_msg_ptr], 
		 "method_expr_is_c_fmt_arg: mismatched receiver parentheses.");
	return FALSE;
      }
    } else {
      if (M_TOK(messages[rcvr_tok_idx]) == ARRAYCLOSE) {
	int __i, n_blocks;
	for (__i = rcvr_tok_idx, array_open_idx = -1, n_blocks = 0; 
	     (__i <= stack_start) && (array_open_idx == -1);
	     __i++) {
	  if (M_TOK(messages[__i]) == ARRAYCLOSE)
	    --n_blocks;
	  if (M_TOK(messages[__i]) == ARRAYOPEN)
	    ++n_blocks;
	  if (!n_blocks)
	    array_open_idx = __i;
	}
	rcvr_tok_idx = prevlangmsg (messages, array_open_idx);
      }
    }

    r = is_fmt_arg (messages, rcvr_tok_idx,
		    stack_start, stack_top);
    context = object_context (messages, rcvr_tok_idx);
    return (r && (context == c_argument_context));
  }
  return FALSE;
}

/*
 *  Rather simple function just fills a semantic hole - only checks
 *  whether a label token at the arglist start is also a method (but
 *  not also an instance or class variable message), which means that
 *  the method can *only* have zero parameters. No reason yet to issue
 *  a warning for a method which returns, "Any," either.  Tokens at
 *  the start of an argument list are taken as complex expressions
 *  anyway.  */
int compound_method_limit (MESSAGE_STACK messages, 
			   int method_msg_ptr, int arglist_start,
			   int n_params, int varargs) {
  METHOD *method, *arg_method;
  int prev_tok_idx, arglist_start_2_idx;
  MESSAGE *m_prev_tok, *m, *m_arg_start;
  OBJECT *m_class_object;
  if (n_params || varargs)
    return ERROR;
  if (M_TOK(messages[arglist_start]) != LABEL) 
    return ERROR;

 if ((prev_tok_idx = prevlangmsg (messages, method_msg_ptr)) != ERROR) {
    m_prev_tok = messages[prev_tok_idx];
    m = messages[method_msg_ptr];
    if (((method = get_instance_method 
	  (m_prev_tok, m_prev_tok -> obj, M_NAME(m), 
	   ERROR, FALSE)) != NULL) ||
	((method = get_class_method 
	  (m_prev_tok, m_prev_tok -> obj, M_NAME(m), 
	   ERROR, FALSE)) != NULL)) {
      
      if (method -> returnclass && 
	  ((m_class_object = get_class_object (method ->returnclass)) 
	   != NULL)) {
	m_arg_start = messages[arglist_start];
	if (((arg_method = get_instance_method 
	      (m_arg_start, m_class_object, M_NAME(m_arg_start), 
	       ERROR, FALSE)) 
	     != NULL) ||
	    ((arg_method = get_class_method 
	      (m_arg_start, m_class_object, M_NAME(m_arg_start), 
	       ERROR, FALSE)) 
	     != NULL)) {
  	  if (!get_instance_variable (M_NAME(m_arg_start),
  				      m_class_object -> __o_name,
  				      FALSE) && 
  	      !get_class_variable (M_NAME(m_arg_start),
  				   m_class_object -> __o_name,
  				   FALSE)) {
	    messages[arglist_start] -> tokentype = METHODMSGLABEL;
	    if ((arglist_start_2_idx = 
		 nextlangmsg (messages, arglist_start)) != ERROR) {
	      if (METHOD_ARG_TERM_MSG_TYPE(messages[arglist_start_2_idx])){
		return arglist_start;
	      } else {
		return method_arglist_limit_2 (messages, 
					       arglist_start,
					       arglist_start_2_idx,
					       arg_method -> n_params,
					       arg_method -> varargs);
	      }
	    } else {
	      return arglist_start;
	    }
  	  }
	}
      }
    }
  }

  return ERROR;
}

