/* $Id: argblk.c,v 1.4 2014/05/08 10:39:11 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "list.h"

#define C_ARG_BLK (argblks[argblk_ptr+1])

int have_complex_arg_block = FALSE;  /* Set and reset in complexmetho.c, also */
				     /* reset in method_call ().              */
int complex_arg_block_start = -1;
int complex_arg_block_rcvr_ptr = -1; /* Set by method_call () and      */ 
                                     /* is_argblk_expr                 */

ARGBLK *argblks[MAXARGS + 1];
int argblk_ptr;

static int argblk_id;

/* extern METHOD *new_methods[MAXARGS+1];*/  /* Declared in lib/rtnwmthd.c.  */
extern NEWMETHOD *new_methods[MAXARGS+1];  /* Declared in lib/rtnwmthd.c.  */
extern int new_method_ptr;

extern OBJECT *rcvr_class_obj;

extern I_PASS interpreter_pass;

int argblk;

static LIST *argblk_buffer;

MESSAGE *argblk_messages[N_MESSAGES + 1];
int argblk_messageptr;

MESSAGE_STACK argblk_message_stack (void) {
  return argblk_messages;
}

int get_argblk_message_ptr (void) {
  return argblk_messageptr;
}

int argblk_message_push (MESSAGE *m) {
  if (argblk_messageptr == 0) {
    _error ("argblk_message_push: stack_overflow.");
    return ERROR;
  }

  if (!m)
    _error ("argblk_message_push: null pointer, argblk_messageptr = %d.", 
	    argblk_messageptr);

  argblk_messages[argblk_messageptr] = m;

#ifdef STACK_TRACE
  debug ("argblk_message_push %d. %s.", argblk_messageptr, 
	 (argblk_messages[argblk_messageptr] && 
	  IS_MESSAGE (argblk_messages[argblk_messageptr])) ?
	 argblk_messages[argblk_messageptr] -> name : "(null)"); 
#endif

  --argblk_messageptr;

  return argblk_messageptr + 1;
}

void init_arg_blocks (void) {
  argblk_ptr = MAXARGS;
  argblk_id = argblk = 0;
  argblk_buffer = NULL;
  argblk_messageptr = N_MESSAGES;
}

void argblk_push (ARGBLK *a) {
  argblks[argblk_ptr--] = a;
}

ARGBLK *argblk_pop (void) {
  static ARGBLK *a;
  a = argblks[++argblk_ptr];
  argblks[argblk_ptr] = NULL;
  return a;
}

ARGBLK *new_arg_blk (void) {
  static ARGBLK *a;
  if ((a = (ARGBLK *) __xalloc (sizeof (struct _argblk))) == NULL)
    return NULL;
  return a;
}

int argblk_stmt (int idx) {
  if (argblk_ptr >= MAXARGS)
    return FALSE;
  if ((idx <= C_ARG_BLK->start_idx) &&
      (idx >= C_ARG_BLK->end_idx))
    return TRUE;
  else
    return FALSE;
}

void buffer_argblk_stmt (char *s) {
  LIST *l;
  l = new_list ();
  l -> data = strdup (s);
  list_push (&argblk_buffer, &l);
}

#define HAVE_COMPLEX_ARG_BLOCK (have_complex_arg_block && \
				(complex_arg_block_rcvr_ptr != -1))

int create_argblk (MESSAGE_STACK messages, int argblk_start_idx) {
  ARGBLK *a;
  int i, n_blocks, stack_end;
  int method_idx,
    rcvr_idx;

  if ((method_idx = prevlangmsg (messages, argblk_start_idx))
      == ERROR) {
    warning (messages[C_ARG_BLK->start_idx],
	     "Argument block syntax error.");
    return ERROR;
  }
    
  if (M_TOK(messages[method_idx]) != METHODMSGLABEL) {
    if (!HAVE_COMPLEX_ARG_BLOCK) {
      warning (messages[method_idx], "Argument block has no method.");
      return ERROR;
    }
  }

  if (HAVE_COMPLEX_ARG_BLOCK) {
    rcvr_idx = complex_arg_block_rcvr_ptr;
  } else {
    if (messages[method_idx] -> receiver_msg) {
      MESSAGE *__method_msg, *__rcvr_msg;
      int __i, __stack_start;
      __stack_start = stack_start (messages);
      rcvr_idx = -1;
      __method_msg = messages[method_idx];
      __rcvr_msg = __method_msg -> receiver_msg;
      for (__i = method_idx + 1; 
	   (__i <= __stack_start) && (rcvr_idx == -1); __i++) {
	if (messages[__i] == __rcvr_msg)
	  rcvr_idx = __i;
      }
    } else {
      if ((rcvr_idx = prevlangmsg (messages, method_idx)) == ERROR) {
	warning (messages[C_ARG_BLK->start_idx],
		 "Argument block syntax error.");
	return ERROR;
      }
    }
  }

  if (!HAVE_COMPLEX_ARG_BLOCK) {
    if (messages[method_idx]->receiver_obj != 
	messages[rcvr_idx]->obj) {
      warning (messages[C_ARG_BLK->start_idx],
	       "Invalid argument block receiver.");
      return ERROR;
    }
  }

  stack_end = get_stack_top (messages);

  for (i = argblk_start_idx - 1, n_blocks = 1; i > stack_end; i--) {

    if (M_ISSPACE(messages[i]))
      continue;

    if (M_TOK(messages[i]) == OPENBLOCK)
      ++n_blocks;
    if (M_TOK(messages[i]) == CLOSEBLOCK)
      --n_blocks;

    if (!n_blocks)
      break;
  }
  a = new_arg_blk ();
  a -> id = argblk_id++;
  a -> rcvr_idx = rcvr_idx;
  a -> start_idx = argblk_start_idx;
  a -> end_idx = i;
  argblk = TRUE;
  argblk_push (a);
  return SUCCESS;
}

static char *argblk_fn_name (void) {
  char srcfile[FILENAME_MAX];
  char *r;
  static char s[MAXLABEL];
  if (interpreter_pass == method_pass) {
    sprintf (s, "%s__ctblk%d", 
	     new_methods[new_method_ptr+1] -> method -> selector,
	     C_ARG_BLK->id);
  } else {
    strcpy (srcfile, source_filename ());
    if ((r = strchr (srcfile, '.')) != NULL)
      *r = '\0';
    sprintf (s, "%s__ctblk%d",
	     srcfile, C_ARG_BLK->id);
  }
  return s;
}

void unbuffer_argblk (MESSAGE_STACK messages) {
  int blklength, methodlength;
  LIST *l;
  METHOD *argblk_method;
  char *blk_buf, method_name_buf[MAXMSG], method_fn_buf[MAXMSG], 
    *method_buf;

  if (interpreter_pass == method_pass) {
    unbuffer_method_vartab ();
  } else {
    if (interpreter_pass == parsing_pass) {
      unbuffer_function_vartab ();
    }
  }


  for (l = argblk_buffer, blklength = 0; l; l = l -> next) {
    blklength += strlen ((char *)l -> data);
  }
  blklength *= 2;
  blk_buf = __xalloc (blklength);
  *blk_buf = '\0';
  for (l = argblk_buffer; l; l = l -> next) {
    if (str_eq ((char *)l -> data, "return")) {
      warning (messages[C_ARG_BLK->rcvr_idx], 
	       "\"return,\" used in argument block. Use, \"break,\" instead.");
    }
    if (str_eq ((char *)l -> data, "break")) {
      strcat (blk_buf, "__ctalkExitArgBlockScope (); return ((void *)-1)");
    } else {
      if (strstr ((char *)l -> data, FIXUPROOT))
	continue;
      strcat (blk_buf, (char *)l -> data);
    }
  }
  strcpy (method_name_buf, argblk_fn_name ());
  sprintf (method_fn_buf, "OBJECT *%s (void) ", method_name_buf);

  strcat (blk_buf, "\n");
  methodlength = blklength + strlen (method_fn_buf) + 2;
  method_buf = __xalloc (methodlength);
  sprintf (method_buf, "%s%s", method_fn_buf, blk_buf);
  format_argblk_method (method_buf);

  if (interpreter_pass == method_pass) {
    /* Only needed for methods right now. */
    if ((argblk_method = (METHOD *)malloc (sizeof (METHOD))) == NULL)
      _error ("new_instance_method: %s.", strerror (errno));
    memset (argblk_method, 0, sizeof (struct _method));
    argblk_method -> sig = METHOD_SIG;

    strcpy (argblk_method -> selector, method_name_buf);
    strcpy (argblk_method -> returnclass, OBJECT_CLASSNAME);
    argblk_method -> attrs = METHOD_ARGBLK_ATTR;
    /*
     * These get deleted in delete_newmethod () in 
     * rtnwmthd.c.
     */
    new_methods[new_method_ptr+1] -> 
      argblks[new_methods[new_method_ptr+1] -> argblk_ptr] = 
      argblk_method;
    --new_methods[new_method_ptr+1] -> argblk_ptr;
  }

  __xfree (blk_buf);
  __xfree (method_buf);
  delete_list (&argblk_buffer);
  argblk_buffer = NULL;
  generate_instance_method_definition_call 
    (messages[C_ARG_BLK->rcvr_idx]->obj->__o_classname,
     argblk_fn_name (), method_name_buf, 0, 0);
}

void replace_argblk (MESSAGE_STACK messages) {
  char *token_buf, method_name_buf[MAXMSG], expr_buf[MAXMSG];
  token_buf = collect_tokens (messages, C_ARG_BLK->rcvr_idx,
			     C_ARG_BLK->start_idx+1); 
  sprintf (method_name_buf, "%s", argblk_fn_name ());
  sprintf (expr_buf, "%s(\"%s%s\");", 
	   EVAL_EXPR_FN, token_buf, 
	   method_name_buf);
  fileout (expr_buf, TRUE, 0);
  __xfree (token_buf);
}

int argblk_end (MESSAGE_STACK messages, int idx) {
  ARGBLK *a;
  if (interpreter_pass == expr_check) 
    return FALSE;
  if (argblk_ptr >= MAXARGS)
    return FALSE;
  if (idx <= C_ARG_BLK->end_idx) {
    /*
     *  Trailing brace should not be mistaken for a 
     *  function or method closing block.
     */
    buffer_argblk_stmt (M_NAME(messages[idx]));
    ++messages[idx]->evaled;
    ++messages[idx]->output;
    replace_argblk (messages);
    unbuffer_argblk (messages);
    argblk = FALSE;
    a = argblk_pop ();
    __xfree (a);
    return TRUE;
  }
  return FALSE;
}

void argblock_init (void) {
  char buf[MAXMSG];
  sprintf (buf, "%s ();\n", ARGBLK_ENTER_SCOPE_FN);
  __fileout (buf);
}

void argblock_exit (void) {
  char buf[MAXMSG];
  sprintf (buf, "%s ();\n", ARGBLK_EXIT_SCOPE_FN);
  __fileout (buf);
  sprintf (buf, "%s;\n", ARGBLK_RETURN_STMT);
  __fileout (buf);
}

int format_argblk_method (char *argblk_buf) {
  int i, stack_top, stack_start, n_blocks;
  MESSAGE *m_tok;
  stack_start = argblk_messageptr;
  stack_top = tokenize (argblk_message_push, argblk_buf);
  n_blocks = 0;
  for (i = stack_start; i >= stack_top; i--) {
    m_tok = argblk_messages[i];
    switch (M_TOK(m_tok))
      {
      case OPENBLOCK:
	++n_blocks;
	__fileout (M_NAME(m_tok));
	if (n_blocks == 1)
	  argblock_init ();
	break;
      case CLOSEBLOCK:
	--n_blocks;
	if (n_blocks == 0)
	  argblock_exit ();
	__fileout (M_NAME(m_tok));
	break;
      default:
	__fileout (M_NAME(m_tok));
      }
  }

  DELETE_MESSAGES(argblk_messages,argblk_messageptr,stack_start)
  return SUCCESS;
}

int is_argblk_ctrl_struct (MESSAGE_STACK messages, int idx) {
  if (argblk_ptr >= MAXARGS)
    return FALSE;
  if (((idx <= C_ARG_BLK->rcvr_idx) && (idx >= C_ARG_BLK->start_idx)) ||
      (idx == C_ARG_BLK->end_idx))
    return TRUE;
  return FALSE;
}

/*
 *  Look for a construct <label> <label> <openblock>.
 *  The function works by syntax only, so it is probably only 
 *  useful if the first label is a method parameter as
 *  a receiver.  There is a fixup in resolve () when the
 *  parser determines the methods receiver object and 
 *  receiver message.
 *
 *  Here we just assume that the expression is a complex 
 *  arg block predicate, TO DO - which we should do 
 *  everywhere anyways.
 */
int is_argblk_expr (MESSAGE_STACK messages, int rcvr_idx) {
  int i, lookahead, lookahead2, stack_top;
  stack_top = get_stack_top (messages);
  for (i = rcvr_idx; i > stack_top; i--) {
    if (M_ISSPACE(messages[i]))
      continue;
    if ((M_TOK(messages[i]) != LABEL) &&
	(M_TOK(messages[i]) != OPENBLOCK))
      return FALSE;
    if ((lookahead = nextlangmsg (messages, i)) == ERROR)
      return FALSE;
    if (M_TOK(messages[i]) == LABEL) {
      if (M_TOK(messages[lookahead]) == LABEL) {
	if ((lookahead2 = nextlangmsg (messages, lookahead)) != ERROR) {
	  if (M_TOK(messages[lookahead2]) == OPENBLOCK) {
	    int i_2;
	    messages[lookahead] -> tokentype = METHODMSGLABEL;
	    have_complex_arg_block = TRUE;
	    complex_arg_block_rcvr_ptr = rcvr_idx;
	    create_argblk (messages, lookahead2);
	    for (i_2 = rcvr_idx; i_2 >= lookahead; i_2--) {
	      ++(messages[i_2]->evaled);
	      ++(messages[i_2]->output);
	    }
	    complex_arg_block_rcvr_ptr = -1;
	    have_complex_arg_block = FALSE;
	    return TRUE;
	  } else {
	    if (M_TOK(messages[lookahead2]) == LABEL) {
	      continue;
	    } else {
	      return FALSE;
	    }
	  }
	} else {
	  return FALSE;
	}
      } else {
	return FALSE;
      }
    }
  }
  return FALSE;
}

/*
 *  Similar to above, but it looks for an already looked-up
 *  method. Called from method_args ().
 */
int is_argblk_expr2 (MESSAGE_STACK messages, int rcvr_idx) {
  int i, lookahead, lookahead2, stack_top;
  stack_top = get_stack_top (messages);
  for (i = rcvr_idx; i > stack_top; i--) {
    if (M_ISSPACE(messages[i]))
      continue;
    if ((M_TOK(messages[i]) != LABEL) &&
	(M_TOK(messages[i]) != OPENBLOCK) && 
	(M_TOK(messages[i]) != METHODMSGLABEL))
      return FALSE;
    if ((lookahead = nextlangmsg (messages, i)) == ERROR)
      return FALSE;
    if ((M_TOK(messages[i]) == LABEL) ||
	(M_TOK(messages[i]) == METHODMSGLABEL)) {
      if ((M_TOK(messages[lookahead]) == METHODMSGLABEL) ||
	  (M_TOK(messages[lookahead]) == LABEL)) {
	if ((lookahead2 = nextlangmsg (messages, lookahead)) != ERROR) {
	  if (M_TOK(messages[lookahead2]) == OPENBLOCK) {
	    int i_2;
	    have_complex_arg_block = TRUE;
	    complex_arg_block_rcvr_ptr = rcvr_idx;
	    create_argblk (messages, lookahead2);
	    for (i_2 = rcvr_idx; i_2 >= lookahead; i_2--) {
	      ++(messages[i_2]->evaled);
	      ++(messages[i_2]->output);
	    }
	    complex_arg_block_rcvr_ptr = -1;
	    have_complex_arg_block = FALSE;
	    return TRUE;
	  } else {
	    if (M_TOK(messages[lookahead2]) == LABEL) {
	      continue;
	    } else {
	      return FALSE;
	    }
	  }
	} else {
	  return FALSE;
	}
      } else {
	return FALSE;
      }
    }
  }
  return FALSE;
}

OBJECT *argblk_rcvr_object (MESSAGE_STACK messages, int idx) {
  if (argblk_ptr == MAXARGS)
    return NULL;
  return messages[argblks[argblk_ptr+1]->rcvr_idx] -> obj;
}

int argblk_super_expr (MESSAGE_STACK messages, int idx) {

  if (interpreter_pass == method_pass) {
    OBJECT_CONTEXT super_context;
    super_context = object_context (messages, idx);
    switch (super_context)
      {
      case receiver_context:
	if (argblk_stmt (idx)) {
	  super_argblk_expr 
	    (messages, idx,
	     new_methods[new_method_ptr+1]->method->rcvr_class_obj);
	  return SUCCESS;
	}
	break;
      default:
	break;
      }
  } else { /* if (interpreter_pass == method_pass) */
    OBJECT_CONTEXT super_context;
    OBJECT *blk_rcvr_object;
    super_context = object_context (messages, idx);
    switch (super_context)
      {
      case receiver_context:
	if (argblk_stmt (idx)) {
	  if ((blk_rcvr_object = 
	       argblk_rcvr_object (messages, idx))
	      != NULL) {
	    super_argblk_expr 
	      (messages, idx, blk_rcvr_object);
	    warning (messages[idx], "Keyword, \"super,\" used in a function"
		     " argument block.");
	    warning (messages[idx], "\tUsing, \"%s,\" as the receiver.",
		     M_NAME(messages[argblks[argblk_ptr+1]->rcvr_idx]));
	    return SUCCESS;
	  }
	}
	break;
      default:
	break;
      }
  } /* if (interpreter_pass == method_pass) */

  return ERROR;
}
