/* $Id: collection.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "objtoc.h"

int collection_rt_expr (METHOD *method, MESSAGE_STACK messages, int msg_ptr,
			OBJECT_CONTEXT o_context) {

  char *expr_buf_1;
  char expr_buf[MAXMSG];
  char output_buf[MAXMSG];
  int receiver_ptr,
    arglist_start,
    n_th_arg;

  if (((receiver_ptr = prevlangmsg (messages, msg_ptr)) == ERROR) ||
      ((arglist_start = nextlangmsg (messages, msg_ptr)) == ERROR))
    _error ("collection_rt_expr: parser error.\n");

  expr_buf_1 = collect_tokens (messages, receiver_ptr, arglist_start + 1);
  strcpy (expr_buf, expr_buf_1);
  for (n_th_arg = 0; n_th_arg < method -> n_args; n_th_arg++) {
    strcat (expr_buf, method -> args[n_th_arg] -> obj -> __o_name);
    if (n_th_arg < (method -> n_args - 1))
      strcat (expr_buf, ",");
  }
    


  strcpy (output_buf, fmt_eval_expr (expr_buf));
  free (expr_buf_1);

  switch (o_context)
    {
    case c_context:
    case c_argument_context:
      fileout
	(array_obj_2_c_wrapper (messages, msg_ptr, 
				messages[msg_ptr]->receiver_obj,
				method, output_buf), 
	 FALSE, msg_ptr);
      break;
    default:
      fileout (output_buf, FALSE, msg_ptr);
      break;
    }

  return SUCCESS;
}

int collection_needs_rt_eval (OBJECT *class_obj, METHOD *m) {

  int r;
  if (!IS_OBJECT(class_obj)) return FALSE;

  r = (IS_COLLECTION_SUBCLASS_OBJ(class_obj) &&
       !IS_STREAM_SUBCLASS_OBJ(class_obj) && 
       !IS_PRIMITIVE_METHOD(m) && strcmp (m -> name, "new")
       && (m -> n_params != 0));

  return r;
}


COLLECTION_CONTEXT collection_context (MESSAGE_STACK messages, int rcvr_idx) {

  int fn_idx, n_th_param, n_args;
  int stack_begin, stack_top;
  CFUNC *fn;
  CVAR *c, *param_c = 0;  /* Avoid a warning. */
  char arg_class[MAXLABEL];

  stack_begin = stack_start (messages);
  stack_top = get_stack_top (messages);

  if ((n_th_param = obj_expr_is_fn_arg (messages, rcvr_idx,
					stack_begin,
					stack_top,
					&fn_idx)) != ERROR) {

    fn = get_function (M_NAME(messages[fn_idx]));
    c = get_global_var (M_NAME(messages[fn_idx]));
    if (fn) {
      for (n_args = 0, param_c = fn -> params; 
	   (n_args < n_th_param) && param_c; n_args++) {
	param_c = param_c -> next;
      }
    } else {
      if (c) {
	for (n_args = 0, param_c = c -> params; 
	     (n_args < n_th_param) && param_c; n_args++) {
	  param_c = param_c -> next;
	}
      } else {
	warning (messages[fn_idx], "Could not find function definition %s.",
		 M_NAME(messages[fn_idx]));
      }
    }
    strcpy (arg_class, 
	    basic_class_from_cvar (messages[rcvr_idx], param_c, 0));
    if (!strcmp (arg_class, STRING_CLASSNAME) ||
	(is_subclass_of (arg_class, STRING_CLASSNAME) == TRUE))
      return char_ptr_collection_context;
    if (!strcmp (arg_class, CHARACTER_CLASSNAME) ||
	(is_subclass_of (arg_class, CHARACTER_CLASSNAME) == TRUE))
      return int_collection_context;
    if (!strcmp (arg_class, INTEGER_CLASSNAME) ||
	(is_subclass_of (arg_class, INTEGER_CLASSNAME) == TRUE))
      return int_collection_context;
    return ptr_collection_context;
  }

  if (IS_SIMPLE_SUBSCRIPT (messages, rcvr_idx,
			   stack_begin,
			   stack_top))
    return int_collection_context;

  if (IS_CHAR_PTR_FMT_ARG (messages, rcvr_idx,
			   stack_begin,
			   stack_top ))
    return char_ptr_collection_context;

  return ptr_collection_context;

}

