/* $Id: cvartab.c,v 1.6 2014/05/08 10:44:38 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright © 2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"
#include "typeof.h"

extern NEWMETHOD *new_methods[MAXARGS+1]; /* Declared in lib/rtnwmthd.c.   */
extern int new_method_ptr;

extern char fn_name[MAXLABEL];

void method_cvar_tab_init_stmt (CVAR *var_list) {

  CVAR *c;
  char vartab_entry[MAXLABEL];
  char vartab_init_entry[MAXMSG];

  for (c = var_list; c; c = c -> next) {

    sprintf (vartab_entry, "%s %s%s_%s;\n",
	     c -> type,
	     (((c -> n_derefs == 2) ? "***" : 
	       ((c -> n_derefs == 1) ? "**" : 
		((c -> n_derefs == 0) ? "*" : "")))), 
	     new_methods[new_method_ptr+1] -> method -> selector,
	     c -> name);

    if (c -> n_derefs > 2)
      _warning 
     	("Warning: Variable %s has an unsupported number of dereferences.\n");

    method_vartab_statement (vartab_entry);

    sprintf (vartab_init_entry, "%s_%s = &%s;\n",
     	     new_methods[new_method_ptr+1] -> method -> selector,
     	     c -> name,
     	     c -> name);


    method_vartab_init_statement (vartab_init_entry);
  }
}

void function_cvar_tab_init_stmt (CVAR *var_list) {

  CVAR *c;
  char vartab_entry[MAXLABEL];
  char vartab_init_entry[MAXMSG];

  for (c = var_list; c; c = c -> next) {

    sprintf (vartab_entry, "%s %s%s_%s;\n",
	     c -> type,
	     (((c -> n_derefs == 2) ? "***" : 
	       ((c -> n_derefs == 1) ? "**" : 
		((c -> n_derefs == 0) ? "*" : "")))), 
	     fn_name, c -> name);

    if (c -> n_derefs > 2)
      _warning 
     	("Warning: Variable %s has an unsupported number of dereferences.\n");

    function_vartab_statement (vartab_entry);

    sprintf (vartab_init_entry, "%s_%s = &%s;\n",
	     fn_name, c -> name, c -> name);


    function_vartab_init_statement (vartab_init_entry);
  }
}

int new_method_has_contain_argblk_attr (void) {

  if (new_method_ptr < MAXARGS) {
    if (new_methods[new_method_ptr+1] -> method -> attrs & 
	METHOD_CONTAINS_ARGBLK_ATTR)
      return TRUE;
  }
  return FALSE;
}

char *method_cvar_alias_basename (METHOD *m, CVAR *c) {
  static char s[MAXLABEL];
  sprintf (s, "%s_%s", m -> selector, c -> name);
  return s;
}

char *function_cvar_alias_basename (CVAR *c) {
  static char s[MAXLABEL];
  sprintf (s, "%s_%s", fn_name, c -> name);
  return s;
}

/* 
   method_contains_argblk () is in pattypes.c.  This fn is here
   because find_function_close () is in cparse.c, and this fn uses
   it.
*/
extern int fn_has_argblk;

int function_contains_argblk (MESSAGE_STACK messages, int fn_decl_idx) {

  int i, lookahead, fn_start_idx, fn_end_idx;
  int prev_idx = -1, prev_idx_2 = -1, prev_idx_3 = -1;

  if ((fn_start_idx = scanforward (messages, fn_decl_idx,
				   get_stack_top (messages),
				   OPENBLOCK)) == ERROR) {
    warning (messages[fn_decl_idx],
	     "Warning: Can't find function opening block.");
    fn_has_argblk = FALSE;
    return FALSE;
  }

  if ((fn_end_idx = find_function_close (messages, fn_decl_idx)) == ERROR) {
    warning (messages[fn_decl_idx],
	     "Warning: Can't find function close.");
    fn_has_argblk = FALSE;
    return FALSE;
  }

  for (i = fn_start_idx; i >= fn_end_idx; i--) {

    if (M_ISSPACE(messages[i]))
      continue;

    for (lookahead = i - 1; lookahead >= fn_end_idx; lookahead--) {

      if (M_ISSPACE(messages[lookahead]))
       	continue;

      if ((M_TOK(messages[i]) == LABEL) && 
	  (M_TOK(messages[lookahead]) == OPENBLOCK)) {

	/*	expressions like, "else {" */
	if (is_c_keyword (messages[i] -> name))
	  break;
	/* expressions like "struct {" or "struct <name> {" */
	if (prev_idx != -1) {
	  if (is_c_keyword (messages[prev_idx] -> name))
	    break;
	}
	if (prev_idx_3 != -1) {
	  if (is_c_keyword (messages[prev_idx_3] -> name))
	    break;
	}
	if (prev_idx_2 != -1) {
	  if (is_c_keyword (messages[prev_idx_2] -> name))
	    break;
	}
	if (prev_idx != -1) {
	  if (is_c_keyword (messages[prev_idx] -> name))
	    break;
	}

	fn_has_argblk = TRUE;
	return TRUE;
      } else {
	break;
      }

    }

    prev_idx_3 = prev_idx_2;
    prev_idx_2 = prev_idx;
    prev_idx = i;
  }

  fn_has_argblk = FALSE;
  return FALSE;
}
