/* $Id: errmsgs.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright © 2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"

void method_args_wrong_number_of_arguments_1 (MESSAGE_STACK messages,
					      int err_idx,
					      METHOD *new_method,
					      METHOD *obj_method,
					      char *fn_name,
					      I_PASS interpreter_pass) {
    switch (interpreter_pass)
      {
      case method_pass:
	warning (messages[err_idx],
		 "In method %s:",
		 new_method -> name);
	warning (messages[err_idx],
		 "Method %s: Wrong number of arguments.",
		 obj_method -> name);
	break;
      case parsing_pass:
	if (fn_name != NULL) {
	  warning (messages[err_idx],
		   "In function %s:", fn_name);
	  warning (messages[err_idx],
		   "Method %s: Wrong number of arguments.",
		   obj_method -> name);
	} else {
	  warning (messages[err_idx],
		   "Method %s: Wrong number of arguments.",
		   obj_method -> name);
	}
	break;
      default:
	warning (messages[err_idx],
		 "Method %s: Wrong number of arguments.",
		 obj_method -> name);
	break;
    } 
}

void method_args_ambiguous_argument_1 (MESSAGE_STACK messages,
				       int err_idx,
				       METHOD *new_method,
				       METHOD *obj_method,
				       int arglist_start_idx,
				       char *fn_name,
				       I_PASS interpreter_pass) {
    switch (interpreter_pass)
      {
      case method_pass:
	warning (messages[err_idx],
		 "In method %s:",
		 new_method -> name);
	warning (messages[err_idx],
 "Message, \"%s,\" follows method, \"%s,\" which returns \"Any\" class object.",
		 M_NAME(messages[arglist_start_idx]), obj_method -> name);
	break;
      case parsing_pass:
	if (fn_name != NULL) {
	  warning (messages[err_idx],
		   "In function %s:", fn_name);
	  warning (messages[err_idx],
   "Message, \"%s,\" follows method, \"%s,\" which returns \"Any\" class object.",
		   M_NAME(messages[arglist_start_idx]), obj_method -> name);
	} else {
	  warning (messages[err_idx],
   "Message, \"%s,\" follows method, \"%s,\" which returns \"Any\" class object.",
		   M_NAME(messages[arglist_start_idx]), obj_method -> name);
	}
	break;
      default:
	warning (messages[err_idx],
   "Message, \"%s,\" follows method, \"%s,\" which returns \"Any\" class object.",
		 M_NAME(messages[arglist_start_idx]), obj_method -> name);
	break;
    } 
}

