/* $Id: include.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright 
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#ifdef __DJGPP__
#include <dirent.h>
#endif
#include "object.h"
#include "message.h"
#include "ctalk.h"

extern int include_dir_opt;       /* Declared in main.c.       */
extern int n_user_include_dirs;        /* Specified with -I option. */
extern char *user_include_dirs[MAXUSERDIRS];
extern int systeminc_opt;         /* Specified with -s option. */
extern int n_system_inc_dirs;
extern char *systeminc_dirs[MAXUSERDIRS]; 

extern char *classlibdir;         /* Declared in rtinfo.c, read-only.      */
extern char *pkgname;
extern char libcachedir[FILENAME_MAX];
extern char ctalkuserhomedir[FILENAME_MAX];

extern char *library_include_paths[MAXUSERDIRS];   /* Declared in rtinfo.c. */

#ifdef __DJGPP__
extern int __opendir_flags;
#endif

static char **last_classlib_path_ptr;

char *find_library_include (char *fname, int find_next) {
  static char pathbuf[FILENAME_MAX];
  if (find_next) {
    ++last_classlib_path_ptr;
    while (*last_classlib_path_ptr) {
      strcpy (pathbuf, *last_classlib_path_ptr);
      STRCAT2(pathbuf, "/", fname);
      if (file_exists(pathbuf)) {
	return pathbuf;
      }
      ++last_classlib_path_ptr;
    }
    /* In case we get another find_next after an unsuccessful find_next. */
    last_classlib_path_ptr = library_include_paths;
   } else {
     last_classlib_path_ptr = library_include_paths;
     while (*last_classlib_path_ptr) {
       strcpy (pathbuf, *last_classlib_path_ptr);
       STRCAT2(pathbuf, "/", fname);
       if (file_exists(pathbuf)) {
	 return pathbuf;
       }
       ++last_classlib_path_ptr;
     }
   }
  return NULL;
}

void init_library_paths (void) {
  int i;
  char libdir[FILENAME_MAX],
    libenvdir[FILENAME_MAX],
    libenvdir2[FILENAME_MAX],
    libenvdir3[FILENAME_MAX],
    *libenv, *p, *q;
  int n_paths;

  for (i = 0, n_paths = 0; i < n_user_include_dirs; i++, n_paths++) {
    strcpy (libdir, expand_path(user_include_dirs[i]));
    library_include_paths[n_paths] = strdup (libdir);
    sprintf (libdir, "%s/%s", 
	     expand_path(user_include_dirs[i]), 
	     pkgname);
    if (is_dir (libdir)) {
      library_include_paths[n_paths++] = strdup (libdir);
    }
  }

  if ((libenv = getenv ("CLASSLIBDIRS")) != NULL) {
    p = libenv;
    do {
      q = strchr (p, ':');
      if (q) {
	substrcpy (libenvdir, p, 0, q - p);
	p = q + 1;
      } else {
	strcpy (libenvdir, p);
      }
      strcpy (libenvdir2, expand_path(libenvdir));
      library_include_paths[n_paths++] = strdup (libenvdir2);
      sprintf (libenvdir3, "%s/%s", libenvdir2, pkgname);
      if (is_dir (libenvdir3)) {
	library_include_paths[n_paths++] = strdup (libenvdir3);
      }
    } while (q != NULL);
  }

  library_include_paths[n_paths++] = strdup (classlibdir);

  sprintf (libdir, "%s/%s", classlibdir, pkgname);
  if (is_dir (libdir)) {
    library_include_paths[n_paths++] = strdup (libdir);
  }

}

/*
 *  init_user_home_path (), below must be called first.
 */
void init_lib_cache_path (void) {
  char buf[FILENAME_MAX];
#ifdef DJGPP
  strcpy (libcache_path, "/djgpp/include/ctalk/libcache");
  if (file_exists (libcachedir) &&
      !is_dir (libcachedir))
    unlink (libcachedir);
  strcpy (libcachedir, buf);
  if (!file_exists (libcachedir)) {
    mkdir (buf, 0700);
  }
#else
  sprintf (buf, "%s/%s/%s", getenv ("HOME"), ".ctalk", "libcache");
  strcpy (libcachedir, buf);
  if (file_exists (libcachedir) &&
      !is_dir (libcachedir))
    unlink (libcachedir);
  if (!is_dir (libcachedir)) {
    if (mkdir (buf, 0700))
      _error ("%s: %s.\n", libcachedir, strerror (errno));
  }
#endif
}

void init_user_home_path (void) {
  char buf[FILENAME_MAX];
#ifdef DJGPP
  strcpy (buf, "/djgpp/include/ctalk/libcache");
  strcpy (ctalkuserhomedir, buf);
  if (file_exists (ctalkuserhomedir) &&
      !is_dir (ctalkuserhomedir))
    unlink (ctalkuserhomedir);
  if (!file_exists (ctalkuserhomedir)) {
    mkdir (buf, 0700);
  }
#else
  sprintf (buf, "%s/%s", getenv ("HOME"), ".ctalk");
  strcpy (ctalkuserhomedir, buf);
  if (file_exists (ctalkuserhomedir) &&
      !is_dir (ctalkuserhomedir))
    unlink (ctalkuserhomedir);
  if (!is_dir (ctalkuserhomedir)) {
    if (mkdir (buf, 0700))
      _error ("%s: %s.\n", buf, strerror (errno));
  }
#endif
}

char *home_libcache_path (void) {
  return libcachedir;
}

char *ctalklib_path (void) {
  static char path[FILENAME_MAX];
  int i;

  sprintf (path, "%s/%s", classlibdir, "ctalklib");
  if (file_exists (path))
    return path;

  for (i = 0; library_include_paths[i]; i++) {
    sprintf (path, "%s/%s", library_include_paths[i], "ctalklib");
    if (file_exists (path))
      return path;
  }

  return NULL;
}

