/* $Id: return.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"

extern I_PASS interpreter_pass;         /* Declared in lib/rtinfo.c.  */
/* extern METHOD *new_methods[MAXARGS+1]; */ /* Declared in lib/rtnwmthd.c.*/
extern NEWMETHOD *new_methods[MAXARGS+1];  /* Declared in lib/rtnwmthd.c.*/
extern int new_method_ptr;
extern OBJECT *rcvr_class_obj;        /* Declared in lib/rtnwmthd.c and    */
                                      /* filled in by                      */
                                      /* new_instance|class_method ().     */

static char *basic_constant_return_class_fn (int tok) {
  METHOD *m;
  static char fn_name[MAXLABEL];
  m = new_methods[new_method_ptr + 1] -> method;
  *fn_name = '\0';
  if (*(m -> returnclass)) {
    if (!strcmp (m -> returnclass, LONGINTEGER_CLASSNAME)) {
      strcpy (fn_name, LLINT_CONSTANT_RETURN_FN);
    }
    if (!strcmp (m -> returnclass, BOOLEAN_CLASSNAME)) {
      strcpy (fn_name, BOOL_CONSTANT_RETURN_FN);
    }
    if (!strcmp (m -> returnclass, INTEGER_CLASSNAME)) {
      strcpy (fn_name, INT_CONSTANT_RETURN_FN);
    }
    if (!strcmp (m -> returnclass, CHARACTER_CLASSNAME)) {
      strcpy (fn_name, CHAR_CONSTANT_RETURN_FN);
    }
    if (!strcmp (m -> returnclass, STRING_CLASSNAME)) {
      strcpy (fn_name, CHAR_PTR_CONSTANT_RETURN_FN);
    }
    if (!strcmp (m -> returnclass, FLOAT_CLASSNAME)) {
      strcpy (fn_name, FLOAT_CONSTANT_RETURN_FN);
    }
  } else {
    if (!strcmp (rcvr_class_obj -> __o_name, LONGINTEGER_CLASSNAME)) {
      strcpy (fn_name, LLINT_CONSTANT_RETURN_FN);
    }
    if (!strcmp (rcvr_class_obj -> __o_name, BOOLEAN_CLASSNAME)) {
      strcpy (fn_name, BOOL_CONSTANT_RETURN_FN);
    }
    if (!strcmp (rcvr_class_obj -> __o_name, INTEGER_CLASSNAME)) {
      strcpy (fn_name, INT_CONSTANT_RETURN_FN);
    }
    if (!strcmp (rcvr_class_obj -> __o_name, CHARACTER_CLASSNAME)) {
      strcpy (fn_name, CHAR_CONSTANT_RETURN_FN);
    }
    if (!strcmp (rcvr_class_obj -> __o_name, STRING_CLASSNAME)) {
      strcpy (fn_name, CHAR_PTR_CONSTANT_RETURN_FN);
    }
    if (!strcmp (rcvr_class_obj -> __o_name, FLOAT_CLASSNAME)) {
      strcpy (fn_name, FLOAT_CONSTANT_RETURN_FN);
    }
  }
  if ((*fn_name == '\0') || 
      !strcmp (m -> returnclass, "Any")) {
    switch (tok)
      {
      case INTEGER:
	strcpy (fn_name, INT_CONSTANT_RETURN_FN);
	break;
      case LONG:
      case LONGLONG:
	strcpy (fn_name, LLINT_CONSTANT_RETURN_FN);
	break;
      case LITERAL:
	strcpy (fn_name, CHAR_PTR_CONSTANT_RETURN_FN);
	break;
      case LITERAL_CHAR:
	strcpy (fn_name, CHAR_CONSTANT_RETURN_FN);
	break;
      case DOUBLE:
	strcpy (fn_name, FLOAT_CONSTANT_RETURN_FN);
	break;
      default:
	_warning ("Warning: Return class of method %s (class %s) not supported, defaulting to Integer.\n", m -> name, rcvr_class_obj -> __o_name);
	strcpy (fn_name, INT_CONSTANT_RETURN_FN);
	break;
      }
  }
  if (! *fn_name) {
    _warning ("Warning: Return class of method %s (class %s) not supported, defaulting to Integer.\n", m -> name, rcvr_class_obj -> __o_name);
    strcpy (fn_name, INT_CONSTANT_RETURN_FN);
  }
  return fn_name;
}

static char *deref_terminal_return_fn (char *classname) {
  static char fn_name[MAXLABEL];
  *fn_name = '\0';
  if (!strcmp (classname, LONGINTEGER_CLASSNAME)) {
    strcpy (fn_name, LLINT_CONSTANT_RETURN_FN);
  }
  if (!strcmp (classname, BOOLEAN_CLASSNAME)) {
    strcpy (fn_name, BOOL_CONSTANT_RETURN_FN);
  }
  if (!strcmp (classname, INTEGER_CLASSNAME)) {
    strcpy (fn_name, INT_CONSTANT_RETURN_FN);
  }
  if (!strcmp (classname, CHARACTER_CLASSNAME)) {
    strcpy (fn_name, CHAR_CONSTANT_RETURN_FN);
  }
  if (!strcmp (classname, STRING_CLASSNAME)) {
    strcpy (fn_name, CHAR_PTR_CONSTANT_RETURN_FN);
  }
  if (!strcmp (classname, FLOAT_CLASSNAME)) {
    strcpy (fn_name, FLOAT_CONSTANT_RETURN_FN);
  }
  if (!strcmp (classname, OBJECT_CLASSNAME)) {
    strcpy (fn_name, OBJECT_MBR_RETURN_FN);
  }
  if (!*fn_name) {
    _warning ("Struct member return class %s not supported, defaulting to String\n", classname);
    strcpy (fn_name, CHAR_PTR_CONSTANT_RETURN_FN);
  }
  return fn_name;
}

/*
 *  Prevents generating code within the old methodReturn* 
 *  macros.
 */
static int backward_compatibility_insurer (MESSAGE_STACK messages,
					   int expr_end_idx) {
  if ((M_TOK(messages[expr_end_idx-1]) == SEMICOLON) &&
      (M_TOK(messages[expr_end_idx-2]) == SEMICOLON) &&
      (M_TOK(messages[expr_end_idx-3]) == SEMICOLON))
    return TRUE;
  return FALSE;
}
/*
 *  Handle "return self" or "return self value."
 */
static int simple_self_return_expr (MESSAGE_STACK messages,
				    int expr_start_idx,
				    int expr_end_idx) {
  int expr_end;
  if (expr_start_idx == expr_end_idx) {
    if (!strcmp (M_NAME(messages[expr_start_idx]), "self")) {
      if (interpreter_pass != expr_check) {
	char expr_buf[MAXMSG];

	/* this is faster at run time than calling __ctalkEvalExpr, 
	   which is what rt_self_expr () would generate. */

	sprintf (expr_buf, "%s ()", SELF_ACCESSOR_FN);
	fileout (expr_buf, 0, expr_start_idx);
	++messages[expr_start_idx] -> evaled;
	++messages[expr_start_idx] -> output;
	return TRUE;
      }
    }
  } else {
    if (!strcmp (M_NAME(messages[expr_start_idx]), "self") &&
	!strcmp (M_NAME(messages[expr_end_idx]), "value")) {
      if (interpreter_pass != expr_check) {
	(void)rt_self_expr (message_stack (), expr_start_idx, &expr_end);
	++messages[expr_start_idx] -> evaled;
	++messages[expr_end_idx] -> evaled;
	return TRUE;
      }
    }
  }
  return FALSE;
}

#define IS_SIGN(c) ((c == '-' || c == '+'))

int is_zero (char *buf) {
  int i, idx, bin_exp;

  while (*buf == '\'')
    TRIM_CHAR (buf);

  if (isdigit ((int)buf[0]) || IS_SIGN(buf[0])) {
    switch (radix_of (buf))
      {
      case hexadecimal:
	if (!sscanf (buf, "%x", &i))
	  sscanf (buf, "%X", &i);
	break;
      case octal:
	sscanf (buf, "%o", &i);
	break;
      case binary:
	for (idx = strlen(buf) - 1, bin_exp = 1, i = 0; idx >= 0; idx--) {
	  if ((buf[idx] == 'b') || (buf[idx] == 'B'))
	    continue;
	  i += (buf[idx] == '1') ? bin_exp: 0;
	  bin_exp *= 2;
	}
	break;
      case decimal:
      default:
	sscanf (buf, "%d", &i);
	break;
      }
  } else {
    if (!strcmp (buf, NULLSTR)) {
      i = 0;
    } else {
      if (*buf == '\0') {
	i = 0;
      } else {
	/* Character constant. */
	if (*buf == '\'') {
	  sscanf (buf, "'%c'", (char *)&i);
	} else {
	  _warning ("is_zero: Incompatible return type.\n");
	}
      }
    }
  } 
  return !i;
}

static int const_false_return_expr (MESSAGE_STACK messages,
				    int keyword_ptr,
				    int expr_start_idx,
				    int expr_end_idx) {
  char buf[MAXMSG];

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if ((M_TOK(messages[expr_start_idx]) == INTEGER) ||
	(M_TOK(messages[expr_start_idx]) == LONG) ||
	(M_TOK(messages[expr_start_idx]) == LONGLONG)) {
      if (is_zero (M_NAME(messages[expr_start_idx]))) {
	/*
	 *  Use the actual token as an Integer object in case what
	 *  the program wants is actually the integer.
	 */
	sprintf (buf, "%s(%d)", BOOL_CONSTANT_RETURN_FN, 0);
	fileout (buf, 0, expr_start_idx);
	++messages[expr_start_idx] -> evaled;
	++messages[expr_start_idx] -> output;
	return TRUE;
      }
    }
  }
    return FALSE;
}

/*
 *  Checks for non-zero constant argument as well as the 
 *  TRUE macro, which expands to "!(0)"
 */

static int const_true_return_expr (MESSAGE_STACK messages,
				    int keyword_ptr,
				    int expr_start_idx,
				    int expr_end_idx) {
  char buf[MAXMSG];
  int i;

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if ((M_TOK(messages[expr_start_idx]) == INTEGER) ||
	(M_TOK(messages[expr_start_idx]) == LONG) ||
	(M_TOK(messages[expr_start_idx]) == LONGLONG)) {
      if (!is_zero (M_NAME(messages[expr_start_idx]))) {
	/*
	 *  Use the actual token as an Integer object in case what
	 *  the program wants is actually the integer.
	 */
	sprintf (buf, "%s(%s)", 
		 basic_constant_return_class_fn (M_TOK(messages[expr_start_idx])),
		 M_NAME(messages[expr_start_idx]));
	fileout (buf, 0, expr_start_idx);
	++messages[expr_start_idx] -> evaled;
	++messages[expr_start_idx] -> output;
	return TRUE;
      }
    }
  } else {
    if ((M_TOK(messages[expr_start_idx]) == EXCLAM) &&
	(M_TOK(messages[expr_start_idx-1]) == OPENPAREN) &&
	((M_TOK(messages[expr_start_idx-2]) == INTEGER) &&
	 is_zero(M_NAME(messages[expr_start_idx-2]))) &&
	(M_TOK(messages[expr_start_idx-3]) == CLOSEPAREN)) {
      sprintf (buf, "%s(%d)", BOOL_CONSTANT_RETURN_FN, 1);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      }
      return TRUE;
    }
  }
  return FALSE;
}

static int char_ptr_const_return_expr (MESSAGE_STACK messages,
				       int keyword_ptr,
				       int expr_start_idx,
				       int expr_end_idx) {
  char buf[MAXMSG];

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if (M_TOK(messages[expr_start_idx]) == LITERAL) {
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (LITERAL), 
	       M_NAME(messages[expr_start_idx]));
      fileout (buf, 0, expr_start_idx);
      ++messages[expr_start_idx] -> evaled;
      ++messages[expr_start_idx] -> output;
      return TRUE;
    }
  }
  return FALSE;
}

static int char_const_return_expr (MESSAGE_STACK messages,
				       int keyword_ptr,
				       int expr_start_idx,
				       int expr_end_idx) {
  char buf[MAXMSG];

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if (M_TOK(messages[expr_start_idx]) == LITERAL_CHAR) {
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (LITERAL_CHAR),
	       M_NAME(messages[expr_start_idx]));
      fileout (buf, 0, expr_start_idx);
      ++messages[expr_start_idx] -> evaled;
      ++messages[expr_start_idx] -> output;
      return TRUE;
    }
  }
  return FALSE;
}

static int int_const_return_expr (MESSAGE_STACK messages,
				       int keyword_ptr,
				       int expr_start_idx,
				       int expr_end_idx) {
  char buf[MAXMSG];

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if (M_TOK(messages[expr_start_idx]) == INTEGER) {
      sprintf (buf, "%s(%s)", basic_constant_return_class_fn (INTEGER),
	       M_NAME(messages[expr_start_idx]));
      fileout (buf, 0, expr_start_idx);
      ++messages[expr_start_idx] -> evaled;
      ++messages[expr_start_idx] -> output;
      return TRUE;
    }
  }
  return FALSE;
}

static int llint_const_return_expr (MESSAGE_STACK messages,
				       int keyword_ptr,
				       int expr_start_idx,
				       int expr_end_idx) {
  char buf[MAXMSG];

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if ((M_TOK(messages[expr_start_idx]) == LONG) ||
	(M_TOK(messages[expr_start_idx]) == LONGLONG)) {
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (M_TOK(messages[expr_start_idx])),
	       M_NAME(messages[expr_start_idx]));
      fileout (buf, 0, expr_start_idx);
      ++messages[expr_start_idx] -> evaled;
      ++messages[expr_start_idx] -> output;
      return TRUE;
    }
  }
  return FALSE;
}

static int float_const_return_expr (MESSAGE_STACK messages,
				       int keyword_ptr,
				       int expr_start_idx,
				       int expr_end_idx) {
  char buf[MAXMSG];

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if ((M_TOK(messages[expr_start_idx]) == DOUBLE) ||
	(M_TOK(messages[expr_start_idx]) == FLOAT)) {
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (M_TOK(messages[expr_start_idx])),
	       M_NAME(messages[expr_start_idx]));
      fileout (buf, 0, expr_start_idx);
      ++messages[expr_start_idx] -> evaled;
      ++messages[expr_start_idx] -> output;
      return TRUE;
    }
  }
  return FALSE;
}

/*
 *  Checks for non-zero constant argument as well as the 
 *  TRUE macro, which expands to "!(0)"
 */

static int object_token_return_expr (MESSAGE_STACK messages,
				    int keyword_ptr,
				    int expr_start_idx,
				    int expr_end_idx) {
  char buf[MAXMSG];
  OBJECT *o;

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if (M_TOK(messages[expr_start_idx]) == LABEL) {
      if ((o = get_object (M_NAME(messages[expr_start_idx]),NULL)) != NULL){
	sprintf (buf, "%s(\"%s\", ((void *)0))",
		 OBJECT_TOK_RETURN_FN, 
		 M_NAME(messages[expr_start_idx]));
	fileout (buf, 0, expr_start_idx);
	++messages[expr_start_idx] -> evaled;
	++messages[expr_start_idx] -> output;
	return TRUE;
      }
    }
  }
  return FALSE;
}

/*
 *  Warn if a method tries to return an array without an 
 *  Array return class, unless the return class is String and 
 *  the array is declared as, char[<size>].
 */
static void array_class_return_warning (MESSAGE *orig, METHOD *m, CVAR *c) {

  if (!strcmp (c -> type, "char") && 
      (c -> n_derefs == 1)) {
    if ((*(m -> returnclass))  && 
	(!strcmp (m -> returnclass, STRING_CLASSNAME)))
      return;
    if (rcvr_class_obj && 
	!strcmp (rcvr_class_obj -> __o_classname, 
		 STRING_CLASSNAME))
      return;
  }

  if (c -> n_derefs > 1) {
    warning (orig,
	     "Method, \"%s,\" returns array, \"%s,\" which has more than one subscript.",
	     m -> name, c -> name);
  }

  if (*(m -> returnclass)) {
    if (strcmp (m -> returnclass, ARRAY_CLASSNAME)) {
      warning (orig,
	       "Method, \"%s,\" returns array, \"%s,\" while its return class is defined as \"%s.\"",
	       m -> name, c -> name, m -> returnclass);
    } else {
      if (rcvr_class_obj) {
	if (strcmp (rcvr_class_obj -> __o_name,
		    ARRAY_CLASSNAME)) {
	  warning (orig,
   "Method, \"%s,\" returns array, \"%s,\" while its return class is defined as, \"%s.\"",
		   m -> name, c -> name, rcvr_class_obj -> __o_name);
	}
      }
    }
  }
}

static int cvar_token_return_expr (MESSAGE_STACK messages,
				    int keyword_ptr,
				    int expr_start_idx,
				    int expr_end_idx) {
  char buf[MAXMSG], c_buf[MAXMSG], *s_ptr;
  CVAR *c;

  if (interpreter_pass != method_pass)
    return ERROR;

  if (expr_start_idx == expr_end_idx) {
    if (M_TOK(messages[expr_start_idx]) == LABEL) {

      if (backward_compatibility_insurer (messages, expr_end_idx))
	return FALSE;

      if (((c = get_local_var (M_NAME(messages[expr_start_idx]))) != NULL)||
	  ((c = get_global_var(M_NAME(messages[expr_start_idx]))) != NULL)){
	strcpy (c_buf, 
		fmt_register_c_method_arg_call 
		(c, M_NAME(messages[expr_end_idx]), LOCAL_VAR));
	/* 
	 * The fmt_register_c_method_arg_call adds a semicolon at the
	 * end of the function.  Remove it in this case.
	 */
	if ((s_ptr = strchr (c_buf, ';')) != NULL)
	  *s_ptr = '\0';
  	if (c -> attrs & CVAR_ATTR_ARRAY_DECL) {
	  array_class_return_warning 
	    (messages[expr_start_idx], 
	     new_methods[new_method_ptr + 1] -> method,
	     c); 
  	  sprintf (buf, "(%s, %s(\"%s\", %d, &%s))",
  		   c_buf, CVAR_ARRAY_TOK_RETURN_FN,
  		   M_NAME(messages[expr_start_idx]),
 		   c -> initializer_size,
 		   M_NAME(messages[expr_start_idx]));
  	} else {  /* if (c -> attrs & CVAR_ATTR_ARRAY_DECL) */
	  sprintf (buf, "(%s, %s(\"%s\"))", 
		   c_buf, CVAR_TOK_RETURN_FN, 
		   M_NAME(messages[expr_start_idx]));
  	}
	fileout (buf, 0, expr_start_idx);
	++messages[expr_start_idx] -> evaled;
	++messages[expr_start_idx] -> output;
	return TRUE;
      }
    }
  }
  return FALSE;
}

/*
 *  Expressions of the form: 
 *    <cvar|fn> -> mbr [ -> mbr]*
 *    <cvar|fn> . mbr [ . mbr ]*
 *    ... etc.
 *
 *  Expressions can start with type casts, but they are handled 
 *  below.
 */

static int deref_expr_return_expr (MESSAGE_STACK messages,
				   int keyword_ptr,
				   int expr_start_idx,
				   int expr_end_idx) {
  int last_op_token_idx;
  CVAR *label_cvar, *type_cvar = NULL, /* Avoid a warning */
    *member_cvar;
  CFUNC *label_cfunc;
  char *return_expr, return_classname[MAXLABEL], buf[MAXMSG];
  int i;

  if (is_struct_or_union_expr (messages, 
			       expr_start_idx,
			       stack_start (messages),
			       get_stack_top (messages))) {
    if (M_TOK(messages[expr_end_idx]) == LABEL) {
      if ((last_op_token_idx=prevlangmsg (messages, expr_end_idx)) != ERROR){
	if ((M_TOK(messages[last_op_token_idx]) == DEREF) ||
	    (M_TOK(messages[last_op_token_idx]) == PERIOD)) {
	  if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	      != NULL) {
	    if (((label_cvar = get_local_var (label_cfunc -> return_type))
		 == NULL) &&
		((label_cvar = get_global_var (label_cfunc -> return_type))
		 == NULL)) {
	      if ((type_cvar = get_typedef (label_cfunc -> return_type)) 
		  == NULL) {
		warning (messages[keyword_ptr], 
			 "Unsupported return type %s.",
			 label_cfunc -> return_type);
		return FALSE;
	      }
	    }
	  } else {/*if ((label_cfunc = get_function ... */
	    if (((label_cvar = 
		  get_local_var (M_NAME(messages[expr_start_idx])))
		 == NULL) &&
		((label_cvar = 
		  get_global_var (M_NAME(messages[expr_start_idx])))
		 == NULL)) {
	      if ((type_cvar = get_typedef (M_NAME(messages[expr_start_idx])))
		  == NULL) {
		warning (messages[keyword_ptr], 
			 "Unsupported return type %s.",
			 M_NAME(messages[expr_start_idx]));
		return FALSE;
	      }
	    } else {
	      type_cvar = get_typedef (label_cvar -> type);
	    }
	  }
	  return_expr = collect_tokens (messages, expr_start_idx,
					expr_end_idx);
	  if ((member_cvar = 
	       struct_member_from_expr (return_expr, type_cvar)) == NULL) {
	    warning (messages[expr_start_idx], 
		     "Could not find member definition for expression $s.",
		     return_expr);
	    free (return_expr);
	    return FALSE;
	  }
	  strcpy (return_classname, 
		  basic_class_from_cvar 
		  (messages[expr_start_idx], member_cvar, 0));
	  sprintf (buf, "%s(%s)", 
		   deref_terminal_return_fn (return_classname),
		   return_expr);
	  free (return_expr);
	  fileout (buf, 0, expr_start_idx);
	  for (i = expr_start_idx; i >= expr_end_idx; i--) {
	    ++messages[i] -> evaled;
	    ++messages[i] -> output;
	  } 
	  return TRUE;
	} /* if ((M_TOK(messages[last_op_token_idx]) == DEREF) ... */
      }
    }
  }
  return FALSE;
}

static int subscript_expr_return_expr (MESSAGE_STACK messages,
				       int keyword_ptr,
				       int expr_start_idx,
				       int expr_end_idx) {
  int last_op_token_idx;
  CVAR *label_cvar;
  CFUNC *label_cfunc;
  char *return_expr, return_classname[MAXLABEL], buf[MAXMSG];
  int i;

  if ((last_op_token_idx = 
       is_subscript_expr (messages, 
			  expr_start_idx,
			  get_stack_top (messages))) > 0) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (((label_cvar = get_local_var (label_cfunc -> return_type))
	   == NULL) &&
	  ((label_cvar = get_global_var (label_cfunc -> return_type))
	   == NULL)) {
	warning (messages[keyword_ptr],
		 "Unsupported type %s for %s.",
		 label_cvar -> type,
		 M_NAME(messages[expr_start_idx]));
	return FALSE;
      }
    } else {/*if ((label_cfunc = get_function ... */
      if (((label_cvar = 
	    get_local_var (M_NAME(messages[expr_start_idx])))
	   == NULL) &&
	  ((label_cvar = 
	    get_global_var (M_NAME(messages[expr_start_idx])))
	   == NULL)) {
	warning (messages[keyword_ptr],
		 "Unsupported type %s for %s.",
		 label_cvar -> type,
		 M_NAME(messages[expr_start_idx]));
      }
    }
    return_expr = collect_tokens (messages, expr_start_idx,
				  expr_end_idx);
    strcpy (return_classname, 
	    basic_class_from_cvar 
	    (messages[expr_start_idx], label_cvar, 1));
    sprintf (buf, "%s(%s)", 
	     deref_terminal_return_fn (return_classname),
	     return_expr);
    free (return_expr);
    fileout (buf, 0, expr_start_idx);
    for (i = expr_start_idx; i >= expr_end_idx; i--) {
      ++messages[i] -> evaled;
      ++messages[i] -> output;
    } 
    return TRUE;
  }
  return FALSE;
}

/*
 *  C-only expressions that return int.  Expressions that 
 *  contain objects get evaluated separately.
 */
static int int_expr_return_expr (MESSAGE_STACK messages,
				 int keyword_ptr,
				 int expr_start_idx,
				 int expr_end_idx) {
  char *s, buf[MAXLABEL];
  CVAR *c;
  CFUNC *fn = NULL;
  int i, after_prefix_tok_idx, typecast_end_idx = -1;

  if (interpreter_pass != method_pass)
    return ERROR;

  /*
   *  Return if the expression refers to any objects.
   */
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    if (M_TOK(messages[i]) == LABEL) {
      if (get_object (M_NAME(messages[i]), NULL) != NULL)
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "self"))
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "super"))
	return FALSE;
    }
  }

  /*
   *  Format expressions with leading type casts immediately.
   *
   */
  if (is_typecast_expr (messages, expr_start_idx, &typecast_end_idx)) {
    if (strcmp (basic_class_from_typecast (messages, expr_start_idx,
					   typecast_end_idx), 
		INTEGER_CLASSNAME)) {
      return FALSE;
    } else {
      s = collect_tokens (messages, expr_start_idx, expr_end_idx);
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (INTEGER), s);
      free (s);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      } 
      return TRUE;
    }
  }

  if (IS_C_UNARY_MATH_OP(M_TOK(messages[expr_start_idx]))) {
    after_prefix_tok_idx = nextlangmsg (messages, expr_start_idx);
  } else {
    after_prefix_tok_idx = expr_start_idx;
  }

  if (((c = get_local_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL) ||
      ((c = get_global_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL)) {
    if (strcmp (c -> type, "int"))
      return FALSE;
    if (*(c -> qualifier) || *(c -> qualifier2) || *(c -> qualifier3)
	|| (c -> qualifier4))
      return FALSE;
    if (c -> n_derefs) return FALSE;
  } else {
    if ((fn = get_function (M_NAME(messages[after_prefix_tok_idx]))) != NULL) {
      if (strcmp (fn -> return_type, "int")) 
	return FALSE;
      if (*(fn -> qualifier_type) || *(fn -> qualifier2_type))
	return FALSE;
      if (fn -> return_derefs)
	return FALSE;
    }
  }
  if (!c && !fn)
    return FALSE;

  s = collect_tokens (messages, expr_start_idx, expr_end_idx);
  sprintf (buf, "%s(%s)", 
	   basic_constant_return_class_fn (INTEGER), s);
  free (s);
  fileout (buf, 0, expr_start_idx);
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  } 
  return TRUE;
}

/*
 *  C-only expressions that return <long long> | <long> int.  Expressions that 
 *  contain objects get evaluated separately.
 */
static int long_int_expr_return_expr (MESSAGE_STACK messages,
				 int keyword_ptr,
				 int expr_start_idx,
				 int expr_end_idx) {
  char *s, buf[MAXLABEL];
  CVAR *c;
  CFUNC *fn = NULL;
  int i, after_prefix_tok_idx, typecast_end_idx = -1;

  if (interpreter_pass != method_pass)
    return ERROR;

  /*
   *  Return if the expression refers to any objects.
   */
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    if (M_TOK(messages[i]) == LABEL) {
      if (get_object (M_NAME(messages[i]), NULL) != NULL)
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "self"))
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "super"))
	return FALSE;
    }
  }

  /*
   *  Format expressions with leading type casts immediately.
   *
   */
  if (is_typecast_expr (messages, expr_start_idx, &typecast_end_idx)) {
    if (strcmp (basic_class_from_typecast (messages, expr_start_idx,
					   typecast_end_idx), 
		LONGINTEGER_CLASSNAME)) {
      return FALSE;
    } else {
      s = collect_tokens (messages, expr_start_idx, expr_end_idx);
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (LONGLONG), s);
      free (s);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      } 
      return TRUE;
    }
  }

  if (IS_C_UNARY_MATH_OP(M_TOK(messages[expr_start_idx]))) {
    after_prefix_tok_idx = nextlangmsg (messages, expr_start_idx);
  } else {
    after_prefix_tok_idx = expr_start_idx;
  }

  if (((c = get_local_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL) ||
      ((c = get_global_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL)) {
    if (strcmp (c -> type, "int"))
      return FALSE;
    if (*(c -> qualifier)) {
      if (strcmp (c -> qualifier, "long")) {
	return FALSE;
      }
    } else {
      return FALSE;
    }
    if (c -> n_derefs) return FALSE;
  } else {
    if ((fn = get_function (M_NAME(messages[after_prefix_tok_idx]))) != NULL) {
      if (strcmp (fn -> return_type, "int")) 
	return FALSE;
      if (*(fn -> qualifier_type)) {
	if (strcmp (fn -> qualifier_type, "long")) {
	  return FALSE;
	} 
      } else {
	return FALSE;
      }
      if (fn -> return_derefs)
	return FALSE;
    }
  }
  if (!c && !fn)
    return FALSE;

  s = collect_tokens (messages, expr_start_idx, expr_end_idx);
  sprintf (buf, "%s(%s)", 
	   basic_constant_return_class_fn (LONGLONG), s);
  free (s);
  fileout (buf, 0, expr_start_idx);
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  } 
  return TRUE;
}

/*
 *  C-only expressions that return char.  Expressions that 
 *  contain objects get evaluated separately.
 */
static int char_expr_return_expr (MESSAGE_STACK messages,
				 int keyword_ptr,
				 int expr_start_idx,
				 int expr_end_idx) {
  char *s, buf[MAXLABEL];
  CVAR *c;
  CFUNC *fn = NULL;
  int i, after_prefix_tok_idx, typecast_end_idx = -1;

  if (interpreter_pass != method_pass)
    return ERROR;

  /*
   *  Return if the expression refers to any objects.
   */
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    if (M_TOK(messages[i]) == LABEL) {
      if (get_object (M_NAME(messages[i]), NULL) != NULL)
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "self"))
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "super"))
	return FALSE;
    }
  }

  /*
   *  Format expressions with leading type casts immediately.
   *
   */
  if (is_typecast_expr (messages, expr_start_idx, &typecast_end_idx)) {
    if (strcmp (basic_class_from_typecast (messages, expr_start_idx,
					   typecast_end_idx), 
		CHARACTER_CLASSNAME)) {
      return FALSE;
    } else {
      s = collect_tokens (messages, expr_start_idx, expr_end_idx);
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (LITERAL_CHAR), s);
      free (s);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      } 
      return TRUE;
    }
  }

  if (IS_C_UNARY_MATH_OP(M_TOK(messages[expr_start_idx]))) {
    after_prefix_tok_idx = nextlangmsg (messages, expr_start_idx);
  } else {
    after_prefix_tok_idx = expr_start_idx;
  }

  if (((c = get_local_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL) ||
      ((c = get_global_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL)) {
    if (strcmp (c -> type, "char")) 
      return FALSE;
    if (*(c -> qualifier))
      return FALSE;
    if (c -> n_derefs) 
      return FALSE;
  } else {
    if ((fn = get_function (M_NAME(messages[after_prefix_tok_idx]))) != NULL) {
      if (strcmp (fn -> return_type, "char")) 
	return FALSE;
      if (*(fn -> qualifier_type))
	return FALSE;
      if (fn -> return_derefs)
	return FALSE;
    }
  }
  if (!c && !fn)
    return FALSE;

  s = collect_tokens (messages, expr_start_idx, expr_end_idx);
  sprintf (buf, "%s(%s)", 
	   basic_constant_return_class_fn (LITERAL_CHAR), s);
  free (s);
  fileout (buf, 0, expr_start_idx);
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  } 
  return TRUE;
}

/*
 *  C-only expressions that return char *.  Expressions that 
 *  contain objects get evaluated separately.
 */
static int char_ptr_expr_return_expr (MESSAGE_STACK messages,
				 int keyword_ptr,
				 int expr_start_idx,
				 int expr_end_idx) {
  char *s, buf[MAXLABEL];
  CVAR *c;
  CFUNC *fn = NULL;
  int i, after_prefix_tok_idx, typecast_end_idx = -1;

  if (interpreter_pass != method_pass)
    return ERROR;

  /*
   *  Return if the expression refers to any objects.
   */
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    if (M_TOK(messages[i]) == LABEL) {
      if (get_object (M_NAME(messages[i]), NULL) != NULL)
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "self"))
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "super"))
	return FALSE;
    }
  }

  /*
   *  Format expressions with leading type casts immediately.
   *
   */
  if (is_typecast_expr (messages, expr_start_idx, &typecast_end_idx)) {
    if (strcmp (basic_class_from_typecast (messages, expr_start_idx,
					   typecast_end_idx), 
		STRING_CLASSNAME)) {
      return FALSE;
    } else {
      s = collect_tokens (messages, expr_start_idx, expr_end_idx);
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (LITERAL), s);
      free (s);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      } 
      return TRUE;
    }
  }

  if (IS_C_UNARY_MATH_OP(M_TOK(messages[expr_start_idx]))) {
    after_prefix_tok_idx = nextlangmsg (messages, expr_start_idx);
  } else {
    after_prefix_tok_idx = expr_start_idx;
  }

  if (((c = get_local_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL) ||
      ((c = get_global_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL)) {
    if (strcmp (c -> type, "char")) 
      return FALSE;
    if (*(c -> qualifier))
      return FALSE;
    if (c -> n_derefs != 1) 
      return FALSE;
  } else {
    if ((fn = get_function (M_NAME(messages[after_prefix_tok_idx]))) != NULL) {
      if (strcmp (fn -> return_type, "char")) 
	return FALSE;
      if (*(fn -> qualifier_type))
	return FALSE;
      if (fn -> return_derefs != 1)
	return FALSE;
    }
  }
  if (!c && !fn)
    return FALSE;

  s = collect_tokens (messages, expr_start_idx, expr_end_idx);
  sprintf (buf, "%s(%s)", 
	   basic_constant_return_class_fn (LITERAL), s);
  free (s);
  fileout (buf, 0, expr_start_idx);
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  } 
  return TRUE;
}

/*
 *  C-only expressions that return float or double.  Expressions that 
 *  contain objects get evaluated separately.
 */
static int float_expr_return_expr (MESSAGE_STACK messages,
				 int keyword_ptr,
				 int expr_start_idx,
				 int expr_end_idx) {
  char *s, buf[MAXLABEL];
  CVAR *c;
  CFUNC *fn = NULL;
  int i, after_prefix_tok_idx, typecast_end_idx = -1;

  if (interpreter_pass != method_pass)
    return ERROR;

  /*
   *  Return if the expression refers to any objects.
   */
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    if (M_TOK(messages[i]) == LABEL) {
      if (get_object (M_NAME(messages[i]), NULL) != NULL)
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "self"))
	return FALSE;
      if (!strcmp (M_NAME(messages[i]), "super"))
	return FALSE;
    }
  }

  /*
   *  Format expressions with leading type casts immediately.
   *
   */
  if (is_typecast_expr (messages, expr_start_idx, &typecast_end_idx)) {
    if (strcmp (basic_class_from_typecast (messages, expr_start_idx,
					   typecast_end_idx), 
		FLOAT_CLASSNAME)) {
      return FALSE;
    } else {
      s = collect_tokens (messages, expr_start_idx, expr_end_idx);
      sprintf (buf, "%s(%s)", 
	       basic_constant_return_class_fn (DOUBLE), s);
      free (s);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      } 
      return TRUE;
    }
  }

  if (IS_C_UNARY_MATH_OP(M_TOK(messages[expr_start_idx]))) {
    after_prefix_tok_idx = nextlangmsg (messages, expr_start_idx);
  } else {
    after_prefix_tok_idx = expr_start_idx;
  }

  if (((c = get_local_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL) ||
      ((c = get_global_var (M_NAME(messages[after_prefix_tok_idx]))) != NULL)) {
    if (strcmp (c -> type, "char") && strcmp (c -> type, "double")) 
	return FALSE;
    if (*(c -> qualifier))
      return FALSE;
    if (c -> n_derefs) 
      return FALSE;
  } else {
    if ((fn = get_function (M_NAME(messages[after_prefix_tok_idx]))) != NULL) {
      if (strcmp (fn -> return_type, "float") && strcmp (fn -> return_type, "double"))
	return FALSE;
      if (*(fn -> qualifier_type))
	return FALSE;
      if (fn -> return_derefs)
	return FALSE;
    }
  }
  if (!c && !fn)
    return FALSE;

  s = collect_tokens (messages, expr_start_idx, expr_end_idx);
  sprintf (buf, "%s(%s)", 
	   basic_constant_return_class_fn (DOUBLE), s);
  free (s);
  fileout (buf, 0, expr_start_idx);
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  } 
  return TRUE;
}

static int obj_fn_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  CFUNC *label_cfunc;
  CVAR *arg_cvar;
  char buf[MAXMSG], param_buf[MAXMSG], *s, s_expr[MAXMSG];
  int args[MAXARGS], n_args, n_th_arg, i, open_paren_idx, close_paren_idx,
    have_object;
  if (M_TOK(messages[expr_start_idx]) == LABEL) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (!strcmp (label_cfunc -> return_type, "OBJECT") && 
	  (label_cfunc -> return_derefs == 1)) {
	sprintf (buf, "%s (", M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
    }
  }
  return FALSE;
}

static int int_fn_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  CFUNC *label_cfunc;
  CVAR *arg_cvar;
  char buf[MAXMSG], param_buf[MAXMSG], *s, s_expr[MAXMSG];
  int args[MAXARGS], n_args, n_th_arg, i, open_paren_idx, close_paren_idx,
    have_object;
  if (M_TOK(messages[expr_start_idx]) == LABEL) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (!strcmp (label_cfunc -> return_type, "int") && 
	  (*(label_cfunc -> qualifier_type) == '\0') &&
	  (*(label_cfunc -> qualifier2_type) == '\0') &&
	  (label_cfunc -> return_derefs == 0)) {
	sprintf (buf, "%s ( %s (", INT_CONSTANT_RETURN_FN,
		 M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	strcat (buf, ")");
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
      if (!strcmp (label_cfunc -> return_type, "int") && 
	  !strcmp (label_cfunc -> qualifier_type, "long") &&
	  (*(label_cfunc -> qualifier2_type) == '\0') &&
	  (label_cfunc -> return_derefs == 0)) {
	sprintf (buf, "%s ( %s(", INT_CONSTANT_RETURN_FN,
		 M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	strcat (buf, ")");
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
    }
  }
  return FALSE;
}

static int long_long_int_fn_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  CFUNC *label_cfunc;
  CVAR *arg_cvar;
  char buf[MAXMSG], param_buf[MAXMSG], *s, s_expr[MAXMSG];
  int args[MAXARGS], n_args, n_th_arg, i, open_paren_idx, close_paren_idx,
    have_object;
  if (M_TOK(messages[expr_start_idx]) == LABEL) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (!strcmp (label_cfunc -> return_type, "int") && 
	  !strcmp (label_cfunc -> qualifier_type, "long") &&
	  !strcmp (label_cfunc -> qualifier_type, "long") &&
	  (label_cfunc -> return_derefs == 0)) {
	sprintf (buf, "%s ( %s(", LLINT_CONSTANT_RETURN_FN,
		 M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	strcat (buf, ")");
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
    }
  }
  return FALSE;
}

static int char_fn_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  CFUNC *label_cfunc;
  CVAR *arg_cvar;
  char buf[MAXMSG], param_buf[MAXMSG], *s, s_expr[MAXMSG];
  int args[MAXARGS], n_args, n_th_arg, i, open_paren_idx, close_paren_idx,
    have_object;
  if (M_TOK(messages[expr_start_idx]) == LABEL) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (!strcmp (label_cfunc -> return_type, "char") &&
	  (label_cfunc -> return_derefs == 0)) {
	sprintf (buf, "%s ( %s(", CHAR_CONSTANT_RETURN_FN,
		 M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	strcat (buf, ")");
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
    }
  }
  return FALSE;
}

static int char_ptr_fn_return_expr (MESSAGE_STACK messages,
				    int keyword_ptr,
				    int expr_start_idx,
				    int expr_end_idx) {
  CFUNC *label_cfunc;
  CVAR *arg_cvar;
  char buf[MAXMSG], param_buf[MAXMSG], *s, s_expr[MAXMSG];
  int args[MAXARGS], n_args, n_th_arg, i, open_paren_idx, close_paren_idx,
    have_object;
  if (M_TOK(messages[expr_start_idx]) == LABEL) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (!strcmp (label_cfunc -> return_type, "char") &&
	  (label_cfunc -> return_derefs == 1)) {
	sprintf (buf, "%s ( %s(", CHAR_PTR_CONSTANT_RETURN_FN,
		 M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	strcat (buf, ")");
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
    }
  }
  return FALSE;
}

static int double_fn_return_expr (MESSAGE_STACK messages,
				    int keyword_ptr,
				    int expr_start_idx,
				    int expr_end_idx) {
  CFUNC *label_cfunc;
  CVAR *arg_cvar;
  char buf[MAXMSG], param_buf[MAXMSG], *s, s_expr[MAXMSG];
  int args[MAXARGS], n_args, n_th_arg, i, open_paren_idx, close_paren_idx,
    have_object;
  if (M_TOK(messages[expr_start_idx]) == LABEL) {
    if ((label_cfunc = get_function (M_NAME(messages[expr_start_idx])))
	!= NULL) {
      if (!strcmp (label_cfunc -> return_type, "double") &&
	  (label_cfunc -> return_derefs == 0)) {
	sprintf (buf, "%s ( %s(", FLOAT_CONSTANT_RETURN_FN,
		 M_NAME(messages[expr_start_idx]));
	open_paren_idx = nextlangmsg (messages, expr_start_idx);
	close_paren_idx = match_paren (messages, open_paren_idx,
				       get_stack_top (messages));
	split_args_idx (messages, open_paren_idx, close_paren_idx,
			args, &n_args);
	arg_cvar = label_cfunc -> params;
	for (n_th_arg = 0; n_th_arg < n_args; n_th_arg += 2) {
	  have_object = FALSE;
	  for (i = args[n_th_arg]; i >= args[n_th_arg+1]; i--) {
	    if (get_object (M_NAME(messages[i]), NULL))
	      have_object = TRUE;
	    if (!strcmp (M_NAME(messages[i]), "self"))
	      have_object = TRUE;
	  }
	  s = collect_tokens (messages, args[n_th_arg], args[n_th_arg+1]);
	  sprintf (s_expr, "%s (\"%s\")", EVAL_EXPR_FN, s);
	  if (have_object) {
	    if (arg_cvar) {
	      strcpy (param_buf, 
		       fmt_rt_return 
		       (s_expr, 
			basic_class_from_cvar 
			(messages[expr_start_idx], arg_cvar, 0),
			TRUE));
	    } else {
	      strcpy (param_buf, s);
	    }
	  } else {
	    strcpy (param_buf, s);
	  }
	  free (s);
	  strcat (buf, param_buf);
	  if (n_th_arg == (n_args - 2)) {
	    strcat (buf, ")");
	  } else {
	    strcat (buf, ",");
	  }
	}
	strcat (buf, ")");
	fileout (buf, 0, expr_start_idx);
	for (i = expr_start_idx; i >= expr_end_idx; i--) {
	  ++messages[i] -> evaled;
	  ++messages[i] -> output;
	} 
	return TRUE;
      }
    }
  }
  return FALSE;
}

/*
 *  A return NULL is also a no-op here.
 */
static int NULL_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  int n_parens, i;
  n_parens = 0;
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    switch (M_TOK(messages[i]))
      {
      case OPENPAREN:
	++n_parens;
	break;
      case CLOSEPAREN:
	--n_parens;
	break;
      case LABEL:
	if (strcmp (M_NAME(messages[i]), "void") ||
	    (n_parens != 2))
	  return FALSE;
	break;
      case ASTERISK:
	if (n_parens != 2)
	  return FALSE;
	break;
      case INTEGER:
	if (strcmp (M_NAME(messages[i]), "0") ||
	    (n_parens != 1))
	  return FALSE;
	break;
      }
  }
  return TRUE;
}

static int eval_expr_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  char *s, buf[MAXMSG];
  int i;
  if (backward_compatibility_insurer (messages, expr_end_idx))
    return FALSE;
  s = collect_tokens (messages, expr_start_idx, expr_end_idx);
  sprintf (buf, "%s(\"%s\")", EVAL_EXPR_FN, s);
  free (s);
  fileout (buf, 0, expr_start_idx);
  for (i = expr_start_idx; i >= expr_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  } 
  return TRUE;
}

static int eval_stmt_return_expr (MESSAGE_STACK messages,
				  int keyword_ptr,
				  int expr_start_idx,
				  int expr_end_idx) {
  char *s, buf[MAXMSG];
  int i, next_tok_idx;
  if(!strcmp (M_NAME(messages[expr_start_idx]), "eval")) {
    if ((next_tok_idx = nextlangmsg (messages, expr_start_idx)) != ERROR) {
      s = collect_tokens (messages, expr_start_idx, expr_end_idx);
      sprintf (buf, "%s(\"%s\")", EVAL_EXPR_FN, s);
      free (s);
      fileout (buf, 0, expr_start_idx);
      for (i = expr_start_idx; i >= expr_end_idx; i--) {
	++messages[i] -> evaled;
	++messages[i] -> output;
      } 
      return TRUE;
    }
  }
  return FALSE;
}

int eval_return_expr (MESSAGE_STACK messages, int keyword_ptr) {
    int expr_start_idx,
    expr_end_idx,
    i,
    stack_end,
    typecast_end_idx;

  stack_end = get_stack_top (messages);
  if ((expr_start_idx = nextlangmsg (messages, keyword_ptr)) 
      == ERROR)
    _error ("eval_return_expr: parser error.\n");

  if ((M_TOK(messages[expr_start_idx]) == OPENPAREN) &&
      !is_typecast_expr (messages, expr_start_idx, &typecast_end_idx)) {
    if ((expr_end_idx = match_paren (messages, expr_start_idx,
				     get_stack_top (messages))) 
	== ERROR) {
      warning (messages[expr_start_idx], "Mismatched parentheses.");
      return ERROR;
    }
    expr_start_idx = nextlangmsg (messages, expr_start_idx);
    expr_end_idx = prevlangmsg (messages, expr_end_idx);
  } else {
    for (i = expr_start_idx; i > stack_end; i--)
      if (M_TOK(messages[i]) == SEMICOLON)
	break;
    expr_end_idx = prevlangmsg (messages, i);
  }

  if (eval_stmt_return_expr (messages, keyword_ptr,
			     expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (simple_self_return_expr (messages, expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (int_const_return_expr (messages, keyword_ptr,
				  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (llint_const_return_expr (messages, keyword_ptr,
				  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (char_ptr_const_return_expr (messages, keyword_ptr,
				  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (char_const_return_expr (messages, keyword_ptr,
				  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (float_const_return_expr (messages, keyword_ptr,
				  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (const_false_return_expr (messages, keyword_ptr,
 			       expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (const_true_return_expr (messages, keyword_ptr,
 			       expr_start_idx, expr_end_idx))
    return SUCCESS;

  if (object_token_return_expr (messages, keyword_ptr,
				expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (cvar_token_return_expr (messages, keyword_ptr,
  			      expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (deref_expr_return_expr (messages, keyword_ptr,
 			      expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (subscript_expr_return_expr (messages, keyword_ptr,
				  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (int_expr_return_expr (messages, keyword_ptr,
			    expr_start_idx, expr_end_idx))
    return SUCCESS;

  if (long_int_expr_return_expr (messages, keyword_ptr,
			    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (char_expr_return_expr (messages, keyword_ptr,
			    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (char_ptr_expr_return_expr (messages, keyword_ptr,
			    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (float_expr_return_expr (messages, keyword_ptr,
			    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (obj_fn_return_expr (messages, keyword_ptr,
			  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (int_fn_return_expr (messages, keyword_ptr,
 			  expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (long_long_int_fn_return_expr (messages, keyword_ptr,
				    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (char_fn_return_expr (messages, keyword_ptr,
				    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (char_ptr_fn_return_expr (messages, keyword_ptr,
			       expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (double_fn_return_expr (messages, keyword_ptr,
				    expr_start_idx, expr_end_idx))
    return SUCCESS;
  if (NULL_return_expr (messages, keyword_ptr,
			expr_start_idx,expr_end_idx))
    return SUCCESS;
  if (eval_expr_return_expr (messages, keyword_ptr,
 			     expr_start_idx, expr_end_idx))
    return SUCCESS;
  return SUCCESS;
}
