/* $Id: rt_expr.c,v 1.6 2014/05/05 10:52:02 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2014 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "objtoc.h"

extern I_PASS interpreter_pass;    /* Declared in main.c.               */

extern PARSER *parsers[MAXARGS+1];           /* Declared in rtinfo.c. */
extern int current_parser_ptr;
extern OBJECT *rcvr_class_obj;        /* Declared in lib/rtnwmthd.c. */

/*
 *  Simple or complex expression that should be evaled at run time.
 */
char *fmt_rt_expr (MESSAGE_STACK messages, int start_idx, int *end_idx) {
  static char buf[MAXMSG];
  char *exprbuf, exprbuf_esc[MAXMSG];
  int idx;

  exprbuf = collect_expression (messages, start_idx, end_idx, 
				stack_start (messages),
				get_stack_top (messages));
  de_newline_buf (exprbuf);
  for (idx = start_idx; idx >= *end_idx; idx--)
    messages[idx]->attrs |= TOK_IS_RT_EXPR;
  strcpy (exprbuf_esc, escape_str_quotes (exprbuf));
  free (exprbuf);
  sprintf (buf, "%s(\"%s\")", EVAL_EXPR_FN, exprbuf_esc);
  return buf;
}

/* extern METHOD *new_methods[MAXARGS+1];*/  /* Declared in lib/rtnwmthd.c. */
extern NEWMETHOD *new_methods[MAXARGS+1];  /* Declared in lib/rtnwmthd.c. */
extern int new_method_ptr;

static int is_method_param_name (char *s) {
  METHOD *m;
  int n_th_param;
  if (interpreter_pass != method_pass)
    return FALSE;

  if ((m = new_methods[new_method_ptr+1] -> method) != NULL) {
    for (n_th_param = 0; n_th_param < m -> n_params; n_th_param++) {
      if (!strcmp (s, m -> params[n_th_param]->name))
	return TRUE;
    }
  }
  return FALSE;
}

char *rt_expr (MESSAGE_STACK messages, int start_idx, int *end_idx) {

  char *exprbuf;
  int i, struct_expr_limit, subscript_expr_limit;
  int message_of_rcvr, message_of_arg;

  exprbuf = fmt_rt_expr (messages, start_idx, end_idx);
  message_of_rcvr = message_of_arg = FALSE;
  for (i = start_idx; i >= *end_idx; i--) {
    messages[i] -> attrs |= TOK_IS_RT_EXPR;
    if (M_TOK(messages[i]) == LABEL) {
      CVAR *__c;
      /* 
       *  See the comments in rt_self_expr, below.
       *  Should only need to check for receivers that
       *  are method parameters.
       */
      if (is_method_param_name (M_NAME(messages[i])))
	message_of_rcvr = TRUE;
      if (message_of_rcvr) {
	if (global_var_is_declared (M_NAME(messages[i]))
	    || get_local_var (M_NAME(messages[i]))) 
	  message_of_rcvr = FALSE;
	else
	  continue;
      }
      /*
       *   Not receiver, instance or class variable message.
       */
      if ((messages[i] -> obj == NULL) && 
	  (messages[i] -> receiver_obj == NULL)) {
	if (((__c = get_local_var (M_NAME(messages[i]))) != NULL) ||
	    ((__c = get_global_var (M_NAME(messages[i]))) != NULL)) {
	  if (is_struct_or_union_expr (messages, i, 
				       stack_start(messages),
				       get_stack_top (messages))) {
	    generate_register_c_method_arg_call 
	      (__c, 
	       struct_or_union_expr (messages, i, 
				     get_stack_top(messages),
				     &struct_expr_limit),
	       LOCAL_VAR,
	       start_idx);
	    i = struct_expr_limit;
	  } else {
	    if (is_subscript_expr (messages, i, 
				   get_stack_top (messages)) > 0) {
	      generate_register_c_method_arg_call 
		(__c, 
		 subscript_cvar_registration_expr (messages, i, 
						   get_stack_top(messages),
						   &subscript_expr_limit),
		 LOCAL_VAR,
		 start_idx);
	      i = subscript_expr_limit;
	    } else {
	      generate_register_c_method_arg_call (__c, M_NAME(messages[i]),
						   LOCAL_VAR,
						   start_idx);
	    }
	  }
	} else {
	  undefined_label_check (messages, i);
	}
      }
    } else {
      /***/
/*       if (!M_ISSPACE(messages[i])) { */
/* 	message_of_rcvr = FALSE; */
/*       } */
      if (!M_ISSPACE(messages[i])) {
	message_of_rcvr = FALSE;
	/*
	 *  This is necessary almost as a last attempt to warn the
	 *  user in the front end of a syntax error, when the front
	 *  end needs to evaluate an expression lazily, for example an
	 *  object mutation.  So this can certainly be expanded.
	 */
	switch (M_TOK(messages[i]))
	  {
	  case ARGSEPARATOR:
	    message_of_arg = TRUE;
	    break;
	  case EQ:
	    if (message_of_arg)
	      warning (messages[i], 
		       "Possible assignment in method argument.");
	    break;
	  }
      }
    }
  }
  fileout (exprbuf, 0, start_idx);
  for (i = start_idx; i >= *end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  }
  return exprbuf;
}

static char *fn_param_return_trans (MESSAGE *m_orig, CFUNC *cfn, char *expr, int arg_idx) {
  static char arg_expr_buf[MAXMSG];
  char arg_classname[MAXLABEL];
  int __n_th_arg;
  CVAR *__arg;
  if (cfn -> params && IS_CVAR(cfn -> params)) {
    for (__n_th_arg = 0, __arg = cfn -> params; 
	 __n_th_arg < arg_idx; __n_th_arg++) 
      __arg = __arg -> next;
    strcpy (arg_classname, basic_class_from_cvar (m_orig, __arg, 0));
    strcpy (arg_expr_buf, fmt_rt_return (expr, arg_classname, TRUE));
  } else {
    warning 
      (m_orig, 
       "Function, \"%s,\" prototype has no parameter definitions.",
       cfn -> decl);
    warning 
      (m_orig,
       "Parameter class defaulting to, \"Integer.\"\n");
    strcpy 
      (arg_expr_buf, 
       fmt_rt_return (expr, INTEGER_CLASSNAME, TRUE));
  }
  return arg_expr_buf;
}

/*
 *  This should always be in c_argument_context.
 *  TODO - Also call from rt_self_expr, below.
 */
char *rt_obj_arg (MESSAGE_STACK messages, 
		   int arg_start_idx, int *arg_end_idx) {
  int i, stack_end, lookahead, n_parens, arg_idx, frame_start_idx,
    fn_idx, struct_expr_limit;
  MESSAGE *m;
  METHOD *method;
  int n_th_param, match_param, max_param;
  char *expr_ptr, arg_buf[MAXMSG], expr_buf[MAXMSG];
  static char arg_expr_buf[MAXMSG];
  stack_end = get_stack_top (messages);
  for (i = arg_start_idx, n_parens = 0; i > stack_end; i--) {
    m = messages[i];
    lookahead = nextlangmsg (messages, i);
    switch (M_TOK(m))
      {
      case OPENPAREN:
	++n_parens;
	break;
      case CLOSEPAREN:
	--n_parens;
      default:
	if ((M_TOK(messages[lookahead]) == CLOSEPAREN) ||
	    (M_TOK(messages[lookahead]) == ARGSEPARATOR)) {
	  if (!n_parens) {
	    *arg_end_idx = i;
	    goto got_arg_limit;
	  }
	}
	break;
      }
  }
 got_arg_limit:
  expr_ptr = collect_tokens (messages, arg_start_idx, *arg_end_idx);
  de_newline_buf (expr_ptr);
  strcpy (arg_buf, expr_ptr);
  free (expr_ptr);


  // ***
  // 
  // Optimize special cases as we can get to them.
  //
  if (interpreter_pass == method_pass) {

    if (arg_start_idx == *arg_end_idx) {

      // If the argument is a single token that is a method parameter,
      // use __ctalk_arg_internal (n).

      if ((method = new_methods[new_method_ptr+1] -> method) != NULL) {

	max_param = method -> n_params - 1;
	
	for (n_th_param = 0, match_param = FALSE; 
	     (n_th_param < method -> n_params) && !match_param; ++n_th_param) {
	  if (str_eq (arg_buf, method->params[n_th_param] -> name)) {
	    sprintf (expr_buf, "%s (%d)", 
		     METHOD_ARG_ACCESSOR_FN, max_param - n_th_param);
	    match_param = TRUE;
	  }
	}
	if (!match_param) {
	  if (str_eq (M_NAME(messages[arg_start_idx]), "self")) {

	    // "self" alone is nearly the same.  Use 
	    // __ctalk_self_internal ().
	    sprintf (expr_buf, "%s ()", SELF_ACCESSOR_FN);

	  } else {
	    sprintf (expr_buf, "%s (\"%s\")", EVAL_EXPR_FN, arg_buf);
	  }
	}
      } else {
	sprintf (expr_buf, "%s (\"%s\")", EVAL_EXPR_FN, arg_buf);
      }
    } else {
      sprintf (expr_buf, "%s (\"%s\")", EVAL_EXPR_FN, arg_buf);
    }
  } else {
    sprintf (expr_buf, "%s (\"%s\")", EVAL_EXPR_FN, arg_buf);
  }
  /*
   *  NOTE - In a very complex expression, this might need to be the 
   *  frame start of the function or control block, if different 
   *  than the frame of the argument.
   */
  frame_start_idx = 
    frame_at (CURRENT_PARSER -> frame) -> message_frame_top;

  for (i = arg_start_idx; i >= *arg_end_idx; i--) {
    if (M_TOK(messages[i]) == LABEL) {
      CVAR *__c;
      if (((__c = get_local_var (M_NAME(messages[i]))) != NULL) ||
 	  ((__c = get_global_var (M_NAME(messages[i]))) != NULL)) {
	if (is_struct_or_union_expr (messages, i, 
				     stack_start(messages),
				     get_stack_top (messages))) {
	  generate_register_c_method_arg_call 
	    (__c, 
	     struct_or_union_expr (messages, i, 
				   get_stack_top(messages),
				   &struct_expr_limit),
	     LOCAL_VAR,
	     frame_start_idx);
	  i = struct_expr_limit;
	} else {
	  generate_register_c_method_arg_call (__c, M_NAME(messages[i]),
					       LOCAL_VAR,
					       frame_start_idx);
	}
      }
    }
  }

  if ((arg_idx =
         obj_expr_is_arg (messages, arg_start_idx,
  			stack_start (messages),
  			get_stack_top (messages),
  			&fn_idx)) != ERROR) {
    CFUNC *cfn;
    if ((cfn = get_function (M_NAME(messages[fn_idx]))) != NULL) {
      strcpy (arg_expr_buf, 
	      fn_param_return_trans (messages[fn_idx], 
				     cfn, expr_buf, arg_idx));
    }
  } else {
    _warning ("Can't find self argument context in rt_self_arg.\n");
    strcpy (arg_expr_buf, expr_buf);
  }

  fileout (arg_expr_buf, 0, arg_start_idx);
  for (i = arg_start_idx; i >= *arg_end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  }
  return arg_expr_buf;
}

char *rt_self_expr (MESSAGE_STACK messages, int start_idx, int *end_idx) {

  char *exprbuf, arg_classname[MAXLABEL];
  char exprbuf_save[MAXLABEL];
  static char arg_expr_buf[MAXMSG], *msgbuf;
  int i;
  int fn_idx, arg_idx, next_idx;
  int lookback;
  int leading_unary_idx;
  int struct_expr_limit;
  int messages_self_is_receiver_of;
  OBJECT_CONTEXT context;
  CVAR *c, *struct_decl;
  CVAR *struct_defn = NULL;   /* Avoid a warning. */

  if (((leading_unary_idx = prevlangmsg (messages, start_idx)) != ERROR) &&
      (messages[leading_unary_idx] -> attrs & TOK_IS_PREFIX_OPERATOR)) {
    exprbuf = fmt_rt_expr (messages, leading_unary_idx, end_idx);
    /*
     *  Do not free ().
     */
  } else {

    /***/
    /*
     *  If the expression is, "self," alone, use __ctalk_self_internal ().
     */
    if (((next_idx = nextlangmsg (messages, start_idx)) != ERROR) &&
     	METHOD_ARG_TERM_MSG_TYPE(messages[next_idx])) {

      exprbuf = __xalloc (MAXLABEL); // Should be enough
      sprintf (exprbuf, "%s ()", SELF_ACCESSOR_FN);
      *end_idx = start_idx;
      leading_unary_idx = -1;

    } else {

      exprbuf = fmt_rt_expr (messages, start_idx, end_idx);
      leading_unary_idx = -1;
    }
  }

  if (!IS_OBJECT (rcvr_class_obj)) /* Can't do a syntax check without it. */
    return exprbuf;

  next_idx = nextlangmsg (messages, start_idx);
  /*
   *  Syntax check kludge for sequence of labels beginning with
   *  "self," which are not, based only on their own syntax, 
   *  distinguishable from a variable declaration, which means
   *  that the parser could register a message as a C variable
   *  during the var_pass, or in parser_pass a few messages later.  
   *  So check here again for the "self" expression.
   *  TO DO - To get this more accurately, it will be necessary to 
   *  evaluate the expression here, and determine the class of self.  
   *  It probably isn't  practical to evaluate the expression while 
   *  registering C variables.  Even though there's another check 
   *  for messages that are declared C variables, this is faster 
   *  anyway.
   */
  messages_self_is_receiver_of = FALSE;
  for (i = start_idx; i >= *end_idx; i--) {
    messages[i] -> attrs |= TOK_IS_RT_EXPR;
    if (M_ISSPACE(messages[i]))
      continue;
    if (!strcmp (M_NAME(messages[i]), "self")) {
      messages_self_is_receiver_of = TRUE;
      continue;
    }
    if (M_TOK(messages[i]) == LABEL) {
      CVAR *__c;
      if (messages_self_is_receiver_of) {
	if (global_var_is_declared (M_NAME(messages[i]))
	    || get_local_var (M_NAME(messages[i]))) {
	  messages_self_is_receiver_of = FALSE;
	} else {
	  /*
	   * For now, just check if the next message is an
	   * instance or class variable, or a method.
	   * Later, we might need to check method 
	   * arguments here.
	   */
	  /*
	   * Also save the expression in case these calls need
	   * to compile new method.
	   */
	  strcpy (exprbuf_save, exprbuf);
	  if ((i == next_idx) && 
	      !get_class_method (messages[i], rcvr_class_obj,
				 M_NAME(messages[i]), ERROR, FALSE) &&
	      !get_instance_method (messages[i], rcvr_class_obj,
				    M_NAME(messages[i]), 
				    ERROR, FALSE) &&
	      !get_instance_variable (M_NAME(messages[i]), 
				      rcvr_class_obj -> __o_name,
				      FALSE) &&
	      !get_class_variable (M_NAME(messages[i]), 
				   rcvr_class_obj -> __o_name,
				   FALSE) &&
	      !is_method_name (M_NAME(messages[i])) &&
	      !is_method_parameter (messages, i) &&
	      !is_method_proto (rcvr_class_obj -> __o_name,
				M_NAME(messages[i])) &&
	      /* Believe it or not... */
	      strcmp (M_NAME(messages[i]), 
		      new_methods[new_method_ptr+1] -> method -> name) &&
	      strcmp (M_NAME(messages[i]), "super") &&
	      strcmp (M_NAME(messages[i]), "eval")) {
	    msgbuf = collect_tokens (messages, start_idx, *end_idx);
	    unresolved_eval_delay_warning_1 
	      (messages[i], 
	       (interpreter_pass == method_pass ? 
		new_methods[new_method_ptr+1] -> method -> name : NULL),
	       rcvr_class_obj -> __o_name,
	       msgbuf);
	    free (msgbuf);
	    messages_self_is_receiver_of = FALSE;
	    goto check_vars;
	  } else {
	    /***/
	    /* Found something - restore the original expression. */
	    /* This buffer size should be all right. */
	    strcpy (exprbuf, exprbuf_save);
	  }
	  continue;
	}
      }
    check_vars:
      if (((__c = get_local_var (M_NAME(messages[i]))) != NULL) ||
 	  ((__c = get_global_var (M_NAME(messages[i]))) != NULL)) {
	if (is_struct_or_union_expr (messages, i, 
				     stack_start(messages),
				     get_stack_top (messages))) {
	  generate_register_c_method_arg_call 
	    (__c, 
	     struct_or_union_expr (messages, i, 
				   get_stack_top(messages),
				   &struct_expr_limit),
	     LOCAL_VAR,
	     start_idx);
	  i = struct_expr_limit;
	} else {
	  generate_register_c_method_arg_call (__c, M_NAME(messages[i]),
					       LOCAL_VAR,
					       start_idx);
	}
      }
    } else {
      if (!M_ISSPACE(messages[i])) {
	messages_self_is_receiver_of = FALSE;
      }
    }
  }
  if ((arg_idx = 
       obj_expr_is_arg (messages, start_idx, 
			stack_start (messages),
			get_stack_top (messages),
			&fn_idx)) != ERROR) {
    CFUNC *cfn;
    if ((cfn = get_function (M_NAME(messages[fn_idx]))) != NULL) {
      strcpy (arg_expr_buf, 
	      fn_param_return_trans (messages[fn_idx], 
				     cfn, exprbuf, arg_idx));
    } else {
      /*
       *  TO DO - Could not find the function prototype - There could
       *  possibly be another warning about a missing function
       *  prototype here, if the lack of prototype isn't caught 
       *  elsewhere.  Or there should be an exception specifically 
       *  for a missing prototype - then we wouldn't need to repeat 
       *  the message for each function argument.
       */
      strcpy (arg_expr_buf, 
	      fmt_rt_return 
	      (exprbuf, INTEGER_CLASSNAME, TRUE));
    }
  } else {
    context = object_context (messages, start_idx);
    switch (context)
      {
      case c_context:
      case c_argument_context:
	/*
	 *  This is the same procedure that object_context
	 *  uses to look for a C argument context.
	 */
	lookback = start_idx + 1;
	while ((lookback = prevlangmsg (messages, lookback)) != ERROR) {
	  if ((messages[lookback] -> tokentype == LABEL) &&
	      (messages[lookback] -> obj == NULL))
	    break;
	}
	if (lookback == ERROR) {
	  strcpy (arg_expr_buf, exprbuf);
	  goto rt_expr_evaled;
	}
	if (((c = get_local_var (M_NAME(messages[lookback]))) != NULL) ||
	    ((c = get_global_var (M_NAME(messages[lookback]))) != NULL)) {
	  strcpy (arg_classname, 
		  basic_class_from_cvar (messages[start_idx], c, 0));
	  strcpy (arg_expr_buf, fmt_rt_return (exprbuf, arg_classname, TRUE));
	} else {
	  if (is_struct_member (M_NAME(messages[lookback]))) {
	    int lookback2 = lookback;  /* Avoid a warning. */
	    int last_expr_tok;
	    char *agg_expr;
	    last_expr_tok = lookback;
	    while ((lookback = prevlangmsg (messages, lookback)) != ERROR) {
	      if (((lookback2 = prevlangmsg (messages, lookback)) != ERROR) 
		  && ((M_TOK(messages[lookback2]) != PERIOD) ||
		      (M_TOK(messages[lookback2]) != DEREF)))
		break;
	    }
	    if (((struct_decl = 
		 get_local_var (M_NAME(messages[lookback2]))) == NULL) &&
		((struct_decl =
		  get_global_var (M_NAME(messages[lookback2]))) == NULL)) {
	      warning (messages[start_idx], 
		       "Definition of struct or union, \"%s,\" not found.",
		       M_NAME(messages[lookback2]));
	      strcpy (arg_expr_buf, exprbuf);
	    } else {
	      if (struct_decl -> attrs == CVAR_ATTR_STRUCT) {
		if (((struct_defn = get_global_var (struct_decl->type))
		    == NULL) &&
		    ((struct_defn = get_local_var (struct_decl->type))
		     == NULL)) {
		  warning (messages[start_idx], 
		   "Definition of struct or union, \"%s,\" not found.",
		   struct_decl -> type);
		  struct_defn = struct_decl;
		}
	      }
	      agg_expr = collect_tokens (messages, lookback2, last_expr_tok);
	      de_newline_buf (agg_expr);
	      c = struct_member_from_expr (agg_expr, struct_defn);
	      strcpy (arg_classname, 
		      basic_class_from_cvar (messages[start_idx], c, 0));
	      strcpy (arg_expr_buf, 
		      fmt_rt_return (exprbuf, arg_classname, TRUE));
	      free (agg_expr);
	    }
	  } else {
	    int __s, __s2;
	    __s = stack_start (messages);
	    __s2 = get_stack_top (messages);
	    if (is_fmt_arg (messages, start_idx, __s, __s2)) {
	      strcpy 
		(arg_expr_buf, fmt_printf_fmt_arg 
		 (messages, start_idx, __s, __s2, exprbuf));
	    } else {
	      strcpy (arg_expr_buf, exprbuf);
	    }
	  }
	}
	break;
      default:
	strcpy (arg_expr_buf, exprbuf);
	break;
      }      
  }
 rt_expr_evaled:
  fileout (arg_expr_buf, 0, start_idx);
  if (leading_unary_idx != -1) {
    for (i = leading_unary_idx; i >= *end_idx; i--) {
      ++messages[i] -> evaled;
      ++messages[i] -> output;
    }
  } else {
    for (i = start_idx; i >= *end_idx; i--) {
      ++messages[i] -> evaled;
      ++messages[i] -> output;
    }
  }
  return arg_expr_buf;
}

static void comma_op_buf (char *s) {
  char *p;
  while ((p = index (s, ';')) != NULL) *p = ',';
  while ((p = index (s, '\n')) != NULL) *p = ' ';
}

/*
 *  Format a run-time expression, for example, a for loop termination
 *  predicate, with CVAR registration function calls if necessary.
 */
char *fmt_default_ctrlblk_expr (MESSAGE_STACK messages, int start_idx, 
				int end_idx, int keep) {

  static char buf[MAXMSG];
  char *exprbuf, *c99name,
    fn_call_buf[MAXMSG],
    *cfn_buf;
  int i;
  MESSAGE *m;
  CVAR *cvar;
  CFUNC *fn;

  for (i = start_idx, *buf = 0; i >= end_idx; i--) {

    m = messages[i];

    if (m && M_ISSPACE(m)) continue;

    if (M_TOK(m) == LABEL) {
      if (((cvar = get_local_var (M_NAME(m))) != NULL) ||
	  ((cvar = get_global_var (M_NAME(m))) != NULL)) {
	if ((cvar -> attrs & CVAR_ATTR_FN_DECL) ||
	    (cvar -> attrs & CVAR_ATTR_FN_PTR_DECL) ||
	    (cvar -> attrs & CVAR_ATTR_FN_PROTOTYPE)) {
	  /*
	   *  First check if we have a template for the function.
	   *  If so, insert the template in the output, and if 
	   *  necessary, register the template in the initialization
	   *  (though now the function registration by clib_fn_expr ()
	   *  is handled earlier in ctrlblk_pred_rt_expr ()).
	   *  If there's no template, issue a warning.
	   *
	   *  The function should be able to simply do a token 
	   *  replacement here, but later, it might be necessary
	   *  to set the value object, if we need to rescan the
	   *  message.
	   */
	  if ((c99name = c99_name (M_NAME(m), FALSE)) != NULL) {
	    if ((cfn_buf = clib_fn_rt_expr (messages, i)) != NULL) {
	      free (m -> name);
	      m -> name = strdup (cfn_buf);
	    }
	  }
	} else {
	  sprintf (fn_call_buf, "%s ", 
	   fmt_register_c_method_arg_call (cvar, cvar -> name, LOCAL_VAR));
	  comma_op_buf (fn_call_buf);
	  strcat (buf, fn_call_buf);
	}
      } else {
	if ((fn = get_function (M_NAME(m))) != NULL) {
	  if ((c99name = c99_name (M_NAME(m), FALSE)) != NULL) {
	    if ((cfn_buf = clib_fn_rt_expr (messages, i)) != NULL) {
	      free (m -> name);
	      m -> name = strdup (cfn_buf);
	    }
	  }
	}
      }
    }
  }

  exprbuf = collect_tokens (messages, start_idx, end_idx);
  de_newline_buf (exprbuf);
  sprintf (fn_call_buf, "%s (%s, %d)", INT_TRANS_FN, 
 	   fmt_eval_expr (exprbuf), keep);
  free (exprbuf);
  strcat (buf, fn_call_buf);

  return buf;
}

/*
 *  This is much simpler than fmt_default_ctrlblk_expr, because user functions
 *  get a template for each occurrence (in fn_tmpl.c), so we don't 
 *  have to worry about registering arguments.  
 */
char *fmt_user_fn_rt_expr (MESSAGE_STACK messages, 
			   int start_idx, int end_idx, char *expr) {
  static char fn_call_buf[MAXMSG];
  char buf[MAXMSG];
  strcpy (buf, expr);
  de_newline_buf (buf);
  sprintf (fn_call_buf, "__ctalkToCInteger (%s, %d)", 
	   fmt_eval_expr (buf), FALSE);
  return fn_call_buf;
}

extern OBJECT *rcvr_class_obj;

extern I_PASS interpreter_pass;


int undefined_label_check (MESSAGE_STACK messages, int idx) {
  METHOD *method;
  int param_n;
  int is_method_param, is_defined_object, is_method, is_message,
    is_keyword;
  int prev_tok_idx;

  if (interpreter_pass == expr_check)
    return SUCCESS;
  /*
   *  Check if a label is a method argument.
   */
  is_defined_object = is_method_param = is_method = is_message = 
    is_keyword = FALSE;

  if (is_ctalk_keyword (M_NAME(messages[idx])) ||
      is_c_keyword (M_NAME(messages[idx])))
    is_keyword = TRUE;

  if (interpreter_pass == method_pass) {
    method = new_methods[new_method_ptr + 1] -> method;
    is_method_param = FALSE;
    for (param_n = 0; param_n < method -> n_params; param_n++) {
      if (!strcmp (M_NAME(messages[idx]), 
		   method -> params[param_n] -> name)) {
	is_method_param = TRUE;
      }
    }
  }
  if (get_instance_method (messages[idx], rcvr_class_obj, 
			   M_NAME(messages[idx]), 
			   ERROR, FALSE) ||
      get_class_method (messages[idx], rcvr_class_obj, 
			M_NAME(messages[idx]), ERROR, FALSE))
    is_method = TRUE;
  if (get_object (M_NAME(messages[idx]), NULL) ||
      get_class_object (M_NAME(messages[idx])) ||
      !strcmp (M_NAME(messages[idx]), "self"))
    is_defined_object = TRUE;
  if (is_pending_class (M_NAME(messages[idx])))
    is_defined_object = TRUE;
  /*
   *  Anything that isn't accounted for is either an instance
   *  or class variable message, or an undefined label.
   *  TODO - For now, just do a syntax check for instance or
   *  class variable messages - leave until run time
   *  when the receiver is actually determined, or until the
   *  parser can account for constructed objects and actual 
   *  parameter objects, and so on ....
   */
  if (IS_OBJECT(messages[idx]->receiver_obj))
    is_defined_object = TRUE;
  if ((prev_tok_idx = prevlangmsg (messages, idx)) != ERROR) {
    if (M_TOK(messages[prev_tok_idx]) == LABEL)
      is_message = TRUE;
  }
  if (!is_method_param && !is_defined_object && !is_method && !is_message
      && !is_keyword) {
    if (interpreter_pass == method_pass) {
      __ctalkExceptionInternal 
	(messages[idx], 
	 undefined_label_x, 
	 M_NAME(messages[idx]), 
	 message_stack_at (frame_at (parser_at (parser_ptr () + 1) -> frame) -> message_frame_top) -> error_line + messages[idx]->error_line);
    } else {
      int next_idx;
      /* Try to find a little context for the label. */
      if ((next_idx = nextlangmsg (messages, idx)) != ERROR) {
	if ((M_TOK(messages[next_idx]) == LABEL) || 
	    (M_TOK(messages[next_idx]) == METHODMSGLABEL)) {
	  if (is_method_name (M_NAME(messages[next_idx]))) {
	    __ctalkExceptionInternal (messages[idx], 
				      undefined_class_or_receiver_x, 
				      M_NAME(messages[idx]), 0);
	  } else {
	    __ctalkExceptionInternal (messages[idx], undefined_label_x, 
				      M_NAME(messages[idx]), 0);
	  }
	} else {
	  __ctalkExceptionInternal (messages[idx], undefined_label_x, 
				    M_NAME(messages[idx]), 0);
	}
      } else {
	__ctalkExceptionInternal (messages[idx], undefined_label_x, 
				  M_NAME(messages[idx]), 0);
      }
    }
  }
  return SUCCESS;
}

int eval_keyword_expr (MESSAGE_STACK messages, int keyword_ptr) {
  int prev_tok_idx, next_tok_idx, end_idx, i;
  char buf[MAXMSG];
  OBJECT_CONTEXT context;
  if ((next_tok_idx = nextlangmsg (messages, keyword_ptr)) != ERROR) {
    strcpy (buf, fmt_rt_expr (messages, next_tok_idx, &end_idx));
    for (i = keyword_ptr; i >= end_idx; i--) {
      ++messages[i]->evaled;
      ++messages[i]->output;
    }
    if ((context = object_context (messages, next_tok_idx))
	== c_argument_context) {
      prev_tok_idx = keyword_ptr;
      while ((prev_tok_idx = prevlangmsg (messages, prev_tok_idx)) 
	     != ERROR) {
	if ((M_TOK(messages[prev_tok_idx]) == SEMICOLON) ||
	    (M_TOK(messages[prev_tok_idx]) == OPENBLOCK)) {
	  fileout (buf, 0, keyword_ptr);
	  return SUCCESS;
	}
	if (is_printf_fmt (M_NAME(messages[prev_tok_idx]))) {
	  char buf2[MAXMSG];
	  strcpy (buf2, 
		  fmt_printf_fmt_arg (messages, next_tok_idx,
				      stack_start (messages),
				      get_stack_top (messages),
				      buf));
	  fileout (buf2, 0, keyword_ptr);
	  return SUCCESS;
	}
      } 
    }
    fileout (buf, 0, keyword_ptr);
  }
  return SUCCESS;
}

static inline int const_expr_op_precedence (int precedence, MESSAGE_STACK messages, int op_ptr) {

  int retval = FALSE;
  MESSAGE *m_op;
  OP_CONTEXT context;

  m_op = messages[op_ptr];

  if (!IS_C_OP (m_op -> tokentype) ||
      ((context = op_context (messages, op_ptr)) == op_not_an_op))
    return FALSE;

  switch (precedence)
    {
    case 0:
      switch (m_op -> tokentype)
	{
	case OPENPAREN:
	case CLOSEPAREN:
	case ARRAYOPEN:
	case ARRAYCLOSE:
	case INCREMENT:
	case DECREMENT:
	case DEREF:
	  retval = TRUE;
	  break;
	default:
	  break;
	}
      break;
    case 1:
      switch (m_op -> tokentype) 
	{
	case SIZEOF:
	case EXCLAM:
	  retval = TRUE;
	  break;
	case PLUS:      
	case MINUS:     
	case ASTERISK:  
	case AMPERSAND:
	  if (context == op_unary_context)
	    retval = TRUE;
	default: 
	  break;
	}
      break;
    case 2:
      if ((m_op -> tokentype == DIVIDE) ||
	  (m_op -> tokentype == MODULUS))
	retval = TRUE;
      if ((m_op -> tokentype == ASTERISK) && (context == op_binary_context))
	retval = TRUE;
      break;
    case 3:
      if ((m_op -> tokentype == PLUS) ||
	  (m_op -> tokentype == MINUS))
	retval = TRUE;
      break;
    case 4:
      if ((m_op -> tokentype == ASL) ||
	  (m_op -> tokentype == ASR))
	retval = TRUE;
      break;
    case 5:
      if ((m_op -> tokentype == LT) ||
	  (m_op -> tokentype == LE) ||
	  (m_op -> tokentype == GT) ||
	  (m_op -> tokentype == GE))
	retval = TRUE;
      break;
    case 6:
      if ((m_op -> tokentype == BOOLEAN_EQ) ||
	  (m_op -> tokentype == INEQUALITY))
	retval = TRUE;
      break;
    case 7:
      if (m_op -> tokentype == BIT_AND)
	retval = TRUE;
      break;
    case 8:
      if ((m_op -> tokentype == BIT_OR) || 
	  (m_op -> tokentype == BIT_XOR))
	retval = TRUE;
      break;
    case 9:
      if (m_op -> tokentype == BOOLEAN_AND)
	retval = TRUE;
      break;
    case 10:
      if (m_op -> tokentype == BOOLEAN_OR)
	retval = TRUE;
      break;
    case 11:
      if ((m_op -> tokentype == CONDITIONAL) ||
	  (m_op -> tokentype == COLON))
	retval = TRUE;
    case 12:
      if ((m_op -> tokentype == EQ) ||
	  (m_op -> tokentype == ASR_ASSIGN) ||
	  (m_op -> tokentype == ASL_ASSIGN) ||
	  (m_op -> tokentype == PLUS_ASSIGN) ||
	  (m_op -> tokentype == MINUS_ASSIGN) ||
	  (m_op -> tokentype == MULT_ASSIGN) ||
	  (m_op -> tokentype == DIV_ASSIGN) ||
	  (m_op -> tokentype == BIT_AND_ASSIGN) ||
	  (m_op -> tokentype == BIT_OR_ASSIGN) ||
	  (m_op -> tokentype == BIT_XOR_ASSIGN))
	retval = TRUE;
      break;
    case 13:
      if (m_op -> tokentype == ARGSEPARATOR)
	retval = TRUE;
      break;
    default:
      break;
    }

  return retval;
}

OBJECT *const_expr_class (MESSAGE_STACK messages, int start_idx, 
			  int end_idx) {
  int i, j,
    subexpr_start_idx,
    subexpr_end_idx,
    matching_paren_idx,
    precedence = 0,
    binary_op_rcvr_idx,
    binary_op_arg_idx,
    starting_eval_level,
    n_th_param;
  OBJECT *subexpr_class,
    *result_class,
    *token_object;
  CVAR *token_cvar;
  char token_cvar_basic_class[MAXLABEL];
  MESSAGE *m, *m_op,
    *m_binary_op_rcvr;
  METHOD *op_method,
    *self_method;

  starting_eval_level = messages[start_idx] -> evaled;

 re_eval:  for (i = start_idx; i >= end_idx; i--) {
    m = messages[i];
    if (M_ISSPACE(m))
      continue;
    switch (M_TOK(m))
      {
      case LABEL:
	if (!precedence && !m -> obj) {
	  if ((token_object = get_object (M_NAME(m), NULL)) != NULL) {
	    m -> obj = get_class_object (token_object -> __o_classname);
	  } else {
	    if (((token_cvar = get_local_var (M_NAME(m))) != NULL) ||
		((token_cvar = get_global_var (M_NAME(m))) != NULL)) {
	      if (((token_cvar -> type_attrs & CVAR_ATTR_STRUCT_TAG) ||
		  (token_cvar -> type_attrs & CVAR_ATTR_STRUCT_PTR_TAG)) &&
		  (token_cvar -> members == NULL)) {
		CVAR *c_defn, *c_mbr;
		if (((c_defn = get_local_var (token_cvar -> type)) != NULL) ||
		    ((c_defn = get_global_var (token_cvar -> type)) != NULL)) {
		  char *__s;
		  int struct_end_idx;
		  if ((struct_end_idx = 
		       is_struct_expr_b (messages, i,
					 get_stack_top (messages))) > 0) {
		    __s = collect_tokens (messages, i, struct_end_idx);
		    if ((c_mbr = struct_member_from_expr (__s, c_defn)) 
			!= NULL) {
		      strcpy (token_cvar_basic_class, 
			      basic_class_from_cvar 
			      (m, c_mbr, 0));
		    } else {/* if ((c_mbr = struct_member_from_expr ... */
		      strcpy (token_cvar_basic_class, 
			      basic_class_from_cvar 
			      (m, token_cvar, 0));
		    }/* if ((c_mbr = struct_member_from_expr ... */
		    free (__s);
		  } else { /* if ((struct_end_idx =  ... */
		    strcpy (token_cvar_basic_class, 
			    basic_class_from_cvar 
			    (m, token_cvar, 0));
		  }  /* if ((struct_end_idx =  ... */ 
		} else { /* if (((c_defn =  ... */
		  strcpy (token_cvar_basic_class, 
			  basic_class_from_cvar 
			  (m, token_cvar, 0));
		} /* if (((c_defn =  ... */ 
	      } else {/* if(((token_cvar -> type_attrs & CVAR_ATTR_STRUCT_TAG)*/
		if ((token_cvar -> attrs & CVAR_ATTR_STRUCT_DECL) &&
		    token_cvar -> members) {
		  char *__s;
		  int struct_end_idx;
		  CVAR *c_mbr;
		  if ((struct_end_idx = 
		       is_struct_expr_b (messages, i,
					 get_stack_top (messages))) > 0) {
		    __s = collect_tokens (messages, i, struct_end_idx);
		    if ((c_mbr = struct_member_from_expr (__s, token_cvar)) 
			!= NULL) {
		      strcpy (token_cvar_basic_class, 
			      basic_class_from_cvar 
			      (messages[i], c_mbr, 0));
		    } else {/* if ((c_mbr = struct_member_from_expr ... */
		      strcpy (token_cvar_basic_class, 
			      basic_class_from_cvar 
			      (messages[i], token_cvar, 0));
		    }/* if ((c_mbr = struct_member_from_expr ... */
		    free (__s);
		  } else {
		    strcpy (token_cvar_basic_class, 
			    basic_class_from_cvar 
			    (messages[i], token_cvar, 0));
		  }
		} else {/* if ((token_cvar -> attrs & CVAR_ATTR_STRUCT_DECL)*/
		  strcpy (token_cvar_basic_class, 
			  basic_class_from_cvar 
			  (m, token_cvar, 0));
		}/* if ((token_cvar -> attrs & CVAR_ATTR_STRUCT_DECL)*/
	      }/* if (((token_cvar -> type_attrs & CVAR_ATTR_STRUCT_TAG) */
	      m -> obj = get_class_object (token_cvar_basic_class);
	    } else {
	      if (!strcmp (M_NAME(m), "self")) {
		m -> obj = rcvr_class_obj;
	      } else {
		if (interpreter_pass == method_pass) {
		  if ((self_method = new_methods[new_method_ptr + 1]->method)
		      != NULL) {
		    for (n_th_param = 0; n_th_param < self_method -> n_params;
			 n_th_param++) {
		      if (!strcmp 
			  (M_NAME(m), 
			   self_method -> params[n_th_param] -> name)) {
			m -> obj = 
			  get_class_object 
			  (self_method -> params[n_th_param] -> class);
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
	break;
      case LITERAL_CHAR:
	if (!precedence && !m -> obj) {
	  if ((m -> obj = get_class_object (CHARACTER_CLASSNAME)) == NULL) {
	    class_object_search (CHARACTER_CLASSNAME, FALSE);
	    m -> obj = get_class_object (CHARACTER_CLASSNAME);
	  }
	}
	break;
      case LITERAL:
	if (!precedence && !m -> obj) {
	  if ((m -> obj = get_class_object (STRING_CLASSNAME)) == NULL)  {
	    class_object_search (STRING_CLASSNAME, FALSE);
	    m -> obj = get_class_object (STRING_CLASSNAME);
	  }
	}
	break;
      case INTEGER:
      case LONG:
	if (!precedence && !m -> obj) {
	  if ((m -> obj = get_class_object (INTEGER_CLASSNAME)) == NULL) {
	    class_object_search (INTEGER_CLASSNAME, FALSE);
	    m -> obj = get_class_object (INTEGER_CLASSNAME);
	  }
	}
	break;
      case FLOAT:
	if (!precedence && !m -> obj) {
	  if ((m -> obj = get_class_object (FLOAT_CLASSNAME)) == NULL) {
	    class_object_search (FLOAT_CLASSNAME, FALSE);
	    m -> obj = get_class_object (FLOAT_CLASSNAME);
	  }
	}
	break;
      case LONGLONG:
	if (!precedence && !m -> obj) {
	  if ((m -> obj = get_class_object (LONGINTEGER_CLASSNAME)) == NULL) {
	    class_object_search (LONGINTEGER_CLASSNAME, FALSE);
	    m -> obj = get_class_object (LONGINTEGER_CLASSNAME);
	  }
	}
	break;
      default:
	if (const_expr_op_precedence (precedence, messages, i)) {
	  m_op = messages[i];
	  switch (M_TOK(m_op))
	    {
	    case OPENPAREN:
	      if ((matching_paren_idx = match_paren (messages, i, end_idx))
		  == ERROR) {
		return NULL;
	      } else {
		int i_2;
		subexpr_start_idx = nextlangmsg (messages, i);
		subexpr_end_idx = prevlangmsg (messages, matching_paren_idx);
		subexpr_class = const_expr_class (messages, subexpr_start_idx,
						  subexpr_end_idx);

		for (i_2 = i; i_2 >= matching_paren_idx; i_2--) {
		  if (messages[i_2] -> obj == NULL) {
		    messages[i_2] -> obj = subexpr_class;
		  } else {
		    messages[i_2] -> value_obj = subexpr_class;
		  }
		  ++(messages[i_2] -> evaled);
		}
		i = subexpr_end_idx;
	      }
	      break;
	    case PLUS:
	    case MINUS:
	    case MULT:
	    case DIVIDE:
	    case ASL:
	    case ASR:
	    case BIT_AND:
	    case BIT_OR:
	    case BIT_XOR:
	    case BOOLEAN_EQ:
	    case GT:
	    case LT:
	    case GE:
	    case LE:
	    case INEQUALITY:
	    case BOOLEAN_AND:
	    case BOOLEAN_OR:
	    case MODULUS:
	      if (((binary_op_rcvr_idx = 
		   prevlangmsg (messages, i)) != ERROR) &&
		  ((binary_op_arg_idx = 
		    nextlangmsg (messages, i)) != ERROR)) {
		m_binary_op_rcvr = messages[binary_op_rcvr_idx];
		if ((op_method = get_instance_method 
		     (m_binary_op_rcvr, m_binary_op_rcvr->obj,
		      M_NAME(m_op), ERROR, FALSE)) != NULL) {
		  OBJECT *op_result_class;
		  int i_2;
		  op_result_class = 
		    get_class_object (op_method -> returnclass);
		  for (i_2 = binary_op_rcvr_idx; 
		       i_2 >= binary_op_arg_idx;
		       i_2--) {
		    if (messages[i_2] -> obj == NULL) {
		      messages[i_2] -> obj = op_result_class;
		    } else {
		      messages[i_2] -> value_obj = op_result_class;
		    }
		    ++(messages[i_2] -> evaled);
		  }
		}
	      }
	      break;
	    }  /* switch (M_TOK(m_op)) */
	} /* 	if (const_expr_op_precedence (precedence, messages, i)) { */
      } /* switch (M_TOK(m)) */
  }
  for (; precedence < 13; ++precedence) {
    j = 0;
    while (j >= end_idx) {
      if (const_expr_op_precedence (precedence, messages, j) &&
	  (messages[j] -> evaled <= starting_eval_level))
	goto re_eval;
      --j;
    }
  }
  result_class = M_VALUE_OBJ (messages[start_idx]);
  for (i = start_idx; i >= end_idx; i--) {
    m = messages[i];
    if (M_ISSPACE(m)) continue;
    if (IS_OBJECT(m -> obj) && IS_CLASS_OBJECT (m -> obj))
      m -> obj = NULL;
    if (IS_OBJECT(m -> value_obj) && IS_CLASS_OBJECT (m -> value_obj))
      m -> value_obj = NULL;
  }
  return result_class;
}

/*
 *  Simple prefix expression : <op><obj>.
 */
int prefix_method_expr_a (MESSAGE_STACK messages, int prefix_idx,
			  int obj_idx) {
  OBJECT_CONTEXT context;
  METHOD *prefix_method;
  char *expr_buf, result_buf[MAXMSG];
  int i;
  if (messages[obj_idx] -> obj) {
    if ((prefix_method = 
	 get_prefix_instance_method (messages[prefix_idx],
		     get_class_object (messages[obj_idx]->obj -> __o_classname),
			     M_NAME(messages[prefix_idx]),
			     FALSE)) != NULL) {
      context = object_context (messages, prefix_idx);
      switch (context)
	{
	case c_argument_context:
	case c_context:
	  expr_buf = collect_tokens (messages, prefix_idx, obj_idx);
	  de_newline_buf (expr_buf);
	  sprintf (result_buf, "%s (\"%s\")", EVAL_EXPR_FN, 
		   escape_str_quotes (expr_buf));
	  free (expr_buf);
	  fileout
	    (obj_2_c_wrapper_trans 
	     (messages, prefix_idx, messages[prefix_idx],
	      messages[obj_idx]->obj, prefix_method, 
	      result_buf, TRUE),
	      0, prefix_idx);
	     for (i = prefix_idx; i >= obj_idx; i--) {
	       ++messages[i] -> evaled;
	       ++messages[i] -> output;
	     }
	     return SUCCESS;
	  break;
	default:
	  break;
	}
    }
  }
  return ERROR;
}

/* Output a run time eval for a statement of the form 
   <rcvr>  <unknown-label>
   when <unknown-label> isn't defined yet.
*/
void deferred_method_eval_a (MESSAGE_STACK messages, 
			     int start_idx,
			     int *end_idx) {
  char output_buf[MAXMSG];
  int i;

  strcpy (output_buf, fmt_rt_expr (messages, start_idx, end_idx));

  fileout (output_buf, 0, start_idx);

  for (i = start_idx; i >= *end_idx; i--) {
    ++messages[i] -> evaled;
    ++messages[i] -> output;
  }

}

/* 
 * eval_arg () sticks the expression object into
 * the receiver message's value_obj slot.
 */
void output_arg_rt_expr (MESSAGE_STACK messages, int rcvr_ptr, 
			 int method_ptr, METHOD *method) {
  char tmp[MAXMSG];
  MESSAGE *m;
    
  m = messages[method_ptr];

  sprintf (tmp, "%s(\"%s\")", 
	   EVAL_EXPR_FN, 
	   M_VALUE_OBJ(messages[rcvr_ptr]) -> __o_name);
  fileout
    (obj_2_c_wrapper_trans 
     (messages, rcvr_ptr,
      m, m -> receiver_obj, method, tmp, FALSE),
     0, method_ptr);
}

/*
 *  Like method_arg_rt_expr, but checks for a "self" and 
 *  "self value" expression first, in which case it outputs
 *  __ctalk_self_internal () or __ctalk_self_internal_value ().
 */

int method_self_arg_rt_expr (MESSAGE_STACK messages, int idx) {

  int i, expr_end, stack_end;
  int n_subs, n_parens;
  int idx_next, idx_next_2;
  char *tokenbuf, exprbuf[MAXMSG];

  if ((idx_next = nextlangmsg (messages, idx)) != ERROR) {

    if (METHOD_ARG_TERM_MSG_TYPE (messages[idx_next])) {
      sprintf (messages[idx] -> name, "%s ()", SELF_ACCESSOR_FN);
      /***/
      /* strcpy (messages[idx] -> name, exprbuf); */
      messages[idx] -> tokentype = RESULT;
      return SUCCESS;
    } else {

      if ((idx_next_2 = nextlangmsg (messages, idx_next)) != ERROR) {

	if (str_eq (M_NAME(messages[idx_next]), "value") &&
	    METHOD_ARG_TERM_MSG_TYPE (messages[idx_next_2])) {

	  sprintf (messages[idx] -> name, "%s ()", SELF_VALUE_ACCESSOR_FN);
	  /***/
	  /* strcpy (messages[idx] -> name, exprbuf); */
	  strcpy (messages[idx_next] -> name, " ");
	  messages[idx_next] -> tokentype = RESULT;

	  return SUCCESS;
	}

      }

    }

  }

  stack_end = get_stack_top (messages);


  for (i = idx, n_subs = 0, n_parens = 0, expr_end = idx; 
       i >= stack_end; i--) {

    if (METHOD_ARG_TERM_MSG_TYPE (messages[i]) && !n_subs && !n_parens) {
      expr_end = i + 1;
      goto e_e;
    }
    switch (messages[i] -> tokentype)
      {
      case ARRAYOPEN:
	++n_subs;
	break;
      case ARRAYCLOSE:
	--n_subs;
	break;
      case OPENPAREN:
	++n_parens;
	break;
      case CLOSEPAREN:
	--n_parens;
	break;
      }

  }

 e_e:

  tokenbuf = collect_tokens (messages, idx, expr_end);
  strcpy (exprbuf, fmt_eval_expr (tokenbuf));
  free (tokenbuf);

  strcpy (messages[idx] -> name, exprbuf);
  messages[idx] -> tokentype = RESULT;

  for (i = idx - 1; i >= expr_end; i--) {
    strcpy (messages[i] -> name, " ");
    messages[i] -> tokentype = RESULT;
  }

  return SUCCESS;
}

int method_param_arg_rt_expr (MESSAGE_STACK messages, int idx,
			      METHOD *method) {

  int i, expr_end, stack_end;
  int n_subs, n_parens;
  int idx_next, idx_next_2;
  int max_param;
  char *tokenbuf, exprbuf[MAXMSG];

  max_param = method -> n_params - 1;

  for (i = 0; i < method -> n_params; i++) {

    if (str_eq (method -> params[i] -> name, M_NAME(messages[idx]))) {

      if ((idx_next = nextlangmsg (messages, idx)) != ERROR) {

	if (METHOD_ARG_TERM_MSG_TYPE (messages[idx_next])) {
	  sprintf (messages[idx] -> name, "%s (%d)", 
		   METHOD_ARG_ACCESSOR_FN, max_param - i);
	  messages[idx] -> tokentype = RESULT;
	  return SUCCESS;
	} else {

	  if ((idx_next_2 = nextlangmsg (messages, idx_next)) != ERROR) {

	    if (str_eq (M_NAME(messages[idx_next]), "value") &&
		METHOD_ARG_TERM_MSG_TYPE (messages[idx_next_2])) {

	      sprintf (messages[idx] -> name, "%s (%d)", 
		       METHOD_ARG_VALUE_ACCESSOR_FN, max_param - i);
	      strcpy (messages[idx_next] -> name, " ");
	      messages[idx_next] -> tokentype = RESULT;

	      return SUCCESS;
	    }
	  
	  }
	
	}
      
      }
    }
  }

  stack_end = get_stack_top (messages);

  for (i = idx, n_subs = 0, n_parens = 0, expr_end = idx; 
       i >= stack_end; i--) {

    if (METHOD_ARG_TERM_MSG_TYPE (messages[i]) && !n_subs && !n_parens) {
      expr_end = i + 1;
      goto e_e;
    }
    switch (messages[i] -> tokentype)
      {
      case ARRAYOPEN:
	++n_subs;
	break;
      case ARRAYCLOSE:
	--n_subs;
	break;
      case OPENPAREN:
	++n_parens;
	break;
      case CLOSEPAREN:
	--n_parens;
	break;
      }

  }

 e_e:

  tokenbuf = collect_tokens (messages, idx, expr_end);
  strcpy (exprbuf, fmt_eval_expr (tokenbuf));
  free (tokenbuf);

  strcpy (messages[idx] -> name, exprbuf);
  messages[idx] -> tokentype = RESULT;

  for (i = idx - 1; i >= expr_end; i--) {
    strcpy (messages[i] -> name, " ");
    messages[i] -> tokentype = RESULT;
  }

  return SUCCESS;
}

char *simple_param_accessor (MESSAGE_STACK messages, int tok_idx, 
			     METHOD *method) {
  int i, max_param = method -> n_params - 1;
  static char param_buf[MAXMSG], expr_buf[MAXMSG];

  for (i = 0; i < method -> n_params; i++) {

    if (str_eq (method -> params[i] -> name, M_NAME(messages[tok_idx]))) {

      sprintf (param_buf, "%s (%d)", 
	       METHOD_ARG_ACCESSOR_FN, max_param - i);
      if (is_fmt_arg (messages, tok_idx, 
		      stack_start (messages),
		      get_stack_top (messages))) {
	// If it's a printf argument, use the format string character
	// to provide the translation.  Otherwise, use the class of the
	// method parameter.
	strcpy 
	  (expr_buf, fmt_printf_fmt_arg
	   (messages, tok_idx,
	    stack_start (messages), get_stack_top (messages), param_buf));
      } else {
	strcpy 
	  (expr_buf, fmt_rt_return 
	   (param_buf, method->params[i]->class, TRUE));
      }
      fileout (expr_buf, 0, tok_idx);
      ++messages[tok_idx] -> evaled;
      ++messages[tok_idx] -> output;
      return expr_buf;
    }
  }
  return NULL;
}
