/* $Id: typecast_expr.c,v 1.1.1.1 2014/05/01 19:46:36 rkiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2012 Robert Kiesling, rk3314042@gmail.com.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "lex.h"
#include "typeof.h"

static int typecast_states[] = {
  #include "../include/typecaststate.h"
  -1, 0, 0
};

#define TYPECAST_STATE_COLS 3

#define TYPECAST_STATE(x) (__check_state ((x), messages, \
              typecast_states, TYPECAST_STATE_COLS))

int is_typecast_expr (MESSAGE_STACK messages, int start, int *end) {
  int i,
    state,
    n_parens,
    lookahead,
    cast_expr_paren_lvl,
    stack_end,
    paren_delim;
  MESSAGE *m_tok;

  stack_end = get_stack_top (messages);

  if (M_TOK(messages[start]) == OPENPAREN) {
    paren_delim = TRUE;
  } else {
    paren_delim = FALSE;
  }

  for (i = start, n_parens = 0, cast_expr_paren_lvl = -1; i >= stack_end; 
       i--) {

    m_tok = messages[i];

    if (M_ISSPACE (m_tok)) continue;

    /*
     *  This is to handle sizeof () arguments.
     */
    if ((M_TOK(messages[i]) == CLOSEPAREN) &&
	(n_parens == 0) &&
	(paren_delim == FALSE)) {
      int i_1;
      for (i_1 = start; i_1 > i; i_1--) {
	if (M_ISSPACE(messages[i_1])) 
	  continue;
	switch (M_TOK(messages[i_1]))
	  {
	  case LABEL:
	    if (!is_c_data_type (M_NAME(messages[i_1])) &&
		!get_typedef (M_NAME(messages[i_1])))
	      return FALSE;
	    break;
	  case MULT:
	  case ARRAYOPEN:
	  case ARRAYCLOSE:
	    break;
	  default:
	    return FALSE;
	    break;
	  }
      }
      return TRUE;
    }

    if ((state = TYPECAST_STATE (i)) == ERROR)
      return FALSE;

    switch (m_tok -> tokentype)
      {
      case LABEL:
      case CTYPE:
	if (!is_c_data_type (M_NAME(m_tok)) &&
	    !get_typedef (M_NAME(m_tok)) &&
	    !get_global_var (M_NAME(m_tok))) {
	  /*
	   *  NOTE: The var pass can also identify this 
	   *  expression as an incomplete type.
	   *
	   *  TO DO: If it becomes an issue, make note 
	   *  of the unknown type, and only issue a warning
	   *  on exit if the expression syntactically
	   *  would be a type cast otherwise.
	   *
	   *  First, do a quick lookahead for anything not 
	   *  a valid typecast token.
	   */
	  if (n_parens) {
	    int lookahead = i - 1;

	    for (;lookahead >= stack_end; lookahead--) {
	      if ((M_TOK(messages[lookahead]) != OPENPAREN) ||
		  (M_TOK(messages[lookahead]) != CLOSEPAREN) ||
		  !M_ISSPACE(messages[lookahead]) ||
		  (M_TOK(messages[lookahead]) != ASTERISK) ||
		  (M_TOK(messages[lookahead]) != LABEL))
		return FALSE;
	    }

	    if ((M_TOK(messages[lookahead]) == LABEL) ||
		(M_TOK(messages[lookahead]) == CTYPE)) {
	      warning (m_tok, "Unknown type, \"%s,\" in typecast.", 
		       M_NAME(m_tok));
	      break;
	    }
	  }
	}
	break;
      case OPENPAREN:
	++n_parens;
	if ((lookahead = nextlangmsg (messages, i)) != ERROR) {
	  if ((M_TOK(messages[lookahead]) == CTYPE) ||
	      (M_TOK(messages[lookahead]) == LABEL)) {
	    if (is_c_data_type (M_NAME(messages[lookahead])) ||
		/***/
		/* __get_rt_typedef (M_NAME(messages[lookahead]))) */
		__ctalkGetTypedef (M_NAME(messages[lookahead])))
	      cast_expr_paren_lvl = n_parens;
	  }
	}
	break;
      case CLOSEPAREN:
	if (n_parens == cast_expr_paren_lvl) {
	  *end = i;
	  return TRUE;
	}
	--n_parens;
	break;
      default:
	break;
      }
  }

  return FALSE;

}

extern CVAR *fn_param_decls,
  *fn_param_decls_ptr;

char *basic_class_from_typecast (MESSAGE_STACK messages, int start,
				int end) {
  CVAR c;
  /*
   *  A typecast should be parseable like a function parameter
   *  prototype.
   */
  fn_params (messages, start, end);
  c = get_param_n (0);
  return basic_class_from_cvar (messages[start], &c, 0);
}
